-- NetDB_Access.sql --- Access security group related functions and procedures.

create or replace package body NetDB_Access as

    -- Checks if the user is active and has a particular type of access.
    function check_access ( p_user_id               in int,
                            p_access_type           in varchar,
                            p_record_id             in int     := null,
                            p_check_all_record_flag in boolean := true ) return boolean
    as
        v_access_type varchar(32) := NetDB_RecordType.access (p_access_type);
        v_user_access int         := 0;
    begin
        -- wtf? inactive users should never get this far.
        if (not NetDB_User.is_active (p_user_id)) then return false; end if;

        -- If the access type passed, maps to a null, everyone has access to it.
        if (v_access_type is null) then return true; end if;

        -- check for 'godhood', if we're honoring it, and
        -- check the access type if it is not System Management...
        if ( p_check_all_record_flag and NetDB_User.has_all_records (p_user_id) and
             v_access_type != 'systemmanagement' )
        then
            return true;
        end if;

        -- see if they have the specific access
        select count (*) into v_user_access
          from Users_AccessGroup
          join AccessGroup on AccessGroup.id = Users_AccessGroup.accessgroup
         where lower(AccessGroup.label) = v_access_type and
               Users_AccessGroup.userid = p_user_id;

        if (v_user_access = 0) then return false; end if;

        -- check sub-type access for records that have sub-types.
        if (p_record_id is not null) then
            case lower (v_access_type)
                when 'node' then
                    select count (*) into NetDB.existence_counter from NodeType
                      join Node_NodeType on Node_NodeType.type = NodeType.type
                     where Node_NodeType.node = p_record_id and
                           NodeType.accessgroup is not null;

                    if (NetDB.existence_counter > 0) then
                        select count (*) into v_user_access from Users_AccessGroup
                          join NodeType      on NodeType.accessgroup = Users_AccessGroup.accessgroup
                          join Node_NodeType on Node_NodeType.type   = NodeType.type
                         where Node_NodeType.node = p_record_id and
                               p_user_id = Users_AccessGroup.userid;

                        -- The user must have access to all sub-types.
                        if (NetDB.existence_counter != v_user_access) then
                            v_user_access := 0;
                        end if;
                    end if;

                else null;
            end case;
        end if;

        return Util.to_boolean (v_user_access);
    end check_access;


    -- A convenient wrapper for check_access.
    -- If the user doesn't have access, raise NetDB.ex_no_access_to_record_type.
    procedure has_access_or_die ( p_user_id               in int,
                                  p_record_type           in varchar,
                                  p_record_id             in int     := null,
                                  p_check_all_record_flag in boolean := true )
    as
        err_msg varchar(255);
    begin
        if (not check_access (p_user_id, p_record_type, p_record_id, p_check_all_record_flag)) then
            err_msg := 'The user "' || NetDB_User.handle (p_user_id) || '" does not have access to ';

            if (p_record_id is not null) then
                err_msg := err_msg || 'the ' || NetDB_RecordType.display (p_record_type) ||
                                      ' "'   || NetDB_Util.handle (p_record_id, p_record_type, true) || '".';
            else
                err_msg := err_msg || '"' || NetDB_RecordType.display (p_record_type) || '" records.';
            end if;

            raise_application_error (NetDB.ex_no_access_to_record_type, err_msg);
        end if;
    end has_access_or_die;


    -- Returns a list of access security groups.
    function list return NetDB.ref_cursor
    as
        out_record  result_set       := result_set ();
        ret_cursor  NetDB.ref_cursor := null;
    begin
        for r in (select * from AccessGroup)
        loop
            out_record.extend(1);
            out_record(out_record.COUNT) := result_row ( 'key', r.label, null,
                                                         'string', r.display_label, null, null );
        end loop;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    function verify_or_die ( p_access_id in int     := null,
                             p_label     in varchar := null ) return int
    as
        v_identifier varchar(256) := p_label;
        v_access_id  int;
    begin
        if (p_access_id is null and p_label is null) then
            NetDB_Util.throw_no_record_id_handle ();
        elsif (p_access_id is not null and p_label is not null) then
            select id into v_access_id from AccessGroup
             where id = p_access_id and lower (label) = lower (p_label);
        elsif (p_label is not null) then
            select id into v_access_id from AccessGroup where lower (label) = lower (p_label);
        else -- p_access_id is not null
            v_identifier := p_access_id;

            select id into v_access_id from AccessGroup where id = p_access_id;
        end if;

        return v_access_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('access', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('access');
    end verify_or_die;


    -- Returns the label for an access group.
    function handle (p_access_id in int) return varchar
    as
        v_label AccessGroup.label%TYPE;
    begin
        select label into v_label from AccessGroup where id = p_access_id;

        return v_label;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('access', p_access_id);
    end handle;


    -- main (interfacing) function to manipulate AccessGroup table
    -- possible p_action values are {'list'}
    function master ( p_action         in varchar,
                      p_op_uid         in int     := null,
                      p_label          in varchar := null,
                      p_label_new      in varchar := null,
                      description      in varchar := null,
                      p_default_access in varchar := null,
                      p_display_label  in varchar := null ) return NetDB.ref_cursor
    as
        ret_cursor NetDB.ref_cursor := null;
    begin
        case lower (p_action)
            when 'list' then
                ret_cursor := list ();
            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;

        if (ret_cursor is null) then
            ret_cursor := NetDB_Util.get_null_generic_cursor ();
        end if;

        return ret_cursor;
    end master;

end NetDB_Access;
/
