-- NetDB_DHCPservice.sql --- contains functions related to dhcp services

create or replace package body NetDB_DHCPservice as

    sys_time_stamp timestamp;

    -- Generates a list of all DHCP services and stuffs it in a generic cursor.
    function list return NetDB.ref_cursor
    as
        v_label   DHCPservice.label%TYPE       := null;
        v_comment DHCPservice.description%TYPE := null;
        v_id      DHCPservice.id%TYPE          := 0;
        v_end     int := 1;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;

        cursor data_cursor is
            select label, id, description from DHCPservice;
    begin
        open data_cursor;
        loop
            fetch data_cursor into v_label, v_id, v_comment;
            exit when data_cursor%notfound;

            out_record.extend(4);
            out_record(v_end + 0) := result_row ('data', v_label,  0,    null,     null,      null, null);
            out_record(v_end + 1) := result_row ('key', 'name',    null, 'string', v_label,   null, null);
            out_record(v_end + 2) := result_row ('key', 'id',      null, 'string', v_id,      null, null);
            out_record(v_end + 3) := result_row ('key', 'comment', null, 'string', v_comment, null, null);

            v_end := v_end + 4;
        end loop;
        close data_cursor;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;

    -- Unused and untested.
    procedure del (p_service_id in int) as
    begin
        delete from DHCPservice where id = p_service_id;
    end del;


    -- Unused and untested.
    function ins ( p_op_uid in int,
                   p_label  in varchar,
                   p_desc   in varchar ) return NetDB.ref_cursor
    as
        v_id int := 0;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin
        NetDB_Util.validate (p_label, 'label');
        NetDB_Util.validate (p_desc,  'description');

        if (p_label is null or p_op_uid is null) then
            raise_application_error (NetDB.ex_bad_parameter,
                'DCHP services must have both a label and record of the user who last modified it.');
        end if;

        insert into DHCPservice (label,   description, modified_by, modified)
                         values (p_label, p_desc,      p_op_uid,    sys_time_stamp)
        returning id into v_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'id', 0, 'string', v_id, null, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Unused and untested.
    procedure upd ( p_op_uid     in int,
                    p_service_id in int,
                    p_label_new  in varchar,
                    p_desc       in varchar )
    as
        v_label DHCPservice.label%TYPE       := null;
        v_desc  DHCPservice.description%TYPE := null;
    begin
        NetDB_Util.validate (p_label_new, 'label');
        NetDB_Util.validate (p_desc, 'description');

        select label, description into v_label, v_desc from DHCPservice where id = p_service_id;

        update DHCPservice set label       = coalesce (p_label_new, v_label),
                               description = p_desc,
                               modified_by = p_op_uid,
                               modified    = sys_time_stamp
         where id = p_service_id;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- verify that a DCHP service exists.
    -- Unused and untested.
    function verify_or_die ( p_service_id in int     := null,
                             p_label      in varchar := null ) return int
    as
        v_identifier varchar(256) := p_label;
        v_service_id int          := 0;
    begin
        if (p_service_id is null and p_label is null) then
            NetDB_Util.throw_no_record_id_handle ();
        elsif (p_service_id is not null and p_label is not null) then
            select id into v_service_id from DHCPservice
             where lower (label) = lower (p_label) and id = p_service_id;
        elsif (p_label is not null) then
            select id into v_service_id from DHCPservice where lower (label) = lower (p_label);
        else -- p_service_id is not null
            v_identifier := p_service_id;

            select id into v_service_id from DHCPservice where id = p_service_id;
        end if;

        return v_service_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('dhcpservice', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('dhcpservice');
    end verify_or_die;


    -- Unused and untested.
    function handle (p_service_id in int) return varchar
    as
        v_service DHCPservice.label%TYPE := null;
    begin
        select label into v_service from DHCPservice where id = p_service_id;

        return v_service;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('dhcpservice', p_service_id);
    end handle;


    -- interfacing function to manipulate content of DHCPservice table
    -- Unused and untested.
    function master ( p_action      in varchar,
                      p_op_uid      in int     := null,
                      p_service_id  in int     := null,
                      p_label       in varchar := null,
                      p_label_new   in varchar := null,
                      p_description in varchar := null ) return NetDB.ref_cursor
    as
        v_action      varchar(16) := lower (p_action);
        v_service_id  int         := 0;

        ret_cursor NetDB.ref_cursor := null;
    begin
        sys_time_stamp := systimestamp;

        if (v_action in ('delete', 'update')) then
            v_service_id := verify_or_die (p_service_id, p_label);
        end if;

        if (v_action in ('insert', 'update', 'delete')) then
            NetDB_Access.has_access_or_die (p_op_uid, 'dhcpservice');
        end if;

        case v_action
            when 'delete'then
                del (v_service_id);

            when 'insert' then
                ret_cursor := ins (p_op_uid, p_label, p_description);

            when 'list' then
                ret_cursor := list ();

            when 'update' then
                upd (p_op_uid, v_service_id, p_label_new, p_description);

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;

        if (ret_cursor is null) then
            ret_cursor := NetDB_Util.get_null_generic_cursor ();
        end if;

        return ret_cursor;
    end master;

end NetDB_DHCPservice;
/
