-- NetDB_IPaddress.sql --- functions dealing with IP addresses

create or replace package body NetDB_IPaddress as

    -- Reserve IP addresses
    function reserve ( p_op_uid     in int,
                       p_start      in varchar,
                       p_count      in int     := 1,
                       p_exact      in int     := 0,
                       p_contiguous in int     := 0 ) return NetDB.ref_cursor
    as
        start_ip IPaddress.ipaddress%TYPE := IP.normalize (IP.host (p_start));
        limit    IPaddress.ipaddress%TYPE := null;
        time     IPaddress.reserved%TYPE  := systimestamp;

        ias_id       int := NetDB_IPaddressSpace.containing (start_ip);
        net_id       int := NetDB_IPaddressSpace.get_parent_id (ias_id);
        is_god   boolean := NetDB_Owner.check_access (p_op_uid, net_id, 'Network') and
                            NetDB_Access.check_access (p_op_uid, 'Network');

        exact    boolean := Util.to_boolean (p_exact);
        contig   boolean := Util.to_boolean (p_contiguous);
        mid_res  boolean := Util.to_boolean (NetDB_Default.get_int ('Reserve_IPv4_Midrange'));
        skip_res boolean := mid_res;  -- if true, skip midrange reserved addresses

        ias_record IPaddressSpace%ROWTYPE;

        ips_reserved int := 0;
        ip_id        int := null;
        timeout      int := NetDB_Default.get_int('LockTimeOut');

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin
        -- test for ownership access
        NetDB_Owner.has_access_or_die (p_op_uid, ias_id, 'addressspace');

        -- individual address reservations may be refreshed, so try that
        --   now and maybe skip running the rest of this function...
        if (p_count = 1) then
            update IPaddress set reserved = time
             where ipaddress = start_ip and reserved_by = p_op_uid and
                   time - reserved < NUMTODSINTERVAL (timeout, 'second')
            returning id into ip_id;

            if (ip_id is not null) then
                out_record.extend(2);
                out_record(1) := result_row ('data', 'ip_addresses',      0,  null,       null, null, null);
                out_record(2) := result_row ('key',   IP.host (start_ip), 0, 'timestamp', null, time, null);
                goto return_cursor;
            end if;
        end if;

        select * into ias_record from IPaddressSpace where id = ias_id;

        -- is the starting address the address space network address?
        if (start_ip = ias_record.address) then
            if (exact) then
                raise_application_error (NetDB.ex_cannot_reserve_address,
                    'The address "' || IP.host (ias_record.address) ||
                    '" is the base address of the address space "' || ias_record.prefix_txt ||
                    '" and as such cannot be reserved.');
            end if;

            start_ip := IP.plus (ias_record.lowres, 1);

        -- is the starting address the address space broadcast address?
        elsif (start_ip = ias_record.broadcast) then
            raise_application_error (NetDB.ex_cannot_reserve_address,
                'The address "' || IP.host (ias_record.broadcast) ||
                '" is the broadcast address of the address space "' || ias_record.prefix_txt ||
                '" and as such cannot be reserved.');

        -- is the starting address in the address space low reserve area?
        elsif (start_ip <= ias_record.lowres) then
            if (is_god) then
                exact    := true;
                skip_res := false;
            else
                if (exact) then
                    raise_application_error (NetDB.ex_cannot_reserve_address,
                        'The user "' || NetDB_User.handle (p_op_uid) ||
                        '" cannot allocate addresses in the low reserve region, ' ||
                        IP.plus (ias_record.address, 1) || ' - ' || IP.host (ias_record.lowres) || '.');
                end if;

                start_ip := IP.plus (ias_record.lowres, 1);
            end if;

        -- is the starting address in the address space high reserve area?
        elsif (start_ip >= ias_record.highres) then
            if (not is_god) then
                raise_application_error (NetDB.ex_cannot_reserve_address,
                    'The user "' || NetDB_User.handle (p_op_uid) ||
                    '" cannot allocate addresses in the high reserve region, ' ||
                    IP.host (ias_record.highres) || ' - ' || IP.plus (ias_record.broadcast, -1) || '.');
            end if;

            exact    := true;
            skip_res := false;

        -- is the starting address in the address space middle reserve area?
        elsif (mid_res and NetDB_IPaddressSpace.is_MidRes (start_ip)) then
            if (is_god) then
                exact    := true;
                skip_res := false;
            else
                if (exact) then
                    raise_application_error (NetDB.ex_cannot_reserve_address,
                        'The user "' || NetDB_User.handle (p_op_uid) ||
                        '" cannot allocate addresses in mid reserve regions.');
                end if;

                start_ip := NetDB_IPaddressSpace.after_MidRes (start_ip);
            end if;

        else -- the starting address must be in the balance of the address space
            null;
        end if;

        -- establish ending address
        if (is_god and not skip_res) then limit := ias_record.broadcast;
        else                              limit := ias_record.highres;
        end if;

        -- nuke stale reservations so they don't get in the way
        delete from IPaddress
         where reserved is not null and -- very important/necessary for speed!
               Util.datediff ('seconds', reserved) > timeout;

        -- loop through the address space from start to limit looking for count free addresses
        start_ip := IP.normalize (start_ip);
        limit    := IP.normalize (limit);

        out_record.extend(p_count + 1);
        out_record(1) := result_row ('data', 'ip_addresses', 0, null, null, null, null);

        while (start_ip < limit and ips_reserved < p_count)
        loop
            begin
                insert into IPaddress (ipaddress, reserved_by, reserved)
                               values (start_ip,  p_op_uid,    time)
                returning id into ip_id;
            exception
                when NetDB.unique_constraint_failed then
                    ip_id := null;
            end;

            if (ip_id is not null) then
                ips_reserved := ips_reserved + 1;

                out_record(ips_reserved + 1) := result_row ('key', IP.host (start_ip), 0, 'timestamp', null, time, null);

            elsif (exact and ips_reserved = 0) then
                raise_application_error (NetDB.ex_cannot_reserve_address,
                    'The address "' || IP.host (start_ip) || '" could not be reserved.');

            elsif (contig and ips_reserved > 0) then
                raise_application_error (NetDB.ex_cannot_reserve_address,
                    'Could not reserve ' || p_count ||
                    ' contiguous addresses starting at "' || IP.host (p_start) || '".');
            end if;

            -- Get the next address
            start_ip := IP.normalize (IP.plus (start_ip, 1));

            -- if necessary, adjust for midrange reserved zones
            if (skip_res and mid_res and NetDB_IPaddressSpace.is_MidRes (start_ip)) then
                start_ip := IP.normalize (NetDB_IPaddressSpace.after_MidRes (start_ip));
            end if;
        end loop;

        -- did it work?
        if (ips_reserved != p_count) then
            delete from IPaddress where reserved_by = p_op_uid and reserved = time;

            raise_application_error (NetDB.ex_cannot_reserve_address,
                'The requested number of addresses is greater than the number of available addresses.');
        end if;

      <<return_cursor>>
        -- return the set of addresses reserved and the reservation timestamp
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end reserve;


    procedure accept ( p_op_uid        in int,
                       p_address       in varchar,
                       p_active        in int := 1,
                       p_PTRpref       in int := 0,
                       p_referent_id   in int,
                       p_lock          in timestamp,
                       p_referent_type in varchar )
    as
        v_rec_id   int          := NetDB_Util.verify_or_die (p_referent_id, null, p_referent_type, null, true);
        v_type     varchar(256) := lower (p_referent_type);
        v_reserved timestamp;
    begin
        -- Verify that the user has reserved the address being accepted and that is has not expired.
        select reserved into v_reserved from IPaddress
         where reserved_by = p_op_uid and ipaddress = IP.normalize (p_address);

        if (Util.datediff ('second', v_reserved) > NetDB_Default.get_int ('LockTimeOut')) then
            raise_application_error (NetDB.ex_record_not_reserved,
                'The address "' || IP.text (p_address) || '" has expired.');
        end if;

        NetDB_Util.locked_or_die (p_op_uid, p_lock, p_referent_id, p_referent_type);

        update IPaddress
           set ipcnode        = case v_type when 'node'         then p_referent_id else null end,
               interface      = case v_type when 'interface'    then p_referent_id else null end,
               ipaddressspace = case v_type when 'addressspace' then p_referent_id else null end,
               active         = p_active,
               PTRpref        = p_PTRpref,
               reserved_by    = null,
               reserved       = null
         where reserved_by = p_op_uid and ipaddress = IP.normalize (p_address);
    exception
        when NO_DATA_FOUND then
            raise_application_error (NetDB.ex_record_not_reserved,
                'The user "' || NetDB_User.handle (p_op_uid) || '" has not reserved the address "' ||
                IP.text (p_address) || '".');
    end accept;


    -- Cancel reservation of a single address for a user.
    procedure cancel ( p_op_uid  in int,
                       p_address in varchar ) as
    begin
        delete from IPaddress where reserved_by = p_op_uid and ipaddress = IP.normalize (p_address);
    exception
        when others then
            null;
    end cancel;


    function closest_name (p_ipaddress in varchar) return varchar
    as
        result Name.fullname%TYPE;
        ipnorm IP.address := IP.normalize (p_ipaddress);
    begin
        -- be reflexive with nulls
        if (ipnorm is null) then return null; end if;

        -- ipaddress name?
        begin
            select fullname into result from Name
             where Name.ipaddress = (select id from IPaddress where ipaddress = ipnorm);

            return result;
        exception
            when NO_DATA_FOUND then
                null;
        end;

        -- interface name?
        begin
            select fullname into result from Name, IPaddress
             where Name.interface = IPaddress.interface and IPaddress.ipaddress = ipnorm;

            return result;
        exception
            when NO_DATA_FOUND then
                null;
        end;

        -- has to be a node name
        begin
            select fullname into result from Name
             where DNSlabel    = 'A' and
                   fullname_lc = ( select min (fullname_lc) from Name, IPaddress, Interface
                                    where Name.node           = Interface.node      and
                                          Interface.id        = IPaddress.interface and
                                          IPaddress.ipaddress = ipnorm );
            return result;
        exception
            when NO_DATA_FOUND then
                return null;
        end;
    end closest_name;


    procedure locked_or_die ( p_op_uid     in int,
                              p_lock       in timestamp,
                              p_address_id in int )
    as
        ip_row IPaddress%ROWTYPE;
    begin
        select * into ip_row from IPaddress where id = p_address_id;

        if (ip_row.ipcnode is not null) then
            NetDB_Node.locked_or_die (p_op_uid, p_lock, ip_row.ipcnode);
        elsif (ip_row.interface is not null) then
            NetDB_Interface.locked_or_die (p_op_uid, p_lock, ip_row.interface);
        elsif (ip_row.ipaddressspace is not null) then
            NetDB_IPaddressSpace.locked_or_die (p_op_uid, p_lock, ip_row.ipaddressspace);
        else
            raise_application_error (NetDB.ex_record_not_found,
                'The IP address "' || IP.host (handle (p_address_id)) || '" is not part of any NetDB record.');
        end if;
    end locked_or_die;


    function info ( p_address_id       in int := null,
                    p_node_id          in int := null,
                    p_interface_id     in int := null,
                    p_address_space_id in int := null ) return result_set pipelined
    as
        v_id      int;
        the_query NetDB.ref_cursor;
        ip_row    IPaddress%ROWTYPE;
    begin
        if (p_address_id is not null) then
            v_id := verify_or_die (p_address_id);

            open the_query for
                select * from IPaddress where id = v_id;

        elsif (p_node_id is not null) then
            v_id := NetDB_Node.verify_or_die (p_node_id);

            open the_query for
                select * from IPaddress where ipcnode = v_id;

        elsif (p_interface_id is not null) then
            v_id := NetDB_Interface.verify_or_die (p_interface_id);

            open the_query for
                select * from IPaddress where interface = v_id;

        elsif (p_address_space_id is not null) then
            v_id := NetDB_IPaddressSpace.verify_or_die (p_address_space_id);

            open the_query for
                select * from IPaddress where ipaddressspace = v_id;

        else
            raise_application_error (NetDB.ex_bad_parameter, 'All parameters cannot be null.');
        end if;

        loop
            fetch the_query into ip_row;

            exit when the_query%notfound;

            pipe row (result_row ('data', ip_row.ipaddress_txt,    1,  null,     null,                    null, null));
            pipe row (result_row ('key', 'active?',             null, 'boolean', null,                    null, ip_row.active));
            pipe row (result_row ('key', 'ip_address',          null, 'string',  ip_row.ipaddress_txt,    null, null));
            pipe row (result_row ('key', 'PTRpref',             null, 'string',  to_char(ip_row.PTRpref), null, null));

            for n in ( select *
                         from table (cast (NetDB_Name.names (ip_row.id, null, null, null, 2) as result_set)) )
            loop
                pipe row (result_row ( n.type, n.key, n.hash_level,
                                       n.value_type, n.value_string, n.value_timestamp, n.value_int ));
            end loop;
        end loop;

        close the_query;
    end info;



    procedure del ( p_op_uid     in int,
                    p_lock       in timestamp,
                    p_address_id in int )
    as
        v_as_id int := NetDB_IPaddressSpace.containing (handle (p_address_id));
    begin
        -- the user needs to have access to the address space to delete the address.
        NetDB_Owner.has_access_or_die (p_op_uid, v_as_id, 'addressspace');
        locked_or_die (p_op_uid, p_lock, p_address_id);

        delete from IPaddress where id = p_address_id;
    end del;


    procedure upd ( p_op_uid     in int,
                    p_lock       in timestamp,
                    p_address_id in int,
                    p_active     in int,
                    p_PTRpref    in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_address_id);

        update IPaddress set active  = p_active,
                             PTRpref = p_PTRpref where id = p_address_id;
    end upd;


    function verify_or_die ( p_address_id in int     := null,
                             p_address    in varchar := null) return int
    as
        v_identifier varchar(256) := IP.host (p_address);
        v_address_id int          := 0;
    begin
        if (p_address_id is null and p_address is null) then
            NetDB_Util.throw_no_record_id_handle ();

        elsif (p_address_id is not null and p_address is not null) then
            select id into v_address_id from IPaddress
             where (ipaddress = p_address or ipaddress_txt = p_address) and
                    id = p_address_id;

        elsif (p_address is not null) then
            select id into v_address_id from IPaddress
             where ipaddress = p_address or ipaddress_txt = p_address;

        else -- p_address_id is not null
            v_identifier := p_address_id;

            select id into v_address_id from IPaddress where id = p_address_id;
        end if;

        return v_address_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('address', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('address');
    end verify_or_die;


    -- Get the normalized ip address for an id.
    function handle ( p_address_id  in int,
                      user_readable in boolean := false ) return varchar
    as
        v_address IPaddress.ipaddress%TYPE;
    begin
        select ipaddress into v_address from IPaddress where id = p_address_id;

        if (user_readable) then
            v_address := IP.host (v_address);
        end if;

        return v_address;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('address', p_address_id);
    end handle;

    function master ( p_action           in varchar,  -- {delete, info, update}
                      p_op_uid           in int       := null,
                      p_lock             in timestamp := null,
                      p_address_id       in int       := null,
                      p_address          in varchar   := null,
                      p_active           in int       := 1,
                      p_PTRpref          in int       := 0,
                      p_node_id          in int       := null,
                      p_interface_id     in int       := null,
                      p_address_space_id in int       := null ) return NetDB.ref_cursor
    as
        v_action     varchar(16) := lower (p_action);
        v_address_id int;

        ret_cursor NetDB.ref_cursor := null;
    begin
        if (v_action in ('delete', 'info', 'update')) then
            if (v_action in ('delete', 'update') or
                (v_action = 'info' and p_address is not null or p_address_id is not null) )
            then
                v_address_id := verify_or_die (p_address_id, p_address);
            end if;

            if (v_action in ('update', 'delete')) then
                NetDB_Owner.has_access_or_die (p_op_uid, v_address_id, 'address');
            end if;
        end if;

        -- Perform the requested action.
        case v_action
            when 'delete' then
                del (p_op_uid, p_lock, v_address_id);

            when 'info' then
                open ret_cursor for
                    select *
                      from table (cast (
                           info (v_address_id, p_node_id, p_interface_id, p_address_space_id)
                        as result_set ));

            when 'update' then
                upd (p_op_uid, p_lock, v_address_id, p_active, p_PTRpref);

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;

        if (ret_cursor is null) then
            ret_cursor := NetDB_Util.get_null_generic_cursor ();
        end if;

        return ret_cursor;
    end master;

end NetDB_IPaddress;
/
