-- NetDB_IPaddressSpace.sql --- functions related to IPaddressSpace table

create or replace package body NetDB_IPaddressSpace as

    -- Move an address space from its current network to a new network.
    procedure move ( p_op_uid       in int,
                     p_lock         in timestamp,
                     p_ias_id       in int     := null,
                     p_prefix       in varchar := null,
                     p_network_id   in int     := null,
                     p_network_name in varchar := null,
                     p_net_lock     in timestamp )
    as
        v_net_id    int := NetDB_Network.verify_or_die (p_network_id, p_network_name);
        v_ias_id    int := verify_or_die (p_ias_id, p_prefix);
        v_parent_id int := get_parent_id (v_ias_id);
    begin
        NetDB_Network.locked_or_die (p_op_uid, p_lock, v_parent_id);
        NetDB_Network.locked_or_die (p_op_uid, p_net_lock, v_net_id);

        update IPaddressSpace set network = v_net_id where id = v_ias_id;

        update Network set modified    = systimestamp,
                           modified_by = p_op_uid
         where id in (v_parent_id, v_net_id);

        NetDB_Log.ins (p_op_uid, 'update', v_net_id, 'network');
        NetDB_Log.ins (p_op_uid, 'update', v_parent_id, 'network');
    end move;


    -- Test if an IP address in a mid reserved area.
    -- Currently, only x.x.x.255, x.x.x.0 and x.x.x.1 are mid res addresses.
    function is_MidRes (p_address in varchar) return boolean as begin
        return Util.to_boolean (regexp_instr (IP.normalize (p_address), '\.(000|001|255)$'));
    end is_MidRes;


    -- Returns the address immediately after the mid res which the IP address is in.
    -- If the address is not in a mid res range, return the address.
    function after_MidRes (p_address in varchar) return varchar
    as
        v_address IPaddress.ipaddress%TYPE := p_address;
    begin
        while (is_MidRes (v_address))
        loop
            v_address := IP.plus (v_address, 1);
        end loop;

        return v_address;
    end after_MidRes;


    -- Find the address space containing an IP address.
    function containing (p_ip in varchar) return int
    as
        v_ias_id int;
    begin
        select id into v_ias_id from IPaddressSpace
         where IP.normalize (p_ip) between address and broadcast;

        return v_ias_id;
    exception
        when NO_DATA_FOUND then
            raise_application_error (NetDB.ex_record_not_found,
                'No address space contains the IP "' || IP.host (p_ip) || '".');
    end containing;


    -- Make sure no other networks contain address spaces overlapping a prefix.
    procedure validate_join ( p_prefix     in varchar,
                              p_network_id in int     := null,
                              p_network    in varchar := null ) as
    begin
        select count (*) into NetDB.existence_counter from IPaddressSpace
         where network != NetDB_Network.verify_or_die (p_network_id, p_network) and
               address between IP.normalize (IP.host (p_prefix)) and IP.normalize (IP.broadcast (p_prefix));

        if (NetDB.existence_counter > 0) then
            raise_application_error ( NetDB.ex_record_not_found,
                'The address space "' || IP.text (p_prefix) ||
                '" overlaps with address spaces in other networks.' );
        end if;
    end validate_join;


    -- Make sure the base and broadcast addresses of a prefix are not in use.
    procedure validate_split (p_prefix in varchar) as
    begin
        select count (*) into NetDB.existence_counter from IPaddress
         where ipaddress_txt = IP.host (p_prefix) and ipaddressspace is null and reserved is null;

        if (NetDB.existence_counter > 0) then
            raise_application_error (NetDB.ex_record_already_exists,
                'The base address "' || IP.host (p_prefix) ||
                '" of the address space "' || IP.text (p_prefix) || '" is in use.');
        end if;

        select count (*) into NetDB.existence_counter from IPaddress
         where ipaddress_txt = IP.broadcast (p_prefix) and ipaddressspace is null and reserved is null;

        if (NetDB.existence_counter > 0) then
            raise_application_error (NetDB.ex_record_already_exists,
                'The broadcast address "' || IP.broadcast (p_prefix) ||
                '" of the address space "' || IP.text (p_prefix) || '" is in use.');
        end if;
    end validate_split;


    -- Get the parent network ID for an address space.
    function get_parent_id (p_ias_id in int) return int
    as
        v_net_id int := 0;
    begin
        select network into v_net_id from IPaddressSpace where id = verify_or_die (p_ias_id);

        return v_net_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('AddressSpace', p_ias_id);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('AddressSpace');
    end get_parent_id;


    procedure locked_or_die ( p_op_uid in int,
                              p_lock   in timestamp,
                              p_ias_id in int ) as
    begin
        NetDB_Network.locked_or_die (p_op_uid, p_lock, get_parent_id (p_ias_id));
    end locked_or_die;


    function info (p_ias_id in int) return result_set pipelined
    as
        space_row  IPaddressSpace%ROWTYPE;
        prefix_len int        := 0;
        res_row    result_row := result_row (null, null, null, null, null, null, null);
        ip_cursor  NetDB.ref_cursor;
    begin
        select * into space_row from IPaddressSpace where id = p_ias_id;

        prefix_len := IP.masklen (space_row.prefix);

        pipe row (result_row ('key', 'addr_space_ip',     null, 'string',  space_row.address_txt, null, null));
        pipe row (result_row ('key', 'prefix_length',     null, 'integer', null,                  null, prefix_len));
        pipe row (result_row ('key', 'addr_space_prefix', null, 'string',  space_row.prefix_txt,  null, null));
        pipe row (result_row ('key', 'low_res',           null, 'integer', null, null,
                IP.difference(space_row.lowres, space_row.address)));
        pipe row (result_row ('key', 'high_res',          null, 'integer', null, null,
                IP.difference(space_row.broadcast, space_row.highres)));
        pipe row (result_row ('key', 'comment',           null, 'string',  space_row.comments,    null, null));

        -- Get the groups.
        pipe row (result_row ('data', 'groups', 1, null, null, null, null));

        for r in ( select * from IPaddressSpace_Owner
                     join Owner on IPaddressSpace_Owner.owner = Owner.id
                    where IPaddressSpace_Owner.IPaddressSpace = space_row.id
                    order by lower (Owner.label) )
        loop
            pipe row (result_row ('key', r.label, null, 'string', ' ', null, null));
        end loop;

        -- Get the dhcp addresses.
        pipe row (result_row ('data', 'dhcp_addresses', 1, null, null, null, null));

        ip_cursor := NetDB_IPaddress.master ('info', null, null, null, null, null, null, null, null, space_row.id);
        loop
            fetch ip_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                 res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when ip_cursor%notfound;

            res_row.hash_level := res_row.hash_level + 1;
            pipe row (res_row);
        end loop;
        close ip_cursor;

        -- Get the dhcp options.
        pipe row (result_row ('data', 'dhcp_options', 1, null, null, null, null));

        for r in ( select *
                     from IPaddressSpace_DHCPoptions i
                     join DHCPoption d on d.id = i.dhcpoption
                    where i.ipaddressspace = space_row.id )
        loop
            pipe row (result_row ('data', r.label || '=' || r.value,    2,  null,    null,    null, null));
            pipe row (result_row ('key', 'option',                   null, 'string', r.label, null, null));
            pipe row (result_row ('key', 'value',                    null, 'string', r.value, null, null));
        end loop;
    end info;


    -- Returns list of address spaces to which user p_uid has access or all addr. spaces if p_uid is null.
    function list (p_user_id in int := null) return result_set pipelined as
    begin
        if (p_user_id is not null and not NetDB_User.is_active (p_user_id)) then return; end if;

        if (p_user_id is null or NetDB_User.has_all_groups (p_user_id)) then
            for r in ( select * from IPaddressSpace )
            loop
                pipe row (result_row ('data', r.address_txt,    0,    null,     null,              null, null));
                pipe row (result_row ('key', 'net_id',         null, 'string', to_char(r.network), null, null));
                pipe row (result_row ('key', 'str_ip_address', null, 'string', r.address_txt,      null, null));
                pipe row (result_row ('key', 'prefix',         null, 'string', r.prefix_txt,       null, null));
                pipe row (result_row ('key', 'ip_addr_space',  null, 'string', r.address_txt,      null, null));
                pipe row (result_row ('key', 'comment',        null, 'string', r.comments,         null, null));
            end loop;

        else
            for r in ( select IPaddressSpace.*
                         from IPaddressSpace
                         join IPaddressSpace_Owner on IPaddressSpace.id = IPaddressSpace_Owner.IPaddressSpace
                         join Users_Owner on IPaddressSpace_Owner.owner = Users_Owner.owner
                        where Users_Owner.userid = p_user_id )
            loop
                pipe row (result_row ('data', r.address_txt,    0,    null,     null,              null, null));
                pipe row (result_row ('key', 'net_id',         null, 'string', to_char(r.network), null, null));
                pipe row (result_row ('key', 'str_ip_address', null, 'string', r.address_txt,      null, null));
                pipe row (result_row ('key', 'prefix',         null, 'string', r.prefix_txt,       null, null));
                pipe row (result_row ('key', 'ip_addr_space',  null, 'string', r.address_txt,      null, null));
                pipe row (result_row ('key', 'comment',        null, 'string', r.comments,         null, null));
            end loop;
        end if;
        return;
    end list;


    -- Delete an IP address space.
    procedure del ( p_op_uid in int,
                    p_lock   in timestamp,
                    p_ias_id in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_ias_id);

        delete from IPaddressSpace where id = p_ias_id;
    end del;


    function ins ( p_op_uid     in int,
                   p_lock       in timestamp,
                   p_prefix     in varchar,
                   p_network_id in int,
                   p_comment    in varchar,
                   p_lowres     in int,
                   p_highres    in int ) return NetDB.ref_cursor
    as
        v_ias_id int;

        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        if (p_lowres < 0 or p_highres < 0) then
            raise_application_error (NetDB.ex_bad_parameter,
                'The sizes of both the low reserve and high reserve areas, must be positive.');
        end if;

        NetDB_Network.locked_or_die (p_op_uid, p_lock, p_network_id);
        NetDB_Util.validate (p_comment, 'comment');

        insert into IPaddressSpace ( network,      comments,  prefix,
                                     lowres,
                                     highres )
                            values ( p_network_id, p_comment, p_prefix,
                                     IP.plus (IP.host (p_prefix), p_lowres),
                                     IP.plus (IP.broadcast (p_prefix), -p_highres) )
        returning id into v_ias_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'id', 0, 'string', v_ias_id, null, null);

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    function upd ( p_op_uid     in int,
                   p_lock       in timestamp,
                   p_ias_id     in int,
                   p_prefix     in varchar,
                   p_network_id in int,
                   p_net_lock   in timestamp,
                   p_comment    in varchar,
                   p_lowres     in int,
                   p_highres    in int ) return NetDB.ref_cursor
    as
        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        if (p_lowres < 0 or p_highres < 0) then
            raise_application_error (NetDB.ex_bad_parameter,
                'The sizes of both the low reserve and high reserve areas, must be positive.');
        end if;

        locked_or_die (p_op_uid, p_lock, p_ias_id);

        NetDB_Network.locked_or_die (p_op_uid, p_net_lock, p_network_id);
        NetDB_Util.validate (p_comment, 'comment');

        update IPaddressSpace set network  = p_network_id,
                                  comments = p_comment,
                                  prefix   = p_prefix,
                                  lowres   = IP.plus (IP.host (p_prefix), p_lowres),
                                  highres  = IP.plus (IP.broadcast (p_prefix), -p_highres)
         where id = p_ias_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'id', 0, 'string', p_ias_id, null, null);

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- Verify IPaddressSpace ID / handle or complain.
    function verify_or_die ( p_ias_id in int     := null,
                             p_prefix in varchar := null ) return int
    as
        v_identifier varchar(256) := p_prefix;
        v_as_id      int;
    begin
        if (p_ias_id is null and p_prefix is null) then
            NetDB_Util.throw_no_record_id_handle ();

        elsif (p_ias_id is not null and p_prefix is not null) then
            select id into v_as_id from IPaddressSpace
             where id = p_ias_id and prefix = IP.normalize (p_prefix);

        elsif (p_prefix is not null) then
            select id into v_as_id from IPaddressSpace
             where prefix = IP.normalize (p_prefix);

        else -- p_ias_id is not null
            v_identifier := p_ias_id;

            select id into v_as_id from IPaddressSpace where id = p_ias_id;
        end if;

        return v_as_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('AddressSpace', v_identifier);
    end verify_or_die;


    function handle ( p_ias_id      in int,
                      user_readable in boolean := false) return varchar
    as
        v_prefix IPaddressSpace.prefix%TYPE;
    begin
        select prefix into v_prefix from IPaddressSpace where id = p_ias_id;

        if (user_readable) then
            v_prefix := IP.text (v_prefix);
        end if;

        return v_prefix;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('AddressSpace', p_ias_id);
    end handle;


    function master ( p_action     in varchar,            -- { delete, insert, update, info, verify, list }
                      p_op_uid     in int       := null,
                      p_lock       in timestamp := null,  -- Lock for the network the address space is currently in.
                      p_ias_id     in int       := null,
                      p_prefix     in varchar   := null,
                      p_network_id in int       := null,
                      p_net_lock   in timestamp := null,  -- The network lock, when updating. This will be
                                                          -- the same as p_lock, unless the address space
                                                          -- is being moved between networks.
                      p_comment    in varchar   := null,
                      p_lowres     in int       := null,
                      p_highres    in int       := null ) return NetDB.ref_cursor
    as
        v_action varchar(16) := lower (p_action);
        v_ias_id int;

        ret_cursor NetDB.ref_cursor := null;
    begin
        if (v_action in ('delete', 'info', 'update', 'verify')) then
            v_ias_id := verify_or_die (p_ias_id, p_prefix);
        end if;

        -- Perform the requested action.
        case v_action
            when 'delete' then
                del (p_op_uid, p_lock, v_ias_id);

            when 'list' then
                open ret_cursor for
                    select * from table (cast (list (p_op_uid) as result_set));

            when 'info' then
                open ret_cursor for
                    select * from table (cast (info (v_ias_id) as result_set));

            when 'insert' then
                ret_cursor := ins ( p_op_uid,  p_lock,   p_prefix, p_network_id,
                                    p_comment, p_lowres, p_highres );

            when 'update' then
                ret_cursor := upd ( p_op_uid,   p_lock,    v_ias_id, p_prefix, p_network_id,
                                    p_net_lock, p_comment, p_lowres, p_highres );

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;

        if (ret_cursor is null) then
            ret_cursor := NetDB_Util.get_null_generic_cursor ();
        end if;

        return ret_cursor;
    end master;

end NetDB_IPaddressSpace;
/
