-- NetDB_Interface.sql --- contains functions related to interfaces

create or replace package body NetDB_Interface as

    -- Get the parent node ID for an interface.
    function get_parent_id (p_intfc_id in int) return int
    as
        v_node_id int := 0;
    begin
        select node into v_node_id from Interface where id = verify_or_die (p_intfc_id);

        return v_node_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('interface', p_intfc_id);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('interface');
    end get_parent_id;


    -- Verify that the node containing a interface is locked.
    procedure locked_or_die ( p_op_uid   in int,
                              p_lock     in timestamp,
                              p_intfc_id in int ) as
    begin
        NetDB_Node.locked_or_die (p_op_uid, p_lock, get_parent_id (p_intfc_id));
    end locked_or_die;


    function info (p_intfc_id in int) return NetDB.ref_cursor
    as
        v_key varchar(256);
        intfc Interface%ROWTYPE;

        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        select * into intfc from Interface where id = p_intfc_id;

        out_record.extend (6);
        out_record(1) := result_row ('data', intfc.id,      0,  null,     null,             null, null);
        out_record(2) := result_row ('key', 'roaming?',  null, 'boolean', null,             null, intfc.roam);
        out_record(3) := result_row ('key', 'use-dhcp?', null, 'boolean', null,             null, intfc.dhcp);
        out_record(4) := result_row ('key', 'comment',   null, 'string',  intfc.comments,   null, null);
        out_record(5) := result_row ('key', 'hw',        null, 'string',  intfc.hw_address, null, null);

        -- dhcp options
        out_record(6) := result_row ('data', 'dhcp-options', 1, null, null, null, null);
        for opt in ( select do.label, ido.value
                       from Interface_DHCPoptions ido
                       join DHCPoption do on do.id =  ido.DHCPoption
                      where ido.interface = intfc.id )
        loop
            v_key := opt.label || '=' || opt.value;

            out_record.extend (3);
            out_record(out_record.COUNT - 2) := result_row ('data', v_key,      2,  null,    null,      null, null);
            out_record(out_record.COUNT - 1) := result_row ('key', 'option', null, 'string', opt.label, null, null);
            out_record(out_record.COUNT)     := result_row ('key', 'value',  null, 'string', opt.value, null, null);
        end loop;

        -- interface names
        for name in ( select *
                        from table (cast (NetDB_Name.names (null, intfc.id, null, null, 1) as result_set)) )
        loop
            out_record.extend (1);
            out_record(out_record.COUNT) := result_row ( name.type, name.key, name.hash_level, name.value_type,
                                                         name.value_string, name.value_timestamp, name.value_int );
        end loop;

        -- ip addresses
        out_record.extend (1);
        out_record(out_record.COUNT) := result_row ('data', 'ips', 1, null, null, null, null);

        for ip in (select * from table (cast (NetDB_IPaddress.info (null, null, intfc.id, null) as result_set)))
        loop
            out_record.extend (1);
            out_record(out_record.COUNT) := result_row ( ip.type, ip.key, ip.hash_level + 1, ip.value_type,
                                                         ip.value_string, ip.value_timestamp, ip.value_int );
        end loop;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end info;


    procedure del ( p_op_uid   in int,
                    p_lock     in timestamp,
                    p_intfc_id in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_intfc_id);

        delete from Interface where id = p_intfc_id;
    end del;


    function ins ( p_op_uid     in int,
                   p_lock       in timestamp,
                   p_type       in varchar,
                   p_node_id    in int,
                   p_roam       in int,
                   p_dhcp       in int,
                   p_HW_address in varchar,
                   p_comment    in varchar ) return NetDB.ref_cursor
    as
        v_intfc_id int := null;
        v_type     InterfaceType.type%TYPE := NetDB_InterfaceType.verify_or_die (p_type);

        v_dhcp int := 1;  -- We always encourage people to use DHCP.
        v_roam int := 0;  -- Only computers weighing less than six pounds, should roam.

        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        NetDB_Node.locked_or_die (p_op_uid, p_lock, p_node_id);
        NetDB_Util.validate (p_comment, 'comment');

        if (p_dhcp is not null) then v_dhcp := p_dhcp; end if;
        if (p_roam is not null) then v_roam := p_roam; end if;

        insert into Interface (type,   node,      roam,   dhcp,   hw_address,   comments)
                       values (p_type, p_node_id, v_roam, v_dhcp, p_HW_address, p_comment)
        returning id into v_intfc_id;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'id', 0, 'string', v_intfc_id, null, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('Interface', p_HW_address);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    procedure upd ( p_op_uid     in int,
                    p_lock       in timestamp,
                    p_intfc_id   in int,
                    p_type       in varchar,
                    p_roam       in int,
                    p_dhcp       in int,
                    p_HW_address in varchar,
                    p_comment    in varchar )
    as
        v_intfc_id int := verify_or_die (p_intfc_id);
        v_type     InterfaceType.type%TYPE := NetDB_InterfaceType.verify_or_die (p_type);
    begin
        NetDB_Node.locked_or_die (p_op_uid, p_lock, get_parent_id (p_intfc_id));
        NetDB_Util.validate (p_comment, 'comment');

        update Interface set type       = p_type,
                             roam       = nvl (p_roam, 0), -- See ins (...)
                             dhcp       = nvl (p_dhcp, 1),
                             hw_address = p_HW_address,
                             comments   = p_comment
         where id = v_intfc_id;
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('Interface', p_HW_address);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- verify that an interface, identified by the id or hardware address, exists.
    function verify_or_die ( p_intfc_id   in int     := null,
                             p_HW_address in varchar := null ) return int
    as
        v_identifier varchar(256) := p_HW_address;
        v_record_id  int          := 0;
    begin
        if (p_intfc_id is null and p_HW_address is null) then
            NetDB_Util.throw_no_record_id_handle ();

        elsif (p_intfc_id is not null and p_HW_address is not null) then
            select id into v_record_id from Interface
             where lower (hw_address) = lower (p_HW_address) and id = p_intfc_id;

        elsif (p_HW_address is not null) then
            select id into v_record_id from Interface where lower (hw_address) = lower (p_HW_address);

        else -- p_intfc_id is not null
            v_identifier := p_intfc_id;

            select id into v_record_id from Interface where id = p_intfc_id;
        end if;

        return v_record_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('interface', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('interface');
    end verify_or_die;


    function handle (p_intfc_id in int) return varchar
    as
        hw_addr Interface.hw_address%TYPE;
    begin
        select hw_address into hw_addr from Interface where id = p_intfc_id;

        return hw_addr;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('interface', p_intfc_id);
    end handle;


    function master ( p_action      in varchar,  -- {delete, info, insert, update}
                      p_op_uid      in int       := null,
                      p_lock        in timestamp := null,
                      p_intfc_id    in int       := null,
                      p_hw_addr     in varchar   := null,
                      p_hw_addr_new in varchar   := null,
                      p_type        in varchar   := null,
                      p_node_id     in int       := null,
                      p_use_roam    in int       := null,
                      p_use_dhcp    in int       := null,
                      p_comment     in varchar   := null ) return NetDB.ref_cursor
    as
        v_action   varchar(16) := lower (p_action);
        v_intfc_id int;

        ret_cursor NetDB.ref_cursor;
    begin
        if (v_action in ('delete', 'info', 'update')) then
            v_intfc_id := verify_or_die (p_intfc_id, p_hw_addr);
        end if;

        case v_action
            when 'delete' then
                del (p_op_uid, p_lock, v_intfc_id);

            when 'info' then
                ret_cursor := info (v_intfc_id);

            when 'insert' then
                ret_cursor := ins (p_op_uid, p_lock, p_type, p_node_id, p_use_roam, p_use_dhcp, p_hw_addr, p_comment);

            when 'update' then
                upd (p_op_uid, p_lock, v_intfc_id, p_type, p_use_roam, p_use_dhcp, p_hw_addr_new, p_comment);

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;

        if (ret_cursor is null) then
            ret_cursor := NetDB_Util.get_null_generic_cursor ();
        end if;

        return ret_cursor;
    end master;

end NetDB_Interface;
/
