-- NetDB_Make.sql --- contains functions related to Make table

create or replace package body NetDB_Make as

    function list return NetDB.ref_cursor
    as
        ret_cursor  NetDB.ref_cursor := null;
        out_record  result_set       := result_set ();
    begin
        for r in (select * from Make)
        loop
            out_record.extend(1);
            out_record(out_record.COUNT) :=
                result_row ('key', r.make, null, 'string', ' ', null, null);
        end loop;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    -- deletes make
    procedure del (p_make in varchar) as
    begin
        delete Make where make = p_make;
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('make', p_make);
    end del;


    -- inserts make
    procedure ins (p_make in varchar) as
    begin
        NetDB_Util.validate (p_make, 'make name', 'A make must have a name');

        insert into Make (make) values (p_make);
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('make', p_make);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- updates makes
    procedure upd ( p_make     in varchar,
                    p_make_new in varchar ) as
    begin
        NetDB_Util.validate (p_make_new, 'make name', 'A make must have a name');

        update Make set make = p_make_new where make = p_make;
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('make', p_make_new);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- given make verifies that it exists and returns itself
    function verify_or_die (p_make in varchar) return varchar
    as
        v_make Make.make%TYPE;
    begin
        select make into v_make from Make where lower (make) = lower (p_make);

        return v_make;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('make', p_make);
    end verify_or_die;


    -- main (interfacing) function to access and manipulate makes
    function master ( p_action   in varchar,
                      p_op_uid   in int     := null,
                      p_make     in varchar := null,
                      p_make_new in varchar := null ) return NetDB.ref_cursor
    as
        v_make Make.make%TYPE := null;

        ret_cursor NetDB.ref_cursor := null;
    begin
        if (p_action in ('delete', 'insert', 'update')) then
            NetDB_Access.has_access_or_die (p_op_uid, 'Make');
        end if;

        if (p_action in ('delete', 'update')) then
            v_make := verify_or_die (p_make);
        end if;

        case p_action
            when 'list' then
                ret_cursor := list ();

            when 'delete' then
                del (p_make);

            when 'insert' then
                ins (translate (p_make, '"', ''''));

            when 'update' then
                upd (p_make, translate (p_make_new, '"', ''''));

            else
                NetDB_Util.throw_unknown_action (p_action);

        end case;

        if (ret_cursor is null) then
            ret_cursor := NetDB_Util.get_null_generic_cursor ();
        end if;

        return ret_cursor;
    end master;

end NetDB_Make;
/
