-- NetDB_Name.sql --- functions dealing with names (related to the Name table)

create or replace package body NetDB_Name as

    sys_time_stamp timestamp;

    function get_dns_label (p_type in varchar) return varchar
    as
        v_dns_label NameType.dnslabel%TYPE := null;
    begin
        select distinct dnslabel into v_dns_label from NameType
         where lower (p_type) in (lower (label), lower (dnslabel));

        return v_dns_label;
    exception
        when NO_DATA_FOUND then
             raise_application_error (NetDB.ex_bad_parameter, 'Unknown type "' || p_type || '".');
        when TOO_MANY_ROWS then
            raise_application_error (NetDB.ex_idiot_programmer, 'NameType table is fubar.');
    end get_dns_label;


    -- Attempt to reserve a name.
    function reserve ( p_op_uid in int,
                       p_name   in varchar,
                       p_domain in varchar,
                       p_type   in varchar ) return NetDB.ref_cursor
    as
        v_domain_id     int;
        lower_name      Name.name%TYPE     := lower (p_name);
        v_fullname      Name.fullname%TYPE := NetDB_Util.fullname (p_name, p_domain);
        v_extant_domain int                := 0;
        v_dns_label     Name.DNSlabel%TYPE := get_dns_label (p_type);

        -- If the initial value of v_id is changed from null,
        --   be sure to change the test after the update.
        v_id       int       := null;
        v_reserved timestamp := null;
        v_name_id  int       := null;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin

        sys_time_stamp := systimestamp;

        -- make sure the name is composed of valid characters and is not too long
        if (not DNS.check_name (lower_name)) then
            raise_application_error (NetDB.ex_bad_name,
                'The name "' || p_name || '" is too long or contains invalid characters.');
        end if;

        -- make sure the domain exists
        v_domain_id := NetDB_Domain.verify_or_die (null, p_domain);

        -- disallow adding top-level names/domains, i.e., names with parent domain '.'
        if (v_domain_id = 1) then
            raise_application_error (NetDB.ex_cannot_use_domain,
                'Names cannot be assigned the parent domain "."');
        end if;

        -- see if there's a domain of the same name
        begin
            v_extant_domain := NetDB_Domain.verify_or_die (null, v_fullname);
        exception
            when NetDB.record_not_found then
                -- the new name doesn't match an existing domain name, so make
                --   sure the user has rights to assign names in the name's domain
                if (not NetDB_Domain.can_use (p_op_uid, v_domain_id)) then
                    raise_application_error (NetDB.ex_cannot_use_domain,
                        'The user "' || NetDB_User.handle (p_op_uid) ||
                        '" cannot assign names in the domain "' ||
                        NetDB_Domain.handle (v_domain_id) || '".');
                end if;
        end;

        -- make sure the name is not globally reserved (restricted)
        if (NetDB_ResName.is_reserved (p_op_uid, lower_name, v_domain_id)) then
            raise_application_error (NetDB.ex_bad_name,
                'The name "' || p_name || '" is reserved/restricted.');
        end if;

        -- if the domain is limited, make sure the name is allowed
        if (Util.to_boolean (NetDB_Default.get_int ('CheckForLimitedDomain')) and
            NetDB_Domain.is_limited (v_domain_id) and
            not NetDB_LimitedName.is_limited (lower_name))
        then
            raise_application_error (NetDB.ex_bad_name,
                'The name "' || p_name || '" is not valid for the limited domain "' ||
                NetDB_Domain.handle (v_domain_id) || '".');
        end if;

        -- nuke expired name reservations so they don't get in the way
        delete from Name
         where reserved is not null and -- very important/necessary for speed!
               Util.datediff ('second', reserved) > NetDB_Default.get_int ('LockTimeOut');

        -- if the new name matches an existing domain name, see if
        --   that domain name can be used as other name types
        if (v_extant_domain > 0 and
            not NetDB_Domain.can_use_as_name (p_op_uid, v_extant_domain)) then
            raise_application_error (NetDB.ex_cannot_use_domain_as_name,
                'The user "' || NetDB_User.handle (p_op_uid) ||
                '" cannot use the domain "' ||
                NetDB_Domain.handle (v_extant_domain) || '" as a name.');
        end if;

        -- if the user already has this name reserved, refresh the reservation and return it
        begin
               update Name set name     = p_name,
                               domain   = v_domain_id,
                               label    = p_type,
                               dnslabel = v_dns_label,
                               reserved = sys_time_stamp
                where lower (name) = lower_name  and
                      domain       = v_domain_id and
                      reserved_by  = p_op_uid
            returning id, reserved into v_id, v_reserved;
        exception
            when NO_DATA_FOUND then
                null;
        end;

        -- if (the update failed)
        if (v_id is null) then
            -- make sure the name doesn't already exist, unless it's an MX
            --(p_name,v_domain_id,dns_label) exists and dns_label != "mx") error
            select count (*) into NetDB.existence_counter from Name
             where lower (name) = lower_name  and
                   domain       = v_domain_id and
                   dnslabel     = v_dns_label;

            if (NetDB.existence_counter > 0 and v_dns_label != 'MX') then
                raise_application_error (NetDB.ex_name_collision,
                    'The name "' || p_name || '" already exists with the domain " ' ||
                    NetDB_Domain.handle (v_domain_id) || '" and the type "' || p_type || '"');
            end if;

            -- alias names cannot coexist with other types [DNS rule]
            select count (*) into NetDB.existence_counter from Name
             where lower (name) = lower_name and domain = v_domain_id and dnslabel = 'CNAME';

            if (NetDB.existence_counter > 0) then
                raise_application_error (NetDB.ex_name_collision,
                    'The name "' || p_name ||
                    '" cannot be reserved because it already exists as an alias.');
            end if;

            select count (*) into NetDB.existence_counter from Name
             where lower (name) = lower_name and domain = v_domain_id;

            if (NetDB.existence_counter > 0 and v_dns_label = 'CNAME') then
                raise_application_error (NetDB.ex_name_collision,
                    'The name "' || p_name ||
                    '" cannot be created as an alias because the name already exists.');
            end if;

            -- if the new name is an "A" name and the name already exists as an MX
            --     if the user has access to any record associated with the MX received_by names
            --         ok
            --     else
            --         error
            if (v_dns_label = 'A') then
                begin
                    select id into v_name_id from Name
                     where lower (name) = lower_name and domain = v_domain_id and dnslabel = 'MX';

                    NetDB_Owner.has_access_or_die (p_op_uid, v_name_id, 'MX');
                exception
                    when NO_DATA_FOUND then null;
                end;
            end if;

            -- if the new name is an MX
            --    if the name already exists as an "A" name
            --        if the user has access to the existing record associated with the "A" name
            --            ok
            --        else
            --            error
            --    else if the name already exists as an MX
            --        if the user has access to any record associated with the MX received_by names
            --            ok
            --        else
            --            error
            if (v_dns_label = 'MX') then
                begin
                    select id into v_name_id from Name
                     where lower (name) = lower_name  and
                           domain       = v_domain_id and
                           dnslabel in ('A', 'MX')    and rownum = 1;

                    NetDB_Owner.has_access_or_die (p_op_uid, v_name_id, 'MX');
                exception
                    when NO_DATA_FOUND then null;
                end;
            end if;

            -- reserve the name by inserting it with the reservation (user,timestamp)
            -- unless the DNS type is "mx" and that name already exists
            select count (*) into NetDB.existence_counter from Name
             where lower (name) = lower_name and domain = v_domain_id and dnslabel = 'MX';

            if (v_dns_label != 'MX' or NetDB.existence_counter <= 0) then
                insert into Name (name,   domain,      label,  dnslabel,    reserved_by, reserved)
                          values (p_name, v_domain_id, p_type, v_dns_label, p_op_uid,    sys_time_stamp)
                returning id, reserved into v_id, v_reserved;
            else
                select id, reserved into v_id, v_reserved from Name
                 where name = p_name and domain = v_domain_id and dnslabel = 'MX';
            end if;
        end if;

        out_record.extend(2);
        out_record(1) := result_row ('key', 'id',       0, 'string',    v_id, null,       null);
        out_record(2) := result_row ('key', 'reserved', 0, 'timestamp', null, v_reserved, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end reserve;


    procedure accept ( p_op_uid        in int,
                       p_name_id       in int       := null,
                       p_fullname      in varchar   := null,
                       p_name_type     in varchar,
                       p_referent_id   in int       := null,
                       p_referent_name in varchar   := null,
                       p_lock          in timestamp,
                       p_mx_pref       in int       := null )
    as
        v_reserved      timestamp;
        v_name_id       int := verify_or_die (p_name_id, p_fullname, p_name_type, true);
        v_referent_id   int;

        v_alias_parent_id   int := 0;
        v_alias_parent_type Name.label%TYPE;

        v_type varchar(256) := lower (p_name_type);
    begin
        begin
            v_referent_id := NetDB_Util.verify_or_die ( p_referent_id, p_referent_name,
                                                        p_name_type,   null, true );
        exception
            when NetDB.single_record_expected then
                -- When attempting to create an MX that shares a name with a node and a domain,
                -- we're going to favor nodes for the referrent.
                if (v_type = 'mx') then
                    v_referent_id := NetDB_Util.verify_or_die ( p_referent_id, p_referent_name,
                                                                p_name_type,  'node', true );
                end if;
        end;

        -- Verify that the user has reserved the name, or that the name exists as an MX.
        select reserved into v_reserved from Name
         where  id          = v_name_id and
               (reserved_by = p_op_uid or dnslabel = 'MX');

        if ( v_type != 'mx' and
             Util.datediff ('second', v_reserved) > NetDB_Default.get_int ('LockTimeOut'))
        then
            raise_application_error (NetDB.ex_record_not_reserved,
                'The name "' || fullname (v_name_id) || '" has expired.');
        end if;

        -- validate the referent_id and verify it is locked.
        if (v_type = 'alias') then
            get_alias_terminal_parent (v_referent_id, v_alias_parent_id, v_alias_parent_type);
            NetDB_Util.locked_or_die (p_op_uid, p_lock, v_alias_parent_id, v_alias_parent_type);

        elsif (v_type in ('address', 'domain', 'interface', 'network', 'node')) then
            NetDB_Util.locked_or_die (p_op_uid, p_lock, v_referent_id, v_type);

        elsif (v_type = 'mx') then
            -- All validation of the MX is done in the insert.
            NetDB_MXrecord.ins (p_op_uid, p_lock, v_name_id, null, v_referent_id, null, p_mx_pref);

        else
            NetDB_Util.throw_unknown_record_type (p_name_type);
        end if;

        update Name set aliasee     = case v_type when 'alias'     then v_referent_id else null end,
                        node        = case v_type when 'node'      then v_referent_id else null end,
                        network     = case v_type when 'network'   then v_referent_id else null end,
                        interface   = case v_type when 'interface' then v_referent_id else null end,
                        ipaddress   = case v_type when 'address'   then v_referent_id else null end,
                        reserved_by = null,
                        reserved    = null
         where id = v_name_id and reserved_by = p_op_uid;
    exception
        when NO_DATA_FOUND then
            raise_application_error (NetDB.ex_record_not_reserved,
                'The user "' || NetDB_User.handle (p_op_uid) || '" has not reserved the name "' ||
                fullname (v_name_id) || '".');
    end accept;


    procedure cancel ( p_op_uid   in int,
                       p_name_id  in int     := null,
                       p_fullname in varchar := null,
                       p_type     in varchar := null )
    as
        v_name_id int;
    begin
        if (p_name_id is null and (p_fullname is null or p_type is null)) then
            raise_application_error (NetDB.ex_idiot_programmer,
                'Both p_fullname and p_type must not be null if p_name_id is null');
        end if;

        v_name_id := verify_or_die (p_name_id, p_fullname, p_type, true);

        delete from Name where id          = v_name_id and
                               label       = p_type    and
                               reserved_by = p_op_uid;
    exception
        when others then
            null;
    end cancel;


    -- Follow a trail of aliases to the first non-alias.
    procedure get_alias_terminal_parent ( p_name_id      in int,
                                          p_parent_id   out int,
                                          p_parent_type out varchar )
    as
        v_name_id int := related_node (p_name_id);
    begin
        if (v_name_id is not null) then
            p_parent_id   := v_name_id;
            p_parent_type := 'Node';
        else
            v_name_id := related_network (p_name_id);

            if (v_name_id is not null) then
                p_parent_id   := v_name_id;
                p_parent_type := 'Network';
            else
                v_name_id := p_name_id;

                loop
                    select aliasee   into v_name_id                  from Name where id = v_name_id;
                    select id, label into p_parent_id, p_parent_type from Name where id = v_name_id;

                    exit when lower (p_parent_type) != 'alias';
                end loop;
            end if;
        end if;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('Name', v_name_id);
    end get_alias_terminal_parent;


    -- fullname: return fully-qualified DNS name for input name ID
    function generate_fullname (p_name_id in int) return varchar
    as
        fullname Name.fullname%TYPE;
        v_name   Name.name%TYPE;
        parent   int;
    begin
        -- be reflexive with nulls
        if (p_name_id is null) then return null; end if;

        -- get the first bit, then loop for the rest
        select name, domain into fullname, parent from Name where id = p_name_id;
        loop
            exit when parent = 1;  -- optimization!
            select name, domain into v_name, parent from Name where id=parent;
         -- exit when name = '.';  -- the right way
            fullname := fullname || '.' || v_name;
        end loop;

        return fullname;
    end generate_fullname;


    function fullname (p_name_id in int) return varchar
    as
        v_fullname Name.fullname%TYPE := null;
    begin
        select fullname into v_fullname from Name where id = p_name_id;

        return v_fullname;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('name', p_name_id);
    end fullname;


    -- related_node: return ID of node that contains name identified by input name ID
    function related_node (p_name_id in int) return int
    as
        related_node_id int;
    begin
        select coalesce ( Name.node,
                          Interface.node,
                          IPaddress.ipcnode,
                         (select node from Interface where IPaddress.interface = Interface.id) )
          into related_node_id
          from ( select * from Name where aliasee is null
                  start with id = p_name_id connect by prior aliasee = id
                     union
                 select * from Name where id = p_name_id and aliasee is null ) Name
          left join Interface on Name.interface = Interface.id
          left join IPaddress on Name.ipaddress = IPaddress.id;

        return related_node_id;
    end related_node;


    -- Find the network a name belongs to.
    function related_network (p_name_id in int) return int
    as
        v_network_id int;
    begin
        select coalesce (Name.network, IPaddressSpace.network) into v_network_id
          from ( select * from Name where aliasee is null
                  start with id = p_name_id connect by prior aliasee = id
                     union
                 select * from Name where id = p_name_id and aliasee is null ) Name
          left join IPaddress      on IPaddress.id      = Name.ipaddress
          left join IPaddressSpace on IPaddressSpace.id = IPaddress.ipaddressspace;

        return v_network_id;
    end related_network;


    -- internal function used by names(...) to pull out all recursively aliases
    -- it is used instead of connect by queries because they are too slow when
    -- going down the tree
    function aliases ( p_name_id    in int,
                       p_hash_level in int ) return result_set pipelined as
    begin
        pipe row (result_row ('data', 'aliases', p_hash_level, null, null, null, null));

        for r in ( select Name.id name_id, Name.name, Domain.domain, Name.fullname
                     from Name
                     join Domain on Name.domain = Domain.id
                    where Name.aliasee = p_name_id )
        loop
            pipe row (result_row ('data', r.fullname, p_hash_level + 1,  null,    null,     null, null));
            pipe row (result_row ('key', 'name',      null,             'string', r.name,   null, null));
            pipe row (result_row ('key', 'domain',    null,             'string', r.domain, null, null));

            -- aliases of this alias
            for d in (select * from table (cast (aliases (r.name_id, p_hash_level + 2) as result_set)))
            loop
                pipe row (result_row ( d.type, d.key, d.hash_level,
                                       d.value_type, d.value_string, d.value_timestamp, d.value_int ));
            end loop;
        end loop;
    end aliases;


    -- This function is to be used internally by NetDB_(Node|Network).info
    -- It returns set of names together with mxes and aliases associated with
    -- either ip address, interface, node or a network, starting with hash level
    -- p_hash_level.
    function names ( p_ip         in int := null,
                     p_interface  in int := null,
                     p_node       in int := null,
                     p_network    in int := null,
                     p_hash_level in int := 0 ) return result_set pipelined as
    begin
        pipe row (result_row ('data', 'names', p_hash_level, null, null, null, null));

        for r in ( select Name.id name_id, Name.name, Domain.domain, Name.fullname
                     from Name join Domain on Name.domain = Domain.id
                    where ipaddress = p_ip or
                          interface = p_interface or
                          node      = p_node or
                          network   = p_network )
        loop
            pipe row (result_row ('data',  r.fullname, p_hash_level + 1,  null,    null,     null, null));
            pipe row (result_row ('key',  'name',      null,             'string', r.name,   null, null));
            pipe row (result_row ('key',  'domain',    null,             'string', r.domain, null, null));

            -- mxes
            pipe row (result_row ('data', 'mxs', p_hash_level + 2, null, null, null, null));

            for d in ( select Name.name, Name.fullname, MXrecord.preference pref, Domain.domain
                         from Name
                         join MXrecord on MXrecord.sent_to = Name.id
                         join Domain   on Domain.id        = Name.domain
                        where MXrecord.received_by = r.name_id )
            loop
                pipe row (result_row ('data',  d.fullname, p_hash_level + 3,  null,    null,     null, null));
                pipe row (result_row ('key',  'name',      null,             'string', d.name,   null, null));
                pipe row (result_row ('key',  'domain',    null,             'string', d.domain, null, null));
                pipe row (result_row ('key',  'pref',      null,             'string', d.pref,   null, null));

            end loop;

            -- aliases
            for d in (select * from table (cast (aliases (r.name_id, p_hash_level + 2) as result_set)))
            loop
                pipe row (result_row ( d.type, d.key, d.hash_level,
                                       d.value_type, d.value_string, d.value_timestamp, d.value_int ));
            end loop;
        end loop;
    end names;


    -- Check to see if a user has a lock for a record.
    procedure locked_or_die ( p_op_uid  in int,
                              p_lock    in timestamp,
                              p_name_id in int )
    as
        name_row      Name%ROWTYPE;

        v_parent_id   int;
        v_parent_type Name.label%TYPE;
    begin
        select * into name_row from Name where id = p_name_id;

        case lower (name_row.label)
            when 'mx' then null;

            when 'alias' then
                get_alias_terminal_parent (p_name_id, v_parent_id, v_parent_type);
                NetDB_Util.locked_or_die (p_op_uid, p_lock, v_parent_id, v_parent_type);

            when 'address' then
                NetDB_IPaddress.locked_or_die (p_op_uid, p_lock, name_row.ipaddress);

            when 'domain' then
                NetDB_Domain.locked_or_die (p_op_uid, p_lock, p_name_id);

            when 'interface' then
                NetDB_Interface.locked_or_die (p_op_uid, p_lock, name_row.interface);

            when 'network' then
                NetDB_Network.locked_or_die (p_op_uid, p_lock, name_row.network);

            when 'node' then
                NetDB_Node.locked_or_die (p_op_uid, p_lock, name_row.node);

            else
                NetDB_Util.throw_unknown_record_type (name_row.label);
        end case;
    end locked_or_die;


    -- Delete a record.
    procedure del ( p_op_uid   in int,
                    p_lock     in timestamp,
                    p_name_id  in int     := null,
                    p_fullname in varchar := null,
                    p_type     in varchar )
    as
        v_name_id       int := verify_or_die (p_name_id, p_fullname, p_type, true);
        v_domain_id     int;
        v_extant_domain int := 0;
    begin
        select domain into v_domain_id from Name where id = v_name_id;

        -- does the name match an existing domain?
        begin
            v_extant_domain := NetDB_Domain.verify_or_die (null, fullname (v_name_id));
        exception
            when NetDB.record_not_found then null;
        end;

        -- if the name matches a domain and isn't actually the the domain,
        if (v_extant_domain > 0 and v_extant_domain != v_name_id) then

            -- allow the deletion if the user can use the domain as a name.
            if (not NetDB_Domain.can_use_as_name (p_op_uid, v_extant_domain)) then
                raise_application_error (NetDB.ex_cannot_use_domain,
                    'The user "' || NetDB_User.handle (p_op_uid) ||
                    '" cannot delete the name "' || fullname (v_name_id) || '".');
            end if;

        -- otherwise, allow the deletion if the user can use the parent domain.
        elsif (not NetDB_Domain.can_use (p_op_uid, v_domain_id)) then
            raise_application_error (NetDB.ex_cannot_use_domain,
                'The user "' || NetDB_User.handle (p_op_uid) ||
                '" cannot delete names from the domain "' ||
                NetDB_Domain.handle (v_domain_id) || '".');
        end if;

        NetDB_Name.locked_or_die (p_op_uid, p_lock, v_name_id);

        delete from Name where id = v_name_id or aliasee = v_name_id;
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('Name', fullname (v_name_id));
    end del;


    -- Update the record identfied by p_fullname or p_name.
    procedure update_domain_name ( p_op_uid   in int,
                                   p_lock     in timestamp,
                                   p_name_id  in int,
                                   p_name_new in varchar )
    as
        v_name_id int := verify_or_die (p_name_id);
        current   Name%ROWTYPE;
    begin
        NetDB_Domain.locked_or_die (p_op_uid, p_lock, v_name_id);

        update Name set name = coalesce (DNS.name (p_name_new), current.name)
         where id = v_name_id;
    end update_domain_name;


    -- Gets the name ID based on the name and label or verify that the name for an id exists.
    -- If check_reserved is false, the name MUST already exist in an accepted state.
    function verify_or_die ( p_name_id      in int     := null,
                             p_fullname     in varchar := null,
                             p_label        in varchar := '%',
                             check_reserved in boolean := false ) return int
    as
        v_identifier  varchar(256) := p_fullname;
        v_name_id     int          := 0;
        v_reserved_by int          := null;
        v_label       varchar(256) := coalesce (lower (p_label), '%');
        regexp_labels varchar(256);

        unaccepted_name EXCEPTION;
    begin
        for r in (select * from NameType where lower (dnslabel) = v_label)
        loop
            regexp_labels := regexp_labels || '|' || r.label;
        end loop;

        if (regexp_labels is not null) then
            v_label := '^(' || ltrim (lower (regexp_labels), '|') || ')$';
        end if;

        -- regexp-ify the search label.
        v_label := replace (replace (v_label, '%', '*'), '_', '.');

        if (p_name_id is null and p_fullname is null) then
            NetDB_Util.throw_no_record_id_handle ();

        elsif (p_name_id is not null and p_fullname is not null) then
            select id, reserved_by into v_name_id, v_reserved_by from Name
             where fullname_lc = lower (p_fullname) and
                   id          = p_name_id          and
                   regexp_like (lower (label), v_label);

        elsif (p_fullname is not null) then
            select id, reserved_by into v_name_id, v_reserved_by from Name
             where fullname_lc = lower (p_fullname) and regexp_like (lower (label), v_label);

        else
            v_identifier := p_name_id;

            select id, reserved_by into v_name_id, v_reserved_by from Name
             where id = p_name_id and regexp_like (lower (label), v_label);
        end if;

        -- Raise an error if the the name is still reserved, and we're not checking reserved names.
        if (v_reserved_by is not null and check_reserved != true) then
            raise unaccepted_name;
        end if;

        return v_name_id;
    exception
        when NO_DATA_FOUND or unaccepted_name then
            raise_application_error (NetDB.ex_record_not_found,
                'The name "' || v_identifier || '" does not exist with the label "' ||
                nvl (p_label, '%') || '".');
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('name');
    end verify_or_die;

end NetDB_Name;
/
