-- NetDB_Network.sql --- Network package body.

create or replace package body NetDB_Network as

    sys_time_stamp timestamp;

    -- Add or remove locations from a network.
    procedure location ( p_action      in varchar, -- { add, remove }
                         p_op_uid      in int,
                         p_lock        in timestamp,
                         p_network_id  in int,
                         p_location_id in int )
    as
        v_net_id int := verify_or_die (p_network_id);
        v_loc_id int := NetDB_Location.verify_or_die (p_location_id);
    begin
        locked_or_die (p_op_uid, p_lock, p_network_id);

        case lower (p_action)
            when 'add' then
                insert into Network_Location (network, location) values (v_net_id, v_loc_id);

            when 'remove' then
                delete from Network_Location where network = v_net_id and location = v_loc_id;

            else
                NetDB_Util.throw_unknown_action (p_action);

        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;
    end location;


    procedure locked_or_die ( p_op_uid     in int,
                              p_lock       in timestamp,
                              p_network_id in int )
    as
        v_lock timestamp := null;
    begin
        select locked into v_lock from Network
         where id = p_network_id and locked_by = p_op_uid and locked = p_lock and
               Util.datediff ('second', p_lock) < NetDB_Default.get_int ('LockTimeOut');
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_not_locked ('network', handle (p_network_id), NetDB_User.handle (p_op_uid));
    end locked_or_die;


    -- Lock a record.
    function lock_record ( p_op_uid in int,
                           p_net_id in int ) return NetDB.ref_cursor
    as
        v_lock      timestamp;
        v_locked_by int;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin
        select locked, locked_by into v_lock, v_locked_by from Network where id = p_net_id;

        -- if not locked, locked by this user, or the lock has expired, we're good to go
        if ( v_lock is null or v_locked_by = p_op_uid or
             Util.datediff ('second', v_lock) > NetDB_Default.get_int ('LockTimeOut') )
        then
               update Network set locked    = sys_time_stamp,
                                  locked_by = p_op_uid
                where id = p_net_id
            returning locked into v_lock;
        else
            NetDB_Util.throw_locked_by_other ( 'network',
                                                handle (p_net_id),
                                                NetDB_User.handle (v_locked_by) );
        end if;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'locked', 0, 'timestamp', null, v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end lock_record;


    -- Unlock a record.
    procedure unlock_record ( p_op_uid in int,
                              p_lock   in timestamp,
                              p_net_id in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_net_id);

        update Network set locked    = null,
                           locked_by = null
         where id = p_net_id and locked = p_lock and locked_by = p_op_uid;
    end unlock_record;


    -- lists names of the networks to which user has access
    function list (p_user_id in int) return NetDB.ref_cursor
    as
        out_record  result_set       := result_set ();
        ret_cursor  NetDB.ref_cursor := null;
    begin
        if (not NetDB_Access.check_access (p_user_id, 'Network')) then
            return null;
        end if;

        for r in ( select distinct fullname from (
                       select fullname from Name
                         join Network_Owner on Name.network        = Network_Owner.network
                         join Users_Owner   on Network_Owner.owner = Users_Owner.owner
                        where Users_Owner.userid = p_user_id
                           union
                       select fullname from Name
                         join Users on Name.id = Name.id
                        where Users.all_groups = 1 and Users.id = p_user_id and network is not null ) )
        loop
            out_record.extend(1);
            out_record(out_record.COUNT) := result_row ('key', r.fullname, 0, 'string', ' ', null, null);
        end loop;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    -- Get the info of a network.
    function info (p_net_id in int) return result_set pipelined
    as
        net_row     Network%ROWTYPE;
        res_row     result_row       := result_row (null, null, null, null, null, null, null);
        data_cursor NetDB.ref_cursor := null;
    begin
        select * into net_row from Network where id = p_net_id;

        pipe row (result_row ('key', 'object',  null, 'string', 'network',            null, null ));
        pipe row (result_row ('key', 'id',      null, 'string',  to_char(net_row.id), null, null ));
        pipe row (result_row ('key', 'comment', null, 'string',  net_row.comments,    null, null ));

        if (net_row.dhcpservice is not null) then
            select label into res_row.value_string from DHCPservice where id = net_row.dhcpservice;
            pipe row (result_row ('key', 'dhcp_service', null, 'string', res_row.value_string, null, null ));
        end if;

        pipe row (result_row ('key', 'created',  null, 'timestamp', null, net_row.created,  null ));
        pipe row (result_row ('key', 'modified', null, 'timestamp', null, net_row.modified, null ));

        if (net_row.locked is not null) then
            pipe row (result_row ('key', 'locked', null, 'timestamp', null, net_row.locked, null ));
        end if;

        -- Get the created_by info.
        pipe row (result_row ('data', 'created.by', 0, null, null, null, null));

        data_cursor := NetDB_Person.master ('info', null, net_row.created_by);
        loop
            fetch data_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                   res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when data_cursor%notfound;

            pipe row (res_row);
        end loop;
        close data_cursor;

        -- Get the modified_by info.
        pipe row (result_row ('data', 'mod.by', 0, null, null, null, null));

        data_cursor := NetDB_Person.master ('info', null, net_row.modified_by);
        loop
            fetch data_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                   res_row.value_string, res_row.value_timestamp, res_row.value_int;

            exit when data_cursor%notfound;

            pipe row (res_row);
        end loop;
        close data_cursor;

        -- Get the groups.
        pipe row (result_row ('data', 'groups', 0, null, null, null, null));

        for r in ( select *
                     from Network_Owner
                     join Owner on Network_Owner.owner = Owner.id
                    where Network_Owner.network = p_net_id
                    order by lower (Owner.label) )
        loop
            pipe row (result_row ('key', r.label, null, 'string', ' ', null, null));
        end loop;

        -- Get the DHCP options.
        pipe row (result_row ('data', 'dhcp_options', 0, null, null, null, null));

        for r in ( select *
                     from Network_DHCPoptions n
                     join DHCPoption d on d.id = n.dhcpoption
                    where n.network = p_net_id )
        loop
            pipe row (result_row ('data', r.label || '=' || r.value,    1,  null,    null,    null, null));
            pipe row (result_row ('key', 'option',                   null, 'string', r.label, null, null));
            pipe row (result_row ('key', 'value',                    null, 'string', r.value, null, null));
        end loop;

        -- Get the names.
        for r in ( select *
                     from table (cast (NetDB_Name.names (null, null, null, p_net_id) as result_set)) )
        loop
            pipe row (result_row ( r.type, r.key, r.hash_level,
                                   r.value_type, r.value_string, r.value_timestamp, r.value_int ));
        end loop;

        -- Get the address spaces.
        pipe row (result_row ('data', 'address_spaces', 0, null, null, null, null));

        for r in (select * from IPaddressSpace where network = p_net_id)
        loop
            pipe row (result_row ('data', r.prefix_txt, 1, null, null, null, null));

            for d in ( select *
                        from table (cast (NetDB_IPaddressSpace.info (r.id) as result_set)) )
            loop
                pipe row (result_row ( d.type, d.key, d.hash_level + 1,
                                       d.value_type, d.value_string, d.value_timestamp, d.value_int ));
            end loop;
        end loop;

        -- Get locations
        pipe row (result_row ('data', 'locations', 0, null, null, null, null));

        for r in ( select Location.name, Location.site, Location.code
                     from Network_Location
                     join Location on Network_Location.location = Location.id
                    where Network_Location.network = p_net_id )
        loop
            pipe row (result_row ('data', r.name,    1,     null, null,   null, null));
            pipe row (result_row ('key',  'name', null, 'string', r.name, null, null));
            pipe row (result_row ('key',  'site', null, 'string', r.site, null, null));
            pipe row (result_row ('key',  'code', null, 'string', r.code, null, null));
        end loop;
    end info;


    -- Delete a record and log the action.
    procedure del ( p_op_uid in int,
                    p_lock   in timestamp,
                    p_net_id in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_net_id);
        NetDB_Log.ins (p_op_uid, 'delete', p_net_id, 'network');

        delete from Network where id = p_net_id;

        NetDB_User.update_prefix ();
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('network', handle (p_net_id));
    end del;


    -- Insert a row into the Network table and set the Network column in the name table.
    function ins ( p_op_uid     in int,
                   p_service_id in int,
                   p_service    in varchar,
                   p_comment    in varchar ) return NetDB.ref_cursor
    as
        v_net_id     int       := 0;
        v_service_id int       := null;
        v_lock       timestamp := null;

        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        NetDB_Util.validate (p_comment, 'comment');

        if (p_service_id is not null or p_service is not null) then
            v_service_id := NetDB_DHCPservice.verify_or_die (p_service_id, p_service);
        end if;

        insert into Network ( DHCPservice,    comments,
                              created_by,     modified_by,     locked_by,
                              created,        modified,        locked )
                     values ( v_service_id,   p_comment,
                              p_op_uid,       p_op_uid,        p_op_uid,
                              sys_time_stamp, sys_time_stamp,  sys_time_stamp )
        returning id, locked into v_net_id, v_lock;

        NetDB_Log.set_action ('insert');

        out_record.extend(2);
        out_record(1) := result_row ('key', 'id',     0, 'string',    v_net_id, null,   null);
        out_record(2) := result_row ('key', 'locked', 0, 'timestamp', null,     v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Update Network info.
    procedure upd ( p_op_uid     in int,
                    p_lock       in timestamp,
                    p_net_id     in int,
                    p_service_id in int,
                    p_service    in varchar,
                    p_comment    in varchar )
    as
        v_service_id int := null;
    begin
        locked_or_die (p_op_uid, p_lock, p_net_id);

        NetDB_Util.validate (p_comment, 'comment');

        if (p_service_id is not null or p_service is not null) then
            v_service_id := NetDB_DHCPservice.verify_or_die (p_service_id, p_service);
        end if;

        update Network
           set modified_by = p_op_uid,
               modified    = sys_time_stamp,
               comments    = p_comment,
               DHCPservice = v_service_id
         where id = p_net_id;

        NetDB_Log.set_action ('update');
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- Verify that a network conforms to the rules and inserts the record into the log.
    procedure verify_and_log ( p_op_uid in int,
                               p_net_id in int )
    as
        in_address_space boolean          := false;
        ret_cursor       NetDB.ref_cursor := null;
    begin
        -- A network must have a name
        select count (*) into NetDB.existence_counter from Name where network = p_net_id;
        if (NetDB.existence_counter < 1) then
            raise_application_error (NetDB.ex_record_requires_name,
                'Networks must have at least one name.');
        end if;

        -- A network must be in at least one group.
        select count (*) into NetDB.existence_counter from Network_Owner where network = p_net_id;

        if (NetDB.existence_counter < 1) then
            raise_application_error (NetDB.ex_record_requires_group,
                'The network "' || handle (p_net_id) || '" must have at least one group.');
        end if;

        -- Make sure there are no orphaned IP addresses.
        for r in ( select * from ( select 'lower_limit' type, address   address from IPaddressSpace union
                                   select 'upper_limit' type, broadcast address from IPaddressSpace union
                                   select 'normal_ip'   type, IPaddress address from IPaddress
                                    where IPaddress.IPaddressSpace is null and reserved is null )
                    order by address, type )
        loop
            if (r.type = 'normal_ip' and not in_address_space) then
                raise_application_error (NetDB.ex_record_in_use, 'Nodes exist in removed address spaces.');

            elsif (r.type = 'lower_limit') then in_address_space := true;
            elsif (r.type = 'upper_limit') then in_address_space := false;
            end if;
        end loop;

        NetDB_Log.ins (p_op_uid, NetDB_Log.get_action (), p_net_id, 'network');
    end verify_and_log;


    function verify_or_die ( p_network_id in int     := null,
                             p_network    in varchar := null ) return int
    as
        v_identifier varchar(256) := p_network;
        v_network_id int          := 0;
    begin
        if (p_network_id is null and p_network is null) then
            NetDB_Util.throw_no_record_id_handle ();

        elsif (p_network_id is not null and p_network is not null) then
            select network into v_network_id from Name
             where fullname_lc = lower (p_network) and network = p_network_id;

        elsif (p_network is not null) then
            select network into v_network_id from Name
             where lower (label) = 'network' and fullname_lc = lower (p_network);

        else -- p_network_id is not null
            v_identifier := p_network_id;

            select id into v_network_id from Network where id = p_network_id;
        end if;

        return v_network_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('network', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('network');
    end verify_or_die;


    function handle (p_network_id in int) return varchar
    as
        v_network Name.name%TYPE := null;
    begin
        select fullname into v_network from Name where network = p_network_id;

        return v_network;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('network', p_network_id);
    end handle;


    function master ( p_action  in varchar,
                      p_op_uid  in int       := null,
                      p_lock    in timestamp := null,
                      p_net_id  in int       := null,
                      p_network in varchar   := null,
                      p_service in varchar   := null,
                      p_comment in varchar   := null ) return NetDB.ref_cursor
    as
        v_action varchar(16) := lower (p_action);
        v_net_id int;

        ret_cursor NetDB.ref_cursor := null;
    begin
        sys_time_stamp := systimestamp;

        if (v_action in ('delete', 'info', 'list', 'insert', 'lock', 'unlock', 'update', 'verify')) then
            if (p_action not in ('insert', 'list')) then
                v_net_id := verify_or_die (p_net_id, p_network);
            end if;

            if (p_action in ('insert', 'lock', 'unlock', 'update', 'delete')) then
                NetDB_Access.has_access_or_die (p_op_uid, 'Network');
            end if;

            if (p_action not in ('info', 'insert', 'list')) then
                NetDB_Owner.has_access_or_die (p_op_uid, v_net_id, 'Network');
            end if;
        end if;

        -- Perform the requested action.
        case v_action
            when 'delete' then
                del (p_op_uid, p_lock, v_net_id);

            when 'info' then
                open ret_cursor for
                    select * from table (cast (info (v_net_id) as result_set));

            when 'insert' then
                ret_cursor := ins (p_op_uid, null, p_service, p_comment);

            when 'list' then
                ret_cursor := list (p_op_uid);

            when 'lock' then
                ret_cursor := lock_record (p_op_uid, v_net_id);

            when 'unlock' then
                unlock_record (p_op_uid, p_lock, v_net_id);

            when 'update' then
                upd (p_op_uid, p_lock, v_net_id, null, p_service, p_comment);

            when 'verify' then
                verify_and_log (p_op_uid, p_net_id);

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;

        if (ret_cursor is null) then
            ret_cursor := NetDB_Util.get_null_generic_cursor ();
        end if;

        return ret_cursor;
    end master;

end NetDB_Network;
/
