-- NetDB_OS.sql --- contains functions related to OS table

create or replace package body NetDB_OS as

    sys_time_stamp timestamp;

    -- Replace "old OS" with "new OS" and update any nodes that had "old OS".
    procedure supplant ( p_op_uid in int,
                         p_os_old in varchar,
                         p_os_new in varchar ) as
    begin
        if (lower (p_os_new) = lower (p_os_old)) then
            return;
        end if;

        -- It's possible that both OSes are listed on the same node. If that's
        -- the case, a straight replace will violate the unique constraint.
        delete from Node_OS where lower (os) = lower (p_os_old) and
                                  node in (select node from Node_OS where os = p_os_new);

        -- At this point we can guarantee that a "replace all" will succeed.
        update Node_OS set os = p_os_new where lower (os) = lower (p_os_old);

        -- Now that no nodes are associated with the OS, we can safely remove it.
        delete from OS where lower (os) = lower (p_os_old);
    end supplant;


    function list (p_os in varchar) return NetDB.ref_cursor
    as
        v_os OS.os%TYPE := translate (p_os, '"', '''');

        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        for r in ( select os, coalesce (description, ' ') description from OS
                    where lower (os) like lower (coalesce (v_os, '%')) )
        loop
            out_record.extend(1);
            out_record(out_record.COUNT) := result_row ( 'key', r.os, null,
                                                         'string', r.description, null, null );
        end loop;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    -- deletes os
    procedure del ( p_op_uid in int,
                    p_os     in varchar ) as
    begin
        delete OS where lower (os) = lower (p_os);
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('os', p_os);
    end del;


    -- inserts os
    procedure ins ( p_op_uid      in int,
                    p_os          in varchar,
                    p_description in varchar ) as
    begin
        NetDB_Util.validate (p_os, 'OS name', 'OSes must have a name');
        NetDB_Util.validate (p_description, 'comment');

        insert into OS (os,   description,   modified_by, modified)
        values         (p_os, p_description, p_op_uid,    sys_time_stamp);
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('OS', p_os);

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- updates os
    procedure upd ( p_op_uid      in int,
                    p_os          in varchar,
                    p_os_new      in varchar,
                    p_description in varchar ) as
    begin
        NetDB_Util.validate (p_os_new, 'OS name', 'OSes must have a name');
        NetDB_Util.validate (p_description, 'comment');

        update OS set os          = p_os_new,
                      description = p_description,
                      modified_by = p_op_uid,
                      modified    = sys_time_stamp
         where lower (os) = lower (p_os);
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('OS', p_os_new);

        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- Verify OS handle or complain.
    function verify_or_die (p_os in varchar) return varchar
    as
        v_os    OS.os%TYPE := translate (p_os, '"', '''');
        v_found OS.os%TYPE;
    begin
        if (p_os is null) then
            NetDB_Util.throw_no_record_id_handle ();
        else
            select os into v_found from OS where lower (os) = lower (v_os);
        end if;

        return v_found;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('os', p_os);
    end verify_or_die;


    -- main (interfacing) function to access and manipulate oses
    function master ( p_action      in varchar,
                      p_op_uid      in int     := null,
                      p_os          in varchar := null,
                      p_os_new      in varchar := null,
                      p_description in varchar := null ) return NetDB.ref_cursor
    as
        v_action varchar(16) := lower (p_action);
        v_os     OS.os%TYPE  := translate (p_os,     '"', '''');
        v_os_new OS.os%TYPE  := translate (p_os_new, '"', '''');
        v_valid  OS.os%TYPE  := null;

        ret_cursor NetDB.ref_cursor := null;
    begin
        sys_time_stamp := systimestamp;

        if (v_action in ('delete', 'update', 'replace')) then
            NetDB_Access.has_access_or_die (p_op_uid, 'OS');
            v_valid := verify_or_die (v_os);

        elsif (v_action = 'insert') then
            -- Users with node access are allowed to insert only.
            if (not NetDB_Access.check_access (p_op_uid, 'OS')) then
                NetDB_Access.has_access_or_die (p_op_uid, 'Node');
            end if;
        end if;

        case v_action
            when 'list' then
                ret_cursor := list (v_os);

            when 'delete' then
                del (p_op_uid, v_os);

            when 'insert' then
                ins (p_op_uid, v_os, p_description);

            when 'replace' then
                supplant (p_op_uid, v_os, v_os_new);

            when 'update' then
                upd (p_op_uid, v_os, v_os_new, p_description);

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;

        if (ret_cursor is null) then
            ret_cursor := NetDB_Util.get_null_generic_cursor ();
        end if;

        return ret_cursor;
    end master;

end NetDB_OS;
/
