-- NetDB_Person.sql --- contains functions which deal with person

create or replace package body NetDB_Person as

    -- Get the name for a person.
    function name (p_person_id in int) return varchar
    as
        v_name Person.name%TYPE;
    begin
        select name into v_name from Person where id = verify_or_die (p_person_id);

        return v_name;
    end name;


    -- returns all of a persons data, which can be used for keys such as mod.by, created.by, etc.
    function info (p_person_id int) return NetDB.ref_cursor
    as
        ret_cursor NetDB.ref_cursor;
        person_row Person%ROWTYPE;
        out_record result_set := result_set ();
    begin
        select * into person_row from Person where id = p_person_id;

        out_record.extend(9);
        out_record(1) := result_row ('key', 'NetID',    0, 'string',  person_row.netid,        null, null);
        out_record(2) := result_row ('key', 'id',       0, 'string',  person_row.id,           null, null);
        out_record(3) := result_row ('key', 'lastname', 0, 'string',  person_row.lastname,     null, null);
        out_record(4) := result_row ('key', 'name',     0, 'string',  person_row.name,         null, null);
        out_record(5) := result_row ('key', 'DirID',    0, 'string',  person_row.directoryid,  null, null);
        out_record(6) := result_row ('key', 'RegID',    0, 'string',  person_row.person_regid, null, null);
        out_record(7) := result_row ('key', 'phone',    0, 'string',  person_row.phone,        null, null);
        out_record(8) := result_row ('key', 'email',    0, 'string',  person_row.email,        null, null);
        out_record(9) := result_row ('key', 'active?',  0, 'boolean', null,                    null, person_row.active);

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end info;


    procedure del (p_person_id in int) as
    begin
        delete from Person where id = p_person_id;
    end del;


    -- Insert or update (yes, update) a person.
    function ins ( p_person_id    in int,
                   p_person_regid in varchar,
                   p_name         in varchar,
                   p_netid        in varchar,
                   p_email        in varchar,
                   p_phone        in varchar,
                   p_active       in int,
                   p_lastname     in varchar,
                   p_directoryid  in varchar ) return NetDB.ref_cursor
    as
        -- If you change the default value of v_person_id,
        -- make sure you change the test for insertion.
        v_person_id int := null;

        out_record  result_set := result_set ();
        ret_cursor  NetDB.ref_cursor;
    begin
        NetDB_Util.validate (p_name, 'name');
        NetDB_Util.validate (p_netid, 'netid');
        NetDB_Util.validate (p_email, 'email');
        NetDB_Util.validate (p_phone, 'phone number');
        NetDB_Util.validate (p_lastname, 'last name');


        if (p_active not in (1, 0)) then
            raise_application_error (NetDB.ex_bad_parameter,
                '"Active" flag values must be either 0 or 1.');
        end if;

        update Person set name         = p_name,
                          netid        = p_netid,
                          email        = p_email,
                          phone        = p_phone,
                          active       = p_active,
                          lastname     = lower (p_lastname),
                          directoryid  = p_directoryid
         where id = p_person_id returning id into v_person_id;

        if (v_person_id is null) then
            insert into Person
                   ( person_regid,   name,     netid,              email,
                     phone,          active,   lastname,           directoryid )
            values ( p_person_regid, p_name,   p_netid,            p_email,
                     p_phone,        p_active, lower (p_lastname), p_directoryid )
            returning id into v_person_id;
        end if;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'id', null, 'string', v_person_id, null, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Verify that an id or person reg_id exists.
    function verify_or_die ( p_person_id in int     := null,
                             p_handle    in varchar := null ) return int
    as
        v_identifier varchar(256) := p_handle;
        v_person_id  int := 0;
    begin
        if (p_person_id is null and p_handle is null) then
            NetDB_Util.throw_no_record_id_handle ();
        elsif (p_person_id is not null and p_handle is not null) then
            select id into v_person_id from Person
             where id = p_person_id and lower (person_regid) = lower (p_handle);
        elsif (p_handle is not null) then
            select id into v_person_id from Person where lower (person_regid) = lower (p_handle);
        else -- p_person_id is not null
            v_identifier := p_person_id;

            select id into v_person_id from Person where id = p_person_id;
        end if;

        return v_person_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('person', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('person');
    end verify_or_die;


    function master ( p_action       in varchar,
                      p_op_uid       in int      := null, -- XXX: Not being used. Remove when the middleware can be updated
                      p_person_id    in int      := null,
                      p_person_regid in varchar  := null,
                      p_name         in varchar  := null,
                      p_netid        in varchar  := null,
                      p_email        in varchar  := null,
                      p_phone        in varchar  := null,
                      p_active       in int      := null,
                      p_lastname     in varchar  := null,
                      p_directoryid  in varchar  := null ) return NetDB.ref_cursor
    as
        v_action    varchar(16) := lower (p_action);
        v_person_id int         := null;

        ret_cursor NetDB.ref_cursor := null;
    begin
        -- If the verify fails, and person_id was given, reraise the error under the assumption
        -- that it really was supposed to be an update (or some other action). If the ID isn't
        -- given, assume that the data is an insert and continue on.
        begin
            v_person_id := verify_or_die (p_person_id, p_person_regid);
        exception
            when NetDB.record_not_found then
                if (p_person_id is not null) then
                    raise;
                end if;
        end;

        if (v_action = 'delete') then
            del (v_person_id);

        elsif (v_action = 'info') then
            ret_cursor := info (v_person_id);

        elsif (v_action in ('insert', 'update')) then
            -- This is not a bug. _Person.ins (...) knows how and when to update.
            ret_cursor := ins ( v_person_id,  p_person_regid, p_name,   p_netid,
                                p_email,      p_phone,        p_active, p_lastname,
                                p_directoryid );

        else
            NetDB_Util.throw_unknown_action (p_action);
        end if;

        if (ret_cursor is null) then
            ret_cursor := NetDB_Util.get_null_generic_cursor ();
        end if;

        return ret_cursor;
    end master;

end NetDB_Person;
/
