-- NetDB_State.sql --- State package body.

create or replace package body NetDB_State as

    -- Checks if the user has access to a particular state.
    function check_access ( p_user_id in int,
                            p_state   in varchar ) return boolean
    as
        v_state   State.state%TYPE := verify_or_die (p_state);
        state_row State%ROWTYPE;

        access_group AccessGroup.label%TYPE;
    begin
        select * into state_row from State where lower (state) = lower (v_state);

        if (state_row.accessgroup is null) then return true; end if;

        access_group := NetDB_Access.handle (state_row.accessgroup);

        return NetDB_Access.check_access (p_user_id, access_group, null, true);
    end check_access;


    -- A convenient wrapper for check_access.
    -- If the user doesn't have access, raise NetDB.ex_no_access_to_record_type.
    procedure has_access_or_die ( p_user_id in int,
                                  p_state   in varchar )
    as
        v_state State.state%TYPE := verify_or_die (p_state);
    begin
        if (not check_access (p_user_id, v_state)) then
            raise_application_error (NetDB.ex_no_access_to_record_type,
                'The user "' || NetDB_User.handle (p_user_id) ||
                '" does not have access to the state "' || v_state || '".');
        end if;
    end has_access_or_die;


    -- returns list of states restricted by p_filter
    -- if p_filter is null then all states are shown
    function list (p_filter in varchar) return NetDB.ref_cursor
    as
        out_record result_set := result_set ();
        ret_cursor NetDB.ref_cursor;
    begin
        for r in (select State.*, AccessGroup.label agroup
                    from State
                    left join AccessGroup on State.accessgroup = AccessGroup.id
                   where lower (state) like lower (nvl (p_filter, '%')) )
        loop
            out_record.extend(7);
            out_record(out_record.COUNT - 6) := result_row ('data',  r.state,      0,     null,     null,           null, null);
            out_record(out_record.COUNT - 5) := result_row ('key',  'name',        null, 'string',  r.state,        null, null);
            out_record(out_record.COUNT - 4) := result_row ('key',  'rank',        null, 'integer', null,           null, r.rank);
            out_record(out_record.COUNT - 3) := result_row ('key',  'denyDHCP',    null, 'boolean', null,           null, r.denyDHCP);
            out_record(out_record.COUNT - 2) := result_row ('key',  'denyDNS',     null, 'boolean', null,           null, r.denyDNS);
            out_record(out_record.COUNT - 1) := result_row ('key',  'description', null, 'string',  r.description,  null, null);
            out_record(out_record.COUNT)     := result_row ('key',  'access',      null, 'string',  r.agroup,       null, null);
        end loop;

        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end list;


    -- Delete a state.
    procedure del (p_state in varchar) as
    begin
        delete State where lower (state) = lower (p_state);
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('State', p_state);
    end del;


    -- Insert a state.
    procedure ins ( p_op_uid       in int,
                    p_state        in varchar,
                    p_rank         in int,
                    p_denyDHCP     in int,
                    p_denyDNS      in int,
                    p_description  in varchar,
                    p_access_group in varchar )
    as
        v_access_group int;
    begin
        NetDB_Util.validate (p_state, 'state', 'States must be named.');
        NetDB_Util.validate (p_description, 'description');

        if (p_access_group is not null) then
            v_access_group := NetDB_Access.verify_or_die (null, p_access_group);
        end if;

        insert into State (  state,   rank,   denyDHCP,   denyDNS,   description,   accessgroup)
                   values (p_state, p_rank, p_denyDHCP, p_denyDNS, p_description, v_access_group);
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('state', p_state);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Update a state.
    procedure upd ( p_op_uid       in int,
                    p_state        in varchar,
                    p_state_new    in varchar,
                    p_rank         in int,
                    p_denyDHCP     in int,
                    p_denyDNS      in int,
                    p_description  in varchar,
                    p_access_group in varchar )
    as
        v_access_group int;
    begin
        NetDB_Util.validate (p_state_new, 'state', 'States must be named.');
        NetDB_Util.validate (p_description, 'description');

        if (p_access_group is not null) then
            v_access_group := NetDB_Access.verify_or_die (null, p_access_group);
        end if;

        update State set state       = p_state_new,
                         rank        = p_rank,
                         denyDHCP    = p_denyDHCP,
                         denyDNS     = p_denyDNS,
                         description = p_description,
                         accessgroup = v_access_group
         where lower (state) = lower (p_state);
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('state', p_state_new);
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    function verify_or_die (p_state in varchar) return varchar
    as
        v_state State.state%TYPE;
    begin
        select state into v_state from State where lower (state) = lower (p_state);

        return v_state;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('state', p_state);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('state');
    end verify_or_die;


    -- main (interfacing) function to access and manipulate states
    function master ( p_action       in varchar,
                      p_op_uid       in int     := null,
                      p_state        in varchar := null,
                      p_state_new    in varchar := null,
                      p_rank         in int     := null,
                      p_denyDHCP     in int     := null,
                      p_denyDNS      in int     := null,
                      p_description  in varchar := null,
                      p_access_group in varchar := null ) return NetDB.ref_cursor
    as
        v_action varchar(16) := lower (p_action);
        v_state  State.state%TYPE;

        ret_cursor NetDB.ref_cursor := null;
    begin
        if (v_action in ('delete', 'update')) then
            v_state := verify_or_die (p_state);
        end if;

        if (v_action in ('delete', 'insert', 'update')) then
            NetDB_Access.has_access_or_die (p_op_uid, 'state');
        end if;

        case v_action
            when 'list' then
                ret_cursor := list (p_state);

            when 'delete' then
                del (p_state);

            when 'insert' then
                ins (p_op_uid, p_state, p_rank, p_denyDHCP, p_denyDNS, p_description, p_access_group);

            when 'update' then
                upd (p_op_uid, p_state, p_state_new, p_rank, p_denyDHCP, p_denyDNS, p_description, p_access_group);

            else
                NetDB_Util.throw_unknown_action (p_action);
        end case;

        return ret_cursor;
    end master;

end NetDB_State;
/
