-- NetDB_User.sql --- contains functions related to users

create or replace package body NetDB_User as

    sys_time_stamp timestamp;

    -- Reset prefixes that don't exists in the IP address table.
    procedure update_prefix as
    begin
        update Users set prefix = null where prefix not in (select prefix from IPaddressSpace);
    end update_prefix;


    -- Update all prefixes in the Users table. Sets all old_prefixes to new_prefix,
    -- and any prefixes that are contained in new_prefix to null.
    procedure update_prefix ( old_prefix in varchar,
                              new_prefix in varchar := null )
    as
        new_base      Users.prefix%TYPE := IP.normalize (IP.host (new_prefix));
        new_broadcast Users.prefix%TYPE := IP.normalize (IP.broadcast (new_prefix));
    begin
        update Users set prefix = new_prefix where prefix = IP.normalize (old_prefix);

        update Users set prefix = null
         where prefix != IP.normalize (new_prefix) and prefix between new_base and new_broadcast;
    end update_prefix;


    -- Gets the status of a users "active" flag.
    function is_active (p_user_id in int) return boolean
    as
        v_active int := 0;
    begin
        select active into v_active from Users where id = verify_or_die (p_user_id);
        return Util.to_boolean (v_active);
    end is_active;


    -- Gets the status of a users "all_groups" flag.
    function has_all_groups (p_user_id int) return boolean
    as
        v_all_groups int := 0;
    begin
        select all_groups into v_all_groups from Users where id = verify_or_die (p_user_id);
        return Util.to_boolean (v_all_groups);
    end has_all_groups;


    -- Gets the status of a users "all_records" flag.
    function has_all_records (p_user_id int) return boolean
    as
        v_all_records int := 0;
    begin
        select all_records into v_all_records from Users where id = verify_or_die (p_user_id);
        return Util.to_boolean (v_all_records);
    end has_all_records;


    -- Gets the default domain for a user. If they don't have one, the default domain in
    -- GlobalDefault is used. If a global default isn't set, "." is used.
    function def_domain (p_user_id in int) return int
    as
        v_def_domain_id int := null;
    begin
        select coalesce (domain, NetDB_Default.get_id ('DefaultDomain')) into v_def_domain_id
          from Users where id = verify_or_die (p_user_id);

        return v_def_domain_id;
    end def_domain;


    -- Update a users profile (default {group, domain, prefix})
    procedure upd_profile ( p_op_uid      in int,
                            p_def_group   in varchar,
                            p_def_domain  in varchar,
                            p_def_prefix  in varchar )
    as
        v_active boolean := is_active (p_op_uid);

        v_owner_id  Users.owner%TYPE  := null;
        v_prefix_id Users.prefix%TYPE := null;
        v_domain_id Users.domain%TYPE := NetDB_Domain.verify_or_die (null, p_def_domain);
    begin
        -- Inactive users are not required to have a default group.
        if (v_active = true or (v_active = false and p_def_group is not null)) then
            v_owner_id := NetDB_Owner.verify_or_die (null, p_def_group);

            if (not NetDB_Owner.is_member_of_group (p_op_uid, v_owner_id)) then
                raise_application_error (NetDB.ex_not_a_member_of_group,
                    'The user "' || NetDB_User.handle (p_op_uid) || '" must be a member of "' ||
                    NetDB_Owner.handle (v_owner_id) || '" in order to set it as the default group.');
            end if;

            -- Users with all groups can set their default group to anything, without actually
            -- being a member. If they don't have all groups, they can't get to this point in
            -- the code.
            begin
                insert into Users_Owner values (p_op_uid, v_owner_id);
            exception
                when NetDB.unique_constraint_failed then
                    null;
            end;
        end if;

        if (p_def_prefix is not null) then
            v_prefix_id := NetDB_IPaddressSpace.verify_or_die (null, p_def_prefix);
        end if;


        update Users set domain      = v_domain_id,
                         owner       = v_owner_id,
                         prefix      = p_def_prefix,
                         modified_by = p_op_uid,
                         modified    = sys_time_stamp
         where id = p_op_uid;

        NetDB_Log.ins (p_op_uid, 'update', p_op_uid, 'user');
    end upd_profile;


    -- Insert, delete or modify LNA entries.
    procedure org ( p_action   in varchar,  -- { insert, update, delete }
                    p_op_uid   in int,
                    p_lock     in timestamp,
                    p_user_id  in int,
                    p_org_id   in int     := null,
                    p_org_name in varchar := null,
                    p_official in int     := 0 )
    as
        v_org_id  int := NetDB_Organization.verify_or_die (p_org_id, p_org_name);
        v_user_id int := NetDB_User.verify_or_die (p_user_id, null);
    begin
        NetDB_User.locked_or_die (p_op_uid, p_lock, p_user_id);

        if (lower (p_action) in ('insert', 'update') and not p_official in (0, 1)) then
            raise_application_error (NetDB.ex_bad_parameter,
                '"Official" flag values must be either 0 or 1.');
        end if;

        case lower (p_action)
            when 'insert' then
                insert into LNA (lna,       organization, official)
                         values (p_user_id, v_org_id,     p_official);

            when 'update' then
                update LNA set official = p_official
                 where lna = p_user_id and organization = v_org_id;

            when 'delete' then
                delete from LNA where lna = p_user_id and organization = v_org_id;

            else
                NetDB_Util.throw_unknown_action (p_action);

        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;
    end org;


    -- Add or remove an Users_AccessGroup row.
    procedure users_access ( p_action      in varchar,  -- { add, remove }
                             p_op_uid      in int,
                             p_lock        in timestamp,
                             p_user_id     in int,
                             p_access_id   in int     := null,
                             p_access_name in varchar := null )
    as
        v_access_id int := NetDB_Access.verify_or_die (p_access_id, NetDB_RecordType.access (p_access_name));
    begin
        locked_or_die (p_op_uid, p_lock, p_user_id);
        NetDB_Access.has_access_or_die (p_op_uid, NetDB_Access.handle (v_access_id));

        case lower (p_action)
            when 'add' then
                insert into Users_AccessGroup (userid, accessgroup) values (p_user_id, v_access_id);

            when 'remove' then
                delete from Users_AccessGroup where userid = p_user_id and accessgroup = v_access_id;

            else
                NetDB_Util.throw_unknown_action (p_action);

        end case;
    exception
        when NetDB.unique_constraint_failed then
            null;
    end users_access;


    -- Sets a users "last login" to "now".
    procedure update_last_login (p_user_id in int) as begin
        update Users set last_login = sys_time_stamp where id = p_user_id and active = 1;
    end update_last_login;


    -- Check to see if a user has a lock for a record.
    procedure locked_or_die ( p_op_uid  in int,
                              p_lock    in timestamp,
                              p_user_id in int )
    as
        v_lock timestamp := null;
    begin
        select locked into v_lock from Users
         where id = p_user_id and locked_by = p_op_uid and locked = p_lock and
               Util.datediff ('second', p_lock) < NetDB_Default.get_int ('LockTimeOut');
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_not_locked ('user', handle (p_user_id), handle (p_op_uid));
    end locked_or_die;


    -- Lock a record.
    function lock_record ( p_op_uid  in int,
                           p_user_id in int ) return NetDB.ref_cursor
    as
        v_lock       timestamp;
        v_locked_by  int;

        out_record result_set       := result_set ();
        ret_cursor NetDB.ref_cursor := null;
    begin
        select locked, locked_by into v_lock, v_locked_by from Users where id = p_user_id;

        -- if not locked, locked by this user, or expired lock, we're good to go
        if ( v_lock is null or v_locked_by = p_op_uid or
             Util.datediff ('second', v_lock) > NetDB_Default.get_int ('LockTimeOut'))
        then
               update users set locked    = sys_time_stamp,
                                locked_by = p_op_uid
                where id = p_user_id
            returning locked into v_lock;
        else
            NetDB_Util.throw_locked_by_other ('user', handle (p_user_id), handle (v_locked_by));
        end if;

        out_record.extend(1);
        out_record(1) := result_row ('key', 'locked', 0, 'timestamp', null, v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    end lock_record;


    -- Unlock a record.
    procedure unlock_record ( p_op_uid  in int,
                              p_lock    in timestamp,
                              p_user_id in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_user_id);

        update users set locked    = null,
                         locked_by = null
         where id = p_user_id and locked = p_lock and locked_by = p_op_uid;
    end unlock_record;


    -- Get the info on a user.
    function info (p_user_id in int) return result_set pipelined
    as
        v_sort_name varchar(256) := null;
        res_row     result_row   := result_row (null, null, null, null, null, null, null);
        data_cursor NetDB.ref_cursor;

        -- User info expressed as a cursor (instead of defining a gazillion variables).
        cursor user_info is
        select u.active,             u.all_groups,            u.all_records,         u.comments,
               u.created,            u.created_by,            u.last_login,          u.modified,
               u.modified_by,        p.active dir_active,     p.directoryid dir_id,  p.email,
               p.name,               p.netid,                 p.person_regid reg_id, p.phone,
               d.domain def_domain, ip.prefix_txt def_prefix, o.label def_group
          from Users u
          left join IPAddressSpace ip on ip.prefix = u.prefix
          left join Domain d          on u.domain  = d.id
          left join Owner  o          on o.id      = u.owner
               join Person p          on p.id      = u.id
         where u.id = p_user_id;

        i user_info%ROWTYPE;
    begin
        open  user_info;
        fetch user_info into i;
        close user_info;

        pipe row (result_row ('key', 'object',             null, 'string',    'user',       null,         null));
        pipe row (result_row ('key', 'id',                 null, 'string',    p_user_id,    null,         null));
        pipe row (result_row ('key', 'name',               null, 'string',    i.name,       null,         null));
        pipe row (result_row ('key', 'phone',              null, 'string',    i.phone,      null,         null));
        pipe row (result_row ('key', 'email',              null, 'string',    i.email,      null,         null));
        pipe row (result_row ('key', 'NetID',              null, 'string',    i.netid,      null,         null));
        pipe row (result_row ('key', 'DirID',              null, 'string',    i.dir_id,     null,         null));
        pipe row (result_row ('key', 'RegID',              null, 'string',    i.reg_id,     null,         null));
        pipe row (result_row ('key', 'active?',            null, 'boolean',   null,         null,         i.active));
        pipe row (result_row ('key', 'dirActive?',         null, 'boolean',   null,         null,         i.dir_active));
        pipe row (result_row ('key', 'all_objects?',       null, 'boolean',   null,         null,         i.all_records));
        pipe row (result_row ('key', 'all_groups?',        null, 'boolean',   null,         null,         i.all_groups ));
        pipe row (result_row ('key', 'default_domain',     null, 'string',    i.def_domain, null,         null));
        pipe row (result_row ('key', 'default_addr_space', null, 'string',    i.def_prefix, null,         null));
        pipe row (result_row ('key', 'default_group',      null, 'string',    i.def_group,  null,         null));
        pipe row (result_row ('key', 'comment',            null, 'string',    i.comments,   null,         null));
        pipe row (result_row ('key', 'created',            null, 'timestamp', null,         i.created,    null));
        pipe row (result_row ('key', 'modified',           null, 'timestamp', null,         i.modified,   null));
        pipe row (result_row ('key', 'last_login',         null, 'timestamp', null,         i.last_login, null));

        -- Get the created_by info.
        pipe row (result_row ('data', 'created.by', 0, null, null, null, null));

        data_cursor := NetDB_Person.master ('info', null, i.created_by);
        loop
            fetch data_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                   res_row.value_string, res_row.value_timestamp, res_row.value_int;
            exit when data_cursor%notfound;

            pipe row (res_row);
        end loop;
        close data_cursor;

        -- Get the modified_by info.
        pipe row (result_row ('data', 'mod.by', 0, null, null, null, null));

        data_cursor := NetDB_Person.master ('info', null, i.modified_by);
        loop
            fetch data_cursor into res_row.type, res_row.key, res_row.hash_level, res_row.value_type,
                                   res_row.value_string, res_row.value_timestamp, res_row.value_int;
            exit when data_cursor%notfound;

            pipe row (res_row);
        end loop;
        close data_cursor;

        -- Get the groups.
        pipe row (result_row ('data', 'groups', 0, null, null, null, null));

        for r in ( select Owner.label
                    from Users_Owner, Owner
                   where Users_Owner.userid = p_user_id and Users_Owner.owner = Owner.id
                order by lower (Owner.label) )
        loop
            pipe row (result_row ('key', r.label, null, 'string', ' ', null, null));
        end loop;

        -- Get the records.
        pipe row (result_row ('data', 'records', 0, null, null, null, null));

        for r in ( select ag.label, ag.display_label
                     from AccessGroup ag, Users_AccessGroup uag
                    where uag.userid = p_user_id and ag.id = uag.accessgroup )
        loop
            pipe row (result_row ('key', r.label, null, 'string', r.display_label, null, null));
        end loop;

        -- Get the departments.
        pipe row (result_row ('data', 'departments', 0, null, null, null, null));

        for r in ( select o.id, o.organization org, o.org_regid, LNA.official
                     from LNA, Organization o
                    where LNA.lna = p_user_id and LNA.organization = o.id )
        loop
            pipe row (result_row ('data', r.org,      1,     null,     null,        null, null));
            pipe row (result_row ('key', 'RegID',     null, 'string',  r.org_regid, null, null));
            pipe row (result_row ('key', 'official?', null, 'boolean', null,        null, r.official));

            -- now consultants for this department
            pipe row (result_row ('data', 'consultants', 2, null, null, null, null));

            for d in ( select c.primary, p.name, p.directoryid dir_id, p.netid, p.person_regid reg_id, p.id
                         from Person p, Consultant c
                        where p.id = c.person and c.organization = r.id )
            loop
                v_sort_name := d.netid || ' ' || d.name || ' ' || d.reg_id || ' ' || r.id;

                pipe row (result_row ('data', v_sort_name, 3,    null,      null,     null, null));
                pipe row (result_row ('key', 'primary?',   null, 'boolean', null,     null, d.primary));
                pipe row (result_row ('key', 'name',       null, 'string',  d.name,   null, null));
                pipe row (result_row ('key', 'DirID',      null, 'string',  d.dir_id, null, null));
                pipe row (result_row ('key', 'NetID',      null, 'string',  d.netid,  null, null));
                pipe row (result_row ('key', 'RegID',      null, 'string',  d.reg_id, null, null));
            end loop;
        end loop;
    end info;


    -- Delete a record and log the action.
    procedure del ( p_op_uid  in int,
                    p_lock    in timestamp,
                    p_user_id in int ) as
    begin
        locked_or_die (p_op_uid, p_lock, p_user_id);
        NetDB_Log.ins (p_op_uid, 'delete', p_user_id, 'user');

        delete from Users where id = p_user_id;
    exception
        when NetDB.children_exist then
            NetDB_Util.throw_still_in_use ('user', handle (p_user_id));
    end del;


    -- Inserting users record.
    function ins ( p_op_uid      in int,
                   p_person_id   in varchar,
                   p_def_group   in varchar,
                   p_def_domain  in varchar,
                   p_def_prefix  in varchar,
                   p_active      in int,
                   p_all_records in int,
                   p_all_groups  in int,
                   p_comment     in varchar ) return NetDB.ref_cursor
    as
        v_user_id   int := 0;
        v_prefix_id int := 0;
        v_owner_id  int := 0;
        v_domain_id int := NetDB_Domain.verify_or_die (null, p_def_domain);
        v_lock      timestamp := null;

        out_record  result_set       := result_set ();
        ret_cursor  NetDB.ref_cursor := null;
    begin
        NetDB_Util.validate (p_comment, 'comment');

        if (p_active = 1 or (p_active = 0 and p_def_group is not null)) then
            v_owner_id := NetDB_Owner.verify_or_die (null, p_def_group);
        end if;

        if (p_all_records = 1 and not has_all_records (p_op_uid)) then
            raise_application_error (NetDB.ex_no_access_to_record,
                '"' || handle (p_op_uid) || '" is not allowed to set the all_records flag.');
        end if;

        if (p_all_groups = 1 and not has_all_groups (p_op_uid)) then
            raise_application_error (NetDB.ex_no_access_to_record,
                '"' || handle (p_op_uid) || '" is not allowed to set the all_groups flag.');
        end if;

        if (p_def_prefix is not null) then
            v_prefix_id := NetDB_IPaddressSpace.verify_or_die (null, p_def_prefix);
        end if;

        -- get person_id, to be passed as user_id for insert
        v_user_id := NetDB_Person.verify_or_die (p_person_id);

        insert into users
               ( id,                domain,                owner,                  prefix,
                 active,            all_groups,            all_records,            comments,
                 created_by,        modified_by,           created,                modified,
                 locked_by,         locked )
        values ( v_user_id,         v_domain_id,           v_owner_id,             p_def_prefix,
                 nvl (p_active, 0), nvl (p_all_groups, 0), nvl (p_all_records, 0), p_comment,
                 p_op_uid,          p_op_uid,              sys_time_stamp,         sys_time_stamp,
                 p_op_uid,          sys_time_stamp )
        returning locked into v_lock;

        NetDB_Log.set_action ('insert');

        out_record.extend(2);
        out_record(1) := result_row ('key', 'id',     0, 'string',    v_user_id, null,   null);
        out_record(2) := result_row ('key', 'locked', 0, 'timestamp', null,      v_lock, null);
        open ret_cursor for
            select * from table (cast (out_record as result_set));

        return ret_cursor;
    exception
        when NetDB.unique_constraint_failed then
            NetDB_Util.throw_record_already_exists ('User', handle (v_user_id));
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end ins;


    -- Update a record
    procedure upd ( p_op_uid      in int,
                    p_lock        in timestamp,
                    p_user_id     in int,
                    p_def_group   in varchar,
                    p_def_domain  in varchar,
                    p_def_prefix  in varchar,
                    p_active      in int,
                    p_all_records in int,
                    p_all_groups  in int,
                    p_comment     in varchar )
    as
        v_active    boolean := is_active (p_user_id);
        v_prefix_id int     := 0;
        v_owner_id  int     := 0;
        v_domain_id int     := NetDB_Domain.verify_or_die (null, p_def_domain);

        user_values Users%ROWTYPE;
    begin
        locked_or_die (p_op_uid, p_lock, p_user_id);

        NetDB_Util.validate (p_comment, 'comment');

        if (v_active = true or (v_active = false and p_def_group is not null)) then
            v_owner_id := NetDB_Owner.verify_or_die (null, p_def_group);
        end if;

        select * into user_values from Users where id = p_user_id;

        if (p_all_records != user_values.all_records and not has_all_records (p_op_uid)) then
            raise_application_error (NetDB.ex_no_access_to_record,
                '"' || handle (p_op_uid) || '" is not allowed to modify the all_records flag.');
        end if;

        if (p_all_groups != user_values.all_groups and not has_all_groups (p_op_uid)) then
            raise_application_error (NetDB.ex_no_access_to_record,
                '"' || handle (p_op_uid) || '" is not allowed to modify the all_groups flag.');
        end if;

        if (p_active not in (1, 0)) then
            raise_application_error (NetDB.ex_bad_parameter,
                '"Active" flag values must be either 0 or 1.');
        end if;

        if (p_def_prefix is not null) then
            v_prefix_id := NetDB_IPaddressSpace.verify_or_die (null, p_def_prefix);
        end if;

        update Users set domain      = v_domain_id,
                         owner       = v_owner_id,
                         prefix      = p_def_prefix,
                         active      = p_active,
                         all_groups  = coalesce (p_all_groups,  0),
                         all_records = coalesce (p_all_records, 0),
                         comments    = p_comment,
                         modified_by = p_op_uid,
                         modified    = sys_time_stamp
         where id = p_user_id;

        NetDB_Log.set_action ('update');
    exception
        when NetDB.value_too_large then
            NetDB_Util.throw_field_too_long (sqlerrm);
    end upd;


    -- Verify user is active, has record in users_owner, department and
    -- has rights to default domain and address space. Then inserts the record into the log.
    procedure verify_and_log ( p_op_uid  in int,
                               p_user_id in int )
    as
        v_def_domain int          := 0;
        v_def_group  int          := 0;
        v_def_prefix varchar(256) := null;

        v_active     boolean := is_active (p_user_id);
        v_domain_id  int     := 0;
        v_owner_id   int     := 0;
        v_prefix_id  int     := 0;
    begin
        select count (*) into NetDB.existence_counter from LNA where lna = p_user_id;
        if (NetDB.existence_counter < 1) then
            raise_application_error (NetDB.ex_user_must_have_dept,
                'The user "' ||  handle (p_user_id) || '" must have at least one department.');
        end if;

        select domain,       prefix,       owner
          into v_def_domain, v_def_prefix, v_def_group
          from Users where id = p_user_id;

        -- Inactive users are not required to have a default group.
        if (v_active = true or (v_active = false and v_def_group is not null)) then
            v_owner_id := NetDB_Owner.verify_or_die (v_def_group);
        end if;

        v_domain_id := NetDB_Domain.verify_or_die (v_def_domain);
        NetDB_Owner.has_access_or_die (p_user_id, v_def_domain, 'Domain', true, false);

        if (v_def_prefix is not null) then
            v_prefix_id := NetDB_IPaddressSpace.verify_or_die (null, v_def_prefix);

            NetDB_Owner.has_access_or_die (p_user_id, v_prefix_id, 'AddressSpace', true, false);
        end if;

        NetDB_Log.ins (p_op_uid, NetDB_Log.get_action (), p_user_id, 'user');
    end verify_and_log;


    -- Verify user ID / handle or complain.
    function verify_or_die ( p_user_id in int     := null,
                             p_netid   in varchar := null ) return int
    as
        v_identifier varchar(256) := p_netid;
        v_user_id    int;
    begin
        if (p_user_id is null and p_netid is null) then
            NetDB_Util.throw_no_record_id_handle ();
        elsif (p_user_id is not null and p_netid is not null) then
            select id into v_user_id from Users where id = p_user_id and netid = p_netid;
        elsif (p_netid is not null) then
            select id into v_user_id from Users where netid = p_netid;
        else -- p_user_id is not null
            v_identifier := p_user_id;

            select id into v_user_id from Users where id = p_user_id;
        end if;

        return v_user_id;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('user', v_identifier);
        when TOO_MANY_ROWS then
            NetDB_Util.throw_single_record_expected ('user');
    end verify_or_die;


    -- Returns the netid for a user.
    function handle (p_user_id in int) return varchar
    as
        v_netid varchar(256) := null;
    begin
        select netid into v_netid from Users where id = p_user_id;

        return v_netid;
    exception
        when NO_DATA_FOUND then
            NetDB_Util.throw_record_does_not_exist ('user', p_user_id);
    end handle;


    -- main (interfacing) function for manipulating and retrieving user information
    function master ( p_action      in varchar,
                      p_op_uid      in int       := null,
                      p_lock        in timestamp := null,
                      p_user_id     in int       := null,
                      p_netid       in varchar   := null,
                      p_person_id   in int       := null,
                      p_def_group   in varchar   := null,
                      p_def_domain  in varchar   := null,
                      p_def_prefix  in varchar   := null,
                      p_active      in int       := null,
                      p_all_records in int       := null,
                      p_all_groups  in int       := null,
                      p_comment     in varchar   := null ) return NetDB.ref_cursor
    as
        v_user_id  int;
        v_action   varchar(16) := lower (p_action);
        v_active   boolean     := false;

        ret_cursor NetDB.ref_cursor;
    begin
        sys_time_stamp := systimestamp;

        if (v_action in ('delete', 'info', 'insert', 'lock', 'login', 'profile', 'unlock', 'update', 'verify')) then
            if (v_action in ('insert', 'update')) then
                if (v_action = 'insert' or (v_action = 'update' and p_active is not null)) then
                    v_active := Util.to_boolean (p_active);
                elsif (v_action = 'update') then
                    v_active := is_active (p_user_id);
                end if;
            end if;

            if (v_action not in ('insert', 'profile')) then
                v_user_id := verify_or_die (p_user_id, p_netid);
            end if;

            if (v_action in ('delete', 'lock', 'insert', 'update')) then
                NetDB_Access.has_access_or_die (p_op_uid, 'User');
            end if;

            if (v_action in ('delete', 'lock', 'unlock', 'update')) then
                NetDB_Owner.has_access_or_die (p_op_uid, v_user_id, 'user');
            end if;
        end if;

        -- End data verification. Begin 'action' logic.

        if (v_action = 'delete') then
            del (p_op_uid, p_lock, v_user_id);

        elsif (v_action in ('info', 'login')) then
            open ret_cursor for
                select * from table (cast (info (v_user_id) as result_set));

            if (v_action = 'login') then
                update_last_login (v_user_id);
            end if;

        elsif (v_action = 'insert') then
            ret_cursor := ins ( p_op_uid,     p_person_id, p_def_group,   p_def_domain,
                                p_def_prefix, p_active,    p_all_records, p_all_groups,
                                p_comment );

        elsif (v_action = 'lock') then
            ret_cursor := lock_record (p_op_uid, v_user_id);

        elsif (v_action = 'profile') then
            upd_profile (p_op_uid, p_def_group, p_def_domain, p_def_prefix);

        elsif (v_action = 'unlock') then
            unlock_record (p_op_uid, p_lock, v_user_id);

        elsif (v_action = 'update') then
            upd ( p_op_uid,     p_lock,    v_user_id,     p_def_group,  p_def_domain,
                  p_def_prefix, p_active,  p_all_records, p_all_groups, p_comment );

        elsif (v_action = 'verify') then
            verify_and_log (p_op_uid, v_user_id);

        else
            NetDB_Util.throw_unknown_action (p_action);
        end if;

        if (ret_cursor is null) then
            ret_cursor := NetDB_Util.get_null_generic_cursor ();
        end if;

        return ret_cursor;
    end master;

end NetDB_User;
/
