Reserve Name Specification
==========================

Signature:

    NetDB_Name.reserve (
                         p_op_uid    int,       -- user reserving name
                         p_name      varchar,   -- name (unqualified) to reserve
                         p_domain_id int,       -- id of the domain part of the name
                         p_type      varchar    -- address alias domain interface mx network node
                       )
        returns (new name id, reservation timestamp)

Implementation:

    -- ORDER of TESTS MATTERS!!!

    -- validate type parameter and convert it to a NetDB DNS label
    case p_type in ("address","interface", "node", "network") dns_label = "A"
                 =  "mx"                                      dns_label = "MX"
                 =  "alias"                                   dns_label = "CNAME"
                 =  "domain"                                  dns_label = "SOA"
                 otherwise error

    -- make sure the name is composed of valid characters and is not too long
    if (not DNS.check_name(p_name)) error

    -- make sure the domain exists
    NetDB_Domain.verify_or_die(p_domain_id);

    -- disallow adding top-level names/domains, i.e., names with parent domain '.'
    if (p_domain_id == 1) error

    -- if the new name doesn't match an existing domain name, make
    --   sure the user has rights to assign names in the name's domain
    if ((p_name,p_domain_id,dns_type="SOA") doesn't exist and
        not NetDB_Domain.can_use (p_op_uid, p_name)) error

    -- make sure the name is not globally reserved (restricted)
    if (NetDB_ResName.reserved(p_op_uid,p_name,p_domain_id)) error

    -- if the domain is limited, make sure the name is allowed
    if (CheckForLimitedDomain from GlobalDefault is set and
        p_domain_id is flagged 'limited' and
        p_name is not in the LimitedNames table) error

    -- nuke expired name reservations so they don't get in the way
    delete from Name
     where reserved is not null  -- very important/necessary for speed!
       and Util.datediff('seconds',reserved) > GlobalDefault.LockTimeOut

    -- if the new name matches an existing domain name, see if
    --   that domain name can be used as other name types
    if ((p_name,p_domain_id,dns_type="SOA") exists and
        not NetDB_Domain.can_use_as_name (p_op_uid, p_domain_id)) error

    -- if the user already has this name reserved, refresh the reservation and return it
    if ((p_name,p_domain_id,reserved_by=p_op_uid,reserved!=null) exists) or
        update row, refreshing the reservation and return the new reservation

    -- make sure the name doesn't already exist, unless it's an MX
    if ((p_name,p_domain_id,dns_label) exists and dns_label != "mx") error

    -- alias names cannot coexist with other types [DNS rule]
    if ((p_name,p_domain_id) exists and
        (the dns_label of the existing name or the new name is "CNAME")) error

    -- if the name is an "A" name and the name already exists as an MX
    --     if the user has access to any record associated with the MX received_by names
    --         ok
    --     else
    --         error
    --
    NetDB_Owner.has_access_or_die (p_op_uid, existing_mx_id, 'MX')

    -- if the name is an MX
    --    if the name already exists as an "A" name
    --        if the user has access to the existing record associated with the "A" name
    --            ok
    --        else
    --            error
    --    else if the name already exists as an MX
    --        if the user has access to any record associated with the MX received_by names
    --            ok
    --        else
    --            error
    --
    NetDB_Owner.has_access_or_die (p_op_uid, existing_A_name_id, 'MX') or
    NetDB_Owner.has_access_or_die (p_op_uid, existing_mx_id, 'MX')

    -- reserve the name by inserting it with the reservation (user,timestamp)
    -- unless the DNS type is "mx" and that name already exists
    insert (p_name,p_domain_id,dns_type) into the name table
      with the reserved_by and reserved columns set to the p_op_uid and "now"
     where not exists (p_name,p_domain_id,dns_type="mx")
