set serveroutput on size 100000;
savepoint start_domain_tests;

declare
    test_number int          := 1;
    err_msg     varchar(256) := null;
    v_lock      timestamp    := null;
    v_count     int          := 0;

    test_cursor NetDB.ref_cursor;
    result      result_row := result_row (null, null, null, null, null, null, null);

    -- Values from the DB.
    db_lock      timestamp(3);
    db_locked_by int;
    db_domain    varchar(256);

    v_reserved   timestamp;

    -- Domains for insert and update.
    test_domain    Name.name%TYPE := 'TEST-1234509876-domain';
    test_domain_lc Name.name%TYPE := lower (test_domain);
    test_domain_id int            := null;
    test_name_id   int            := null;

    unrecoverable_error EXCEPTION;
begin
    select count (*) into v_count from Name where lower (name) like test_domain_lc;

    if (v_count > 0) then
        err_msg := 'The test domain "' || test_domain_lc || '" exists. If it is not a valid ' ||
                   'domain, remove the entry and run the tests again. If it is valid, the ' ||
                   'tests must be modified.';
        raise unrecoverable_error;
    end if;

    NetDB_Test.test_info ('Domain', test_number);
    NetDB_Test.test_locking ('Domain', test_number);

    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'Domain');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a domain not in the Name table.';

        test_cursor := NetDB_Domain.master ('insert', NetDB_Test.user_1.id, null, NetDB_Test.domain_1.id, test_domain_lc, null, 'test data');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a domain.';

        test_cursor    := NetDB_Name.reserve (NetDB_Test.user_1.id, test_domain_lc, NetDB_Test.domain_1.domain, 'Domain');
        result         := NetDB_Test.get_generic_row (test_cursor);
        test_domain_id := result.value_string;
        result         := NetDB_Test.get_generic_row (test_cursor);
        v_reserved     := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Domain.master ('insert', NetDB_Test.user_1.id, null, null, NetDB_Name.fullname (test_domain_id), null, 'test data');
        result      := NetDB_Test.get_generic_row (test_cursor); -- we already have the ID.
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, test_domain_id, null, 'Domain', test_domain_id, null, v_lock);

        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, test_domain_id, 'Domain');
        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    v_lock := NetDB_Test.force_lock (NetDB_Test.user_0.id, test_domain_id, 'Domain');
    NetDB_Owner.remove (NetDB_Test.user_0.id, v_lock, NetDB_Test.group_1.id, null, test_domain_id, 'domain_use');
    NetDB_Test.force_unlock (test_domain_id, 'Domain');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Try to name a node the same name as a domain (normal user).';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_domain_lc, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.cannot_use_domain_as_name then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Try to name a node the same name as a domain (god user).';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_0.id, test_domain_lc, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.cannot_use_domain_as_name then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    v_lock := NetDB_Test.force_lock (NetDB_Test.user_0.id, test_domain_id, 'Domain');
    NetDB_Owner.add (NetDB_Test.user_0.id, v_lock, NetDB_Test.group_1.id, null, test_domain_id, 'domain_use');
    NetDB_Test.force_unlock (test_domain_id, 'Domain');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Name a node the same name as a domain (normal user).';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_domain_lc, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.cancel (NetDB_Test.user_1.id, null, NetDB_Util.fullname (test_domain_lc, NetDB_Test.domain_1.domain), 'Node');

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Name a node the same name as a domain (god user, no record type access).';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_0.id, test_domain_lc, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.cannot_use_domain_as_name then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    -- Being a god user isn't enough. The user really, REALLY must have access to the record.
    NetDB_Test.grant_access (NetDB_Test.user_0.id, 'Domain');
    NetDB_Test.add_to_group (NetDB_Test.user_0.id, NetDB_Test.group_1.id);

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Name a node the same name as a domain (god user).';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_0.id, test_domain_lc, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.cancel (NetDB_Test.user_0.id, null, NetDB_Util.fullname (test_domain_lc, NetDB_Test.domain_1.domain), 'Node');

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a domain with a name change.';

        test_cursor := NetDB_Domain.master ('lock', NetDB_Test.user_1.id, null, test_domain_id);
        NetDB_Test.empty_cursor (test_cursor);

        select locked, locked_by into db_lock, db_locked_by from Domain where id = test_domain_id;

        test_cursor := NetDB_Domain.master ('update', NetDB_Test.user_1.id, db_lock, test_domain_id, null, 'new-test-domain', 'new test data (failed)', 0);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_renaming_domains then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a domain with a case change.';

        test_cursor := NetDB_Domain.master ('update', NetDB_Test.user_1.id, db_lock, test_domain_id, null, test_domain, 'new test data (ok)', 0);
        NetDB_Test.empty_cursor (test_cursor);

        select domain into db_domain from Domain where id = test_domain_id;

        if (substr (db_domain, 1, length (test_domain)) = test_domain) then NetDB_Test.pass ();
        else                                                                NetDB_Test.failed (err_msg);
        end if;
    exception
        when NetDB.no_renaming_domains then
            NetDB_Test.failed (err_msg);
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete a domain.';

        test_cursor := NetDB_Domain.master ('delete', NetDB_Test.user_1.id, db_lock, test_domain_id);
        NetDB_Test.empty_cursor (test_cursor);

        select count (*) into v_count from Domain where id = test_domain_id;

        if (v_count = 0) then NetDB_Test.pass ();
        else                  NetDB_Test.failed (err_msg);
        end if;

        delete from Name where id = test_domain_id;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'List the domains for a user.';

        test_cursor := NetDB_Domain.master ('list', NetDB_Test.user_1.id);
        NetDB_Test.empty_cursor (test_cursor);

        if (NetDB_Test.rows_skipped > 0) then NetDB_Test.pass ();
        else                                  NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete directly from the Domain table.';

        test_cursor    := NetDB_Name.reserve (NetDB_Test.user_1.id, test_domain_lc, NetDB_Test.domain_1.domain, 'Domain');
        result         := NetDB_Test.get_generic_row (test_cursor);
        test_domain_id := result.value_string;
        result         := NetDB_Test.get_generic_row (test_cursor);
        v_reserved     := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Domain.master ('insert', NetDB_Test.user_1.id, null, null, NetDB_Name.fullname (test_domain_id), null, 'test data');
        result      := NetDB_Test.get_generic_row (test_cursor); -- we already have the ID.
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, test_domain_id, null, 'Domain', test_domain_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, test_domain_id, 'Domain');

        delete from Domain where id = test_domain_id;

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.idiot_programmer then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;
exception
    when unrecoverable_error then
        dbms_output.put_line ('ERROR: All remaining tests aborted. An unrecoverable error occured: ' || err_msg);
    when others then
        dbms_output.put_line ('ERROR: Unhandled exception "' || sqlcode || '" with message "' || sqlerrm || '"');
        dbms_output.put_line (err_msg);
end;
/

rollback to start_domain_tests;
