set serveroutput on size 100000;
savepoint start_addressspace_tests;

declare
    test_number int          := 1;
    err_msg     varchar(256) := null;
    v_count     int          := 0;
    v_lock      timestamp;
    net_lock    timestamp;

    test_cursor NetDB.ref_cursor;
    result      result_row := result_row (null, null, null, null, null, null, null);

    -- values for insert and update.
    network_name Name.name%TYPE := 'valid-test-network-name';
    net_fullname Name.fullname%TYPE;
    net_name_id  int;
    network_id   int;
    v_reserved   timestamp;

    low_res  int := 5;
    high_res int := 5;

    a_comment     IPaddressSpace.comments%TYPE   := 'a comment';
    v_prefix      IPaddressSpace.prefix_txt%TYPE := '1.1.1.0/24';
    as_id         int;
    containing_id int;
    parent_id     int;
    verified_id   int;

    existing_as IPaddressSpace%ROWTYPE;

    unrecoverable_error EXCEPTION;
begin
    select count (*) into v_count from IPaddressSpace where prefix_txt = v_prefix;

    if (v_count > 0) then
        err_msg := 'The test address space "' || v_prefix || '" exists. If it is not a valid ' ||
                   'address space, remove the entry and run the tests again. If it is valid, the ' ||
                   'tests must be modified.';
        raise unrecoverable_error;
    end if;

    select count (*) into v_count from Name where name = network_name;

    if (v_count > 0) then
        err_msg := 'The test network "' || network_name || '" exists. If it is not a valid ' ||
                   'network, remove the entry and run the tests again. If it is valid, the ' ||
                   'tests must be modified.';
        raise unrecoverable_error;
    end if;

    select * into existing_as from IPaddressSpace where rownum = 1;

    NetDB_Test.test_info ('AddressSpace', test_number);

    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'network');
    NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a Network.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, network_name, NetDB_Test.domain_1.domain, 'Network');
        result      := NetDB_Test.get_generic_row (test_cursor);
        net_name_id := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        net_fullname := NetDB_Name.fullname (net_name_id);

        test_cursor := NetDB_Network.master ('insert', NetDB_Test.user_1.id, null, null, net_fullname, null, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        network_id  := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, net_name_id, null, 'Network', network_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, network_id, 'Network');
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_2.id, null, network_id, 'Network');

        test_cursor := NetDB_Network.master ('verify', NetDB_Test.user_1.id, v_lock, network_id);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'AddressSpace');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert an address space with bad low reserve size.';

        test_cursor := NetDB_IPaddressSpace.master ('insert', NetDB_Test.user_1.id, v_lock, null, v_prefix, network_id, v_lock, a_comment, -1, high_res);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.bad_parameter then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert an address space with bad high reserve size.';

        test_cursor := NetDB_IPaddressSpace.master ('insert', NetDB_Test.user_1.id, v_lock, null, v_prefix, network_id, v_lock, a_comment, low_res, -1);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.bad_parameter then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert an address space.';

        test_cursor := NetDB_IPaddressSpace.master ('insert', NetDB_Test.user_1.id, v_lock, null, v_prefix, network_id, v_lock, a_comment, low_res, high_res);
        result := NetDB_Test.get_generic_row (test_cursor);
        as_id  := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, as_id, 'AddressSpace');

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Get the parent ID for an address space.';

        parent_id := NetDB_IPaddressSpace.get_parent_id (as_id);

        if (parent_id = network_id) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Find the address space containing an address.';

        containing_id := NetDB_IPaddressSpace.containing (IP.plus (IP.host (v_prefix), 30));

        if (containing_id = as_id) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an address space.';

        test_cursor := NetDB_IPaddressSpace.master ('update', NetDB_Test.user_1.id, v_lock, null, v_prefix, network_id, v_lock, a_comment, low_res, high_res);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an address space with bad low reserve size.';

        test_cursor := NetDB_IPaddressSpace.master ('update', NetDB_Test.user_1.id, v_lock, null, v_prefix, network_id, v_lock, a_comment, -1, high_res);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.bad_parameter then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an address space with bad high reserve size.';

        test_cursor := NetDB_IPaddressSpace.master ('update', NetDB_Test.user_1.id, v_lock, null, v_prefix, network_id, v_lock, a_comment, low_res, -1);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.bad_parameter then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete an address space.';

        test_cursor := NetDB_IPaddressSpace.master ('delete', NetDB_Test.user_1.id, v_lock, as_id);
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_IPaddressSpace.verify_or_die (as_id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Move an address space between networks.';

        v_lock   := NetDB_Test.force_lock (NetDB_Test.user_1.id, NetDB_IPaddressSpace.get_parent_id (NetDB_Test.net_1_space_1.id), 'network');
        net_lock := NetDB_Test.force_lock (NetDB_Test.user_1.id, NetDB_Test.name_net_2.network, 'network');

        NetDB_IPaddressSpace.move (NetDB_Test.user_1.id, v_lock, null, NetDB_Test.net_1_space_1.prefix, null, NetDB_Test.name_net_2.fullname, net_lock);

        if (NetDB_IPaddressSpace.get_parent_id (NetDB_Test.net_1_space_1.id) = NetDB_Test.network_2.id) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

exception
    when unrecoverable_error then
        dbms_output.put_line ('ERROR: All remaining tests aborted. An unrecoverable error occured: ' || err_msg);
    when others then
        dbms_output.put_line ('ERROR: Unhandled exception "' || sqlcode || '" with message "' || sqlerrm || '"');
        dbms_output.put_line (err_msg);
end;
/

rollback to start_addressspace_tests;
