set serveroutput on size 100000;
savepoint start_interface_tests;

declare
    test_number int          := 1;
    err_msg     varchar(256) := null;
    v_lock      timestamp    := null;
    v_count     int          := 0;

    test_cursor NetDB.ref_cursor;
    result      result_row := result_row (null, null, null, null, null, null, null);

    -- Values for insert and update.
    test_hw_addr   Interface.hw_address%TYPE := 'hwaddress';
    test_hw_addr_2 Interface.hw_address%TYPE := 'hwaddress_other';
    test_comment   Interface.comments%TYPE   := 'test comment';
    test_type      Interface.type%TYPE;
    test_intfc_id  int;

    verified_id    int;

    unrecoverable_error EXCEPTION;
begin
    select type into test_type from InterfaceType where rownum = 1;

    NetDB_Test.test_info ('Interface', test_number);

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Check the node column in the interface table.';

        v_count := 0;
        for r in (select * from Interface where node != NetDB_Interface.get_parent_id (id))
        loop
            v_count := v_count + 1;

            if (v_count > 1) then
                dbms_output.put (rpad ('.', NetDB_Test.test_text_length, ' '));
            end if;

            dbms_output.put_line ('FAIL: id ' || rpad (r.id, 10, ' ') || ' expected "' || r.node || '"');
            dbms_output.put      (rpad ('.', NetDB_Test.test_text_length + 25, ' '));
            dbms_output.put_line ('got "' || NetDB_Interface.get_parent_id (r.id) || '"');
        end loop;

        if (v_count = 0) then
            NetDB_Test.pass ();
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    -- Usually there would be tests for a user with insufficient access here. However,
    -- there are no direct access checks made with Interfaces. They are indirectly
    -- done through node locking. If the user has a lock, they clearly have access to
    -- Nodes. If they have access to Nodes, they must have access to the children
    -- (Interfaces), which cannot exist outside of a node.

    -- non-test
    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'Node');

    test_cursor := NetDB_Node.master ('lock', NetDB_Test.user_1.id, null, NetDB_Test.node_3.id);
    result      := NetDB_Test.get_generic_row (test_cursor);
    v_lock      := result.value_timestamp;
    NetDB_Test.empty_cursor (test_cursor);

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert an interface.';

        test_cursor   := NetDB_Interface.master ('insert', NetDB_Test.user_1.id, v_lock, null, test_hw_addr, null, test_type, NetDB_Test.node_3.id, null, null, test_comment);
        result        := NetDB_Test.get_generic_row (test_cursor);
        test_intfc_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_Interface.verify_or_die (test_intfc_id, test_hw_addr);

        if (test_intfc_id = verified_id) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an interface.';

        test_cursor := NetDB_Interface.master ('update', NetDB_Test.user_1.id, v_lock, test_intfc_id, null, test_hw_addr, test_type, NetDB_Test.node_3.id);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete an interface.';

        test_cursor := NetDB_Interface.master ('delete', NetDB_Test.user_1.id, v_lock, test_intfc_id);
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_Interface.verify_or_die (test_intfc_id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert an interface with an existing hw addr.';

        test_cursor := NetDB_Interface.master ('insert', NetDB_Test.user_1.id, v_lock, null, test_hw_addr, null, test_type, NetDB_Test.node_3.id, null, null, test_comment);
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Interface.master ('insert', NetDB_Test.user_1.id, v_lock, null, test_hw_addr, null, test_type, NetDB_Test.node_3.id, null, null, test_comment);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_already_exists then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an interface with an existing hw addr.';

        test_cursor := NetDB_Interface.master ('insert', NetDB_Test.user_1.id, v_lock, null, test_hw_addr_2, null, test_type, NetDB_Test.node_3.id, null, null, test_comment);
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Interface.master ('update', NetDB_Test.user_1.id, v_lock, null, test_hw_addr, test_hw_addr_2, test_type, NetDB_Test.node_3.id, null, 0, test_comment);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_already_exists then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;
exception
    when unrecoverable_error then
        dbms_output.put_line ('ERROR: All remaining tests aborted. An unrecoverable error occured: ' || err_msg);
    when others then
        dbms_output.put_line ('ERROR: Unhandled exception "' || sqlcode || '" with message "' || sqlerrm || '"');
        dbms_output.put_line (err_msg);
end;
/

rollback to start_interface_tests;
