set serveroutput on size 100000;
savepoint start_location_tests;

declare
    test_number int          := 1;
    err_msg     varchar(256) := null;
    v_count     int          := 0;

    test_cursor NetDB.ref_cursor;
    result      result_row := result_row (null, null, null, null, null, null, null);

    -- values for insert and update.
    test_loc     varchar(256) := 'Test "location"';
    quoted_loc   varchar(256) := translate (test_loc, '"', '''');
    test_loc_2   varchar(256) := 'Test loca"tion';
    quoted_loc_2 varchar(256) := translate (test_loc_2, '"', '''');
    test_site    varchar(256) := 'site';
    test_code    varchar(256) := 'code';
    test_id      int := null;

    existing_loc Location%ROWTYPE;

    unrecoverable_error EXCEPTION;
begin
    select count (*) into v_count from Location
     where name in (test_loc, quoted_loc, test_loc_2, quoted_loc_2);

    if (v_count > 0) then
        err_msg := 'The test location "' || test_loc || '" exists. If it is not a valid ' ||
                   'location, remove the entry and run the tests again. If it is valid, the ' ||
                   'tests must be modified.';
        raise unrecoverable_error;
    end if;

    select * into existing_loc from Location where rownum = 1;
    NetDB_Test.revoke_access (NetDB_Test.user_1.id, 'Location');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a location without permission';

        test_cursor := NetDB_Location.master ('insert', NetDB_Test.user_1.id, null, test_loc, null, test_site, test_code);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a location without permission.';

        test_cursor := NetDB_Location.master ('update', NetDB_Test.user_1.id, existing_loc.id, null, test_loc_2, null, null);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete a location without permission.';

        test_cursor := NetDB_Location.master ('delete', NetDB_Test.user_1.id, existing_loc.id);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'Location');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a location';

        test_cursor := NetDB_Location.master ('insert', NetDB_Test.user_1.id, null, test_loc, null, test_site, test_code);
        result  := NetDB_Test.get_generic_row (test_cursor);
        test_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        if (NetDB_Location.handle (test_id) = quoted_loc) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a location.';

        test_cursor := NetDB_Location.master ('update', NetDB_Test.user_1.id, test_id, null, test_loc_2, test_site, test_code);
        NetDB_Test.empty_cursor (test_cursor);

        if (NetDB_Location.handle (test_id) = quoted_loc_2) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete a location.';

        test_cursor := NetDB_Location.master ('delete', NetDB_Test.user_1.id, test_id);
        NetDB_Test.empty_cursor (test_cursor);

        test_id := NetDB_Location.verify_or_die (test_id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;
exception
    when unrecoverable_error then
        dbms_output.put_line ('ERROR: All remaining tests aborted. An unrecoverable error occured: ' || err_msg);
    when others then
        dbms_output.put_line ('ERROR: Unhandled exception "' || sqlcode || '" with message "' || sqlerrm || '"');
        dbms_output.put_line (err_msg);
end;
/

rollback to start_location_tests;
