set serveroutput on size 100000;
savepoint start_model_tests;

declare
    test_number int          := 1;
    err_msg     varchar(256) := null;
    v_count     int          := 0;

    test_cursor NetDB.ref_cursor;
    result      result_row := result_row (null, null, null, null, null, null, null);

    -- values for insert and update.
    test_make    Make.make%TYPE;
    test_model   Model.modelname%TYPE := 'unlikely model name - test';
    test_model_2 Model.modelname%TYPE := 'test - unlikely model name';
    test_id      int := null;
    test_id_2    int := null;
    verified_id  int := null;

    existing_model Model%ROWTYPE;

    unrecoverable_error EXCEPTION;
begin
    select count (*) into v_count from Model where modelname = test_model;

    if (v_count > 0) then
        err_msg := 'The test model "' || test_model || '" exists. If it is not a valid ' ||
                   'model, remove the entry and run the tests again. If it is valid, the ' ||
                   'tests must be modified.';
        raise unrecoverable_error;
    end if;

    select *    into existing_model from Model where rownum = 1;
    select make into test_make      from Make  where rownum = 1;

    NetDB_Test.revoke_access (NetDB_Test.user_1.id, 'Model');
    NetDB_Test.revoke_access (NetDB_Test.user_1.id, 'Node');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a model without permission';

        test_cursor := NetDB_Model.master ('insert', NetDB_Test.user_1.id, null, test_make, test_model);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a model without permission.';

        test_cursor := NetDB_Model.master ('update', NetDB_Test.user_1.id, null, existing_model.make, existing_model.modelname, existing_model.make, test_model_2);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete a model without permission.';

        test_cursor := NetDB_Model.master ('delete', NetDB_Test.user_1.id, existing_model.id);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'Node');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a model with node access.';

        test_cursor := NetDB_Model.master ('insert', NetDB_Test.user_1.id, null, test_make, test_model);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a model and type with node access.';

        test_cursor := NetDB_Model.master ('insert', NetDB_Test.user_1.id, null, test_make, test_model, null, null, 'not null');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a model with node access.';

        test_cursor := NetDB_Model.master ('update', NetDB_Test.user_1.id, null, existing_model.make, existing_model.modelname, existing_model.make, test_model_2);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete a model with node access.';

        test_cursor := NetDB_Model.master ('delete', NetDB_Test.user_1.id, existing_model.id);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    rollback to start_model_tests;
    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'Model');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a model';

        test_cursor := NetDB_Model.master ('insert', NetDB_Test.user_1.id, null, test_make, test_model);
        result  := NetDB_Test.get_generic_row (test_cursor);
        test_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_Model.verify_or_die (test_id, test_make, test_model);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a model.';

        test_cursor := NetDB_Model.master ('update', NetDB_Test.user_1.id, null, test_make, test_model, test_make, test_model_2);
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_Model.verify_or_die (test_id, test_make, test_model_2);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete a model.';

        test_cursor := NetDB_Model.master ('delete', NetDB_Test.user_1.id, test_id);
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_Model.verify_or_die (test_id);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    rollback to start_model_tests;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a model to an existing model.';

        test_cursor := NetDB_Model.master ('insert', NetDB_Test.user_0.id, null, test_make, test_model);
        result  := NetDB_Test.get_generic_row (test_cursor);
        test_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Model.master ('insert', NetDB_Test.user_0.id, null, test_make, test_model_2);
        result    := NetDB_Test.get_generic_row (test_cursor);
        test_id_2 := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Model.master ('update', NetDB_Test.user_0.id, null, test_make, test_model, test_make, test_model_2);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_already_exists then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    rollback to start_model_tests;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Replace a model';

        test_cursor := NetDB_Model.master ('insert', NetDB_Test.user_0.id, null, test_make, test_model);
        result  := NetDB_Test.get_generic_row (test_cursor);
        test_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Model.master ('insert', NetDB_Test.user_0.id, null, test_make, test_model_2);
        result    := NetDB_Test.get_generic_row (test_cursor);
        test_id_2 := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Model.master ('replace', NetDB_Test.user_0.id, null, test_make, test_model, test_make, test_model_2);

        begin
            verified_id := NetDB_Model.verify_or_die (test_id_2, test_make, test_model_2);
        exception
            when others then
                NetDB_Test.failed (err_msg || '1');
                verified_id := null;
        end;

        if (verified_id is not null) then
            verified_id := NetDB_Model.verify_or_die (test_id, test_make, test_model);

            NetDB_Test.failed (err_msg || '3');
        end if;
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;
exception
    when unrecoverable_error then
        dbms_output.put_line ('ERROR: All remaining tests aborted. An unrecoverable error occured: ' || err_msg);
    when others then
        dbms_output.put_line ('ERROR: Unhandled exception "' || sqlcode || '" with message "' || sqlerrm || '"');
        dbms_output.put_line (err_msg);
end;
/

rollback to start_model_tests;
