set serveroutput on size 100000;
savepoint start_name_tests;

declare
    test_number   int          := 1;
    err_msg       varchar(256) := null;

    -- Name for insert and update.
    ip_name       Name.name%TYPE     := 'valid-test-ip-name';
    test_name     Name.name%TYPE     := 'valid-name-that-should-not-exist';
    alias_name    Name.name%TYPE     := 'alias-to-the-test-name';
    test_fullname Name.fullname%TYPE := NetDB_Util.fullname (test_name, NetDB_Test.domain_1_name);
    v_lock        timestamp          := null;
    v_count       int                := 0;
    com_domain    Domain.domain%TYPE := 'COM';

    ip_reserved  timestamp;
    v_reserved   timestamp;
    v_name_id    int;
    node_name_id int;
    ip_name_id   int;
    verified_id  int;

    a_comment     Node.comments%TYPE        := 'a comment';
    room          Node.room%TYPE            := 'the room';
    valid_addr    IPaddress.ipaddress%TYPE  := '191.64.132.254';
    hw_addr       Interface.hw_address%TYPE := 'hwaddress';
    intfc_type    Interface.type%TYPE;
    node_handle   Name.name%TYPE;

    ip_id         int;
    node_id       int;
    domain_id     int;
    mx_name_id    int;
    test_intfc_id int;

    v_os      OS.os%TYPE;
    v_loc     Location.Name%TYPE;
    v_state   State.state%TYPE;
    v_make    Model.make%TYPE;
    v_model   Model.modelname%TYPE;
    v_node_id int;

    v_string  varchar(256) := null;
    v_int     int          := 0;

    test_cursor NetDB.ref_cursor;
    result      result_row := result_row (null, null, null, null, null, null, null);

    unrecoverable_error EXCEPTION;
begin
    select count (*) into v_count from Name where lower (name) = test_name;

    if (v_count > 0) then
        err_msg := 'The test name "' || test_name || '" exists. If it is not a valid name, remove the entry and run the tests again. If it is valid, the tests must be modified.';

        raise unrecoverable_error;
    end if;

    if (NetDB_Domain.is_limited (NetDB_Test.domain_1.id)) then
        dbms_output.put_line ('*** WARNING ***: The domain "' || NetDB_Test.domain_1.id || '" had its limited flag set. Reverting it back to the expected state and continuing.');

        update Domain set limited = 0 where id = NetDB_Test.domain_1.id;
    end if;

    select name            into v_loc           from Location      where rownum = 1;
    select state           into v_state         from State         where rownum = 1;
    select make, modelname into v_make, v_model from Model         where rownum = 1;
    select type            into intfc_type      from InterfaceType where rownum = 1;

    NetDB_Test.grant_access (NetDB_Test.user_2.id, 'Domain');
    NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);
    NetDB_Test.add_to_group (NetDB_Test.user_2.id, NetDB_Test.group_1.id);
    NetDB_Test.remove_from_group (NetDB_Test.user_2.id, NetDB_Test.group_2.id);

    begin -- non-test
        err_msg := 'Locking the node "' || NetDB_Test.node_4_name || '".';

        test_cursor := NetDB_Node.master ('lock', NetDB_Test.user_1.id, null, NetDB_Test.node_4.id);
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);
    exception
        when others then
            raise unrecoverable_error;
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Create an MX.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'MX');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, v_name_id, null, 'MX', NetDB_Test.name_node_4.id, null, v_lock, 10);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve an A name for an existing MX without permission to any of the received_by records.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_2.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    NetDB_Test.add_to_group (NetDB_Test.user_2.id, NetDB_Test.group_2.id);

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve an A name for an existing MX.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_2.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    -- Test clean up
    rollback to start_name_tests;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a name with invalid characters.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, 'name with bad charaters', NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.bad_name then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a name with the parent of ".".';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Domain.handle (1), 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.cannot_use_domain then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a name twice.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_0.id, 'a2', 'COM', 'Domain');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_0.id, 'a2', 'COM', 'Domain');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Cancel a reservation with the wrong name type.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.cancel (NetDB_Test.user_1.id, v_name_id, null, 'Domain');

        -- The name should still exists, and the verify should succeed.
        v_int := NetDB_Name.verify_or_die (v_name_id, null, null, true);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Cancel a reservation.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.cancel (NetDB_Test.user_1.id, v_name_id, null, 'Node');

        v_int := NetDB_Name.verify_or_die (v_name_id, null, null, true);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a valid name with a valid domain.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Accept a name.';

        test_cursor := NetDB_Node.master ('insert', NetDB_Test.user_1.id, null,                   null,
                                                    test_fullname,        NetDB_Test.dept_1_name, v_loc,
                                                   'the room',            v_state,                v_make,
                                                    v_model,             'comment' );
        result    := NetDB_Test.get_generic_row (test_cursor);
        v_node_id := result.value_string;
        result    := NetDB_Test.get_generic_row (test_cursor);
        v_lock    := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, v_name_id, null, 'node', v_node_id, null, v_lock);

        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, v_node_id, 'node');

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        update Name set reserved_by = null where name = test_name;

        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a name identical to an existing name.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.name_collision then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve an MX for an existing name.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'MX');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        update Name set reserved_by = null where name = test_name;

        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve an MX for an existing name again.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'MX');
        NetDB_Test.empty_cursor (test_cursor);

        select count (*) into v_count from Name where name = test_name;
        if (v_count = 2) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve an existing name as an alias.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_0.id, NetDB_Test.name_node_1.name, NetDB_Test.domain_1.domain, 'Alias');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.name_collision then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        delete from Name where name = test_name;

        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a new name as an alias.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Alias');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        update Name set reserved_by = null where name = test_name;

        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a valid name that exists as an alias.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.name_collision then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        update Name set reserved    = add_months (systimestamp, -1),
                        reserved_by = NetDB_Test.user_1.id
         where name = test_name;

        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update an expired name.';
        -- Technically this isn't an update, since the old reservation
        -- is nuked as part of the reservation process.

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        select count (*) into v_count from Name where lower (name) = test_name;

        if (v_count = 1) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a reserved/restricted name.';

        test_cursor   := NetDB_Name.reserve (NetDB_Test.user_1.id, 'fuck', NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.bad_name then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        delete from Name where name = test_name;

        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a name with a domain that the user does not have access to.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_2.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.cannot_use_domain then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Name an IP address.';

        rollback to start_name_tests;

        select name            into v_loc           from location      where rownum = 1;
        select state           into v_state         from State         where rownum = 1;
        select make, modelname into v_make, v_model from Model         where rownum = 1;
        select os              into v_os            from OS            where rownum = 1;
        select type            into intfc_type      from InterfaceType where rownum = 1;

        NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result       := NetDB_Test.get_generic_row (test_cursor);
        node_name_id := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        v_reserved   := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        select fullname into node_handle from Name where name = test_name and domain = NetDB_Test.domain_1.id;

        test_cursor := NetDB_Node.master ('insert', NetDB_Test.user_1.id, null, null, node_handle, NetDB_Test.dept_1_name, v_loc, room, v_state, v_make, v_model, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        node_id     := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor   := NetDB_Interface.master ('insert', NetDB_Test.user_1.id, v_lock, null, hw_addr, null, intfc_type, node_id);
        result        := NetDB_Test.get_generic_row (test_cursor);
        test_intfc_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_IPaddress.reserve (NetDB_Test.user_1.id, valid_addr, 1);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_IPaddress.accept (NetDB_Test.user_1.id, valid_addr, 1, 0, test_intfc_id, v_lock, 'interface');

        NetDB_Name.accept (NetDB_Test.user_1.id, node_name_id, null, 'Node', node_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, node_id, 'Node');
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_2.id, null, node_id, 'Node');
        NetDB_Node.admin ('add', NetDB_Test.user_1.id, v_lock, node_id, null, null, null, NetDB_Test.user_1.id);
        NetDB_Node.OS ('add', NetDB_Test.user_1.id, v_lock, node_id, null, v_os, null);

        test_cursor := NetDB_Node.master ('verify', NetDB_Test.user_1.id, null, node_id);
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, ip_name, NetDB_Test.domain_1.domain, 'Address');
        result       := NetDB_Test.get_generic_row (test_cursor);
        ip_name_id   := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        ip_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, ip_name_id, null, 'address', null, valid_addr, v_lock);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        update Domain set limited = 1 where id = NetDB_Test.domain_1.id;

        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a name, found in the LimitedName table, in a limited domain.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, 'www', NetDB_Test.domain_1.domain, 'Domain');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a name, not found in the LimitedName table, in a limited domain.';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, 'different-name', NetDB_Test.domain_1.domain, 'Domain');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.bad_name then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Verify that no exceptions are thrown from cancel (...).';

        NetDB_Name.cancel (1, -1, null);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Cancel an accepted name.';

        NetDB_Name.cancel (NetDB_Test.user_1.id,  1, null);

        if (NetDB_Name.verify_or_die (1) = 1) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Create an mx with the same name as the node (using the name parameter for acceptance).';

        rollback to start_name_tests;

        select name            into v_loc           from location where rownum = 1;
        select state           into v_state         from State    where rownum = 1;
        select make, modelname into v_make, v_model from Model    where rownum = 1;
        select os              into v_os            from OS       where rownum = 1;

        NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result       := NetDB_Test.get_generic_row (test_cursor);
        node_name_id := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        v_reserved   := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        select fullname into node_handle from Name where name = test_name and domain = NetDB_Test.domain_1.id;

        test_cursor := NetDB_Node.master ('insert', NetDB_Test.user_1.id, null, null, node_handle, NetDB_Test.dept_1_name, v_loc, room, v_state, v_make, v_model, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        node_id     := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, node_name_id, null, 'Node', node_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, node_id, 'Node');
        NetDB_Node.admin ('add', NetDB_Test.user_1.id, v_lock, node_id, null, null, null, NetDB_Test.user_1.id);
        NetDB_Node.OS ('add', NetDB_Test.user_1.id, v_lock, node_id, null, v_os, null);

        test_cursor := NetDB_Node.master ('verify', NetDB_Test.user_1.id, null, node_id);
        NetDB_Test.empty_cursor (test_cursor);

        -- Create the MX.
        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'MX');
        result      := NetDB_Test.get_generic_row (test_cursor);
        mx_name_id  := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, mx_name_id, null, 'MX', null, node_handle, v_lock, 10);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Create an alias for a node with an mx of the same name.';

        rollback to start_name_tests;

        select name            into v_loc           from location      where rownum = 1;
        select state           into v_state         from State         where rownum = 1;
        select make, modelname into v_make, v_model from Model         where rownum = 1;
        select os              into v_os            from OS            where rownum = 1;
        select type            into intfc_type      from InterfaceType where rownum = 1;

        NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result       := NetDB_Test.get_generic_row (test_cursor);
        node_name_id := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        v_reserved   := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        select fullname into node_handle from Name where name = test_name and domain = NetDB_Test.domain_1.id;

        test_cursor := NetDB_Node.master ('insert', NetDB_Test.user_1.id, null, null, node_handle, NetDB_Test.dept_1_name, v_loc, room, v_state, v_make, v_model, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        node_id     := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor   := NetDB_Interface.master ('insert', NetDB_Test.user_1.id, v_lock, null, hw_addr, null, intfc_type, node_id);
        result        := NetDB_Test.get_generic_row (test_cursor);
        test_intfc_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_IPaddress.reserve (NetDB_Test.user_1.id, valid_addr, 1);
        result      := NetDB_Test.get_generic_row (test_cursor);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_IPaddress.accept (NetDB_Test.user_1.id, valid_addr, 1, 0, test_intfc_id, v_lock, 'interface');

        NetDB_Name.accept (NetDB_Test.user_1.id, node_name_id, null, 'Node', node_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, node_id, 'Node');
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_2.id, null, node_id, 'Node');
        NetDB_Node.admin ('add', NetDB_Test.user_1.id, v_lock, node_id, null, null, null, NetDB_Test.user_1.id);
        NetDB_Node.OS ('add', NetDB_Test.user_1.id, v_lock, node_id, null, v_os, null);

        test_cursor := NetDB_Node.master ('verify', NetDB_Test.user_1.id, null, node_id);
        NetDB_Test.empty_cursor (test_cursor);

        -- Create the MX.
        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'MX');
        result      := NetDB_Test.get_generic_row (test_cursor);
        mx_name_id  := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, mx_name_id, null, 'MX', node_name_id, null, v_lock, 10);

        -- Create the alias.
        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, alias_name, NetDB_Test.domain_1.domain, 'Alias');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, v_name_id, null, 'Alias', node_name_id, null, v_lock);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Create an alias of the same name, but different case, as a node.';

        rollback to start_name_tests;

        select name            into v_loc           from location      where rownum = 1;
        select state           into v_state         from State         where rownum = 1;
        select make, modelname into v_make, v_model from Model         where rownum = 1;
        select os              into v_os            from OS            where rownum = 1;
        select type            into intfc_type      from InterfaceType where rownum = 1;

        NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result       := NetDB_Test.get_generic_row (test_cursor);
        node_name_id := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        v_reserved   := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        select fullname into node_handle from Name where name = test_name and domain = NetDB_Test.domain_1.id;

        test_cursor := NetDB_Node.master ('insert', NetDB_Test.user_1.id, null, null, node_handle, NetDB_Test.dept_1_name, v_loc, room, v_state, v_make, v_model, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        node_id     := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor   := NetDB_Interface.master ('insert', NetDB_Test.user_1.id, v_lock, null, hw_addr, null, intfc_type, node_id);
        result        := NetDB_Test.get_generic_row (test_cursor);
        test_intfc_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_IPaddress.reserve (NetDB_Test.user_1.id, valid_addr, 1);
        result      := NetDB_Test.get_generic_row (test_cursor);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_IPaddress.accept (NetDB_Test.user_1.id, valid_addr, 1, 0, test_intfc_id, v_lock, 'interface');

        NetDB_Name.accept (NetDB_Test.user_1.id, node_name_id, null, 'Node', node_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, node_id, 'Node');
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_2.id, null, node_id, 'Node');
        NetDB_Node.admin ('add', NetDB_Test.user_1.id, v_lock, node_id, null, null, null, NetDB_Test.user_1.id);
        NetDB_Node.OS ('add', NetDB_Test.user_1.id, v_lock, node_id, null, v_os, null);

        test_cursor := NetDB_Node.master ('verify', NetDB_Test.user_1.id, null, node_id);
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, upper (test_name), NetDB_Test.domain_1.domain, 'Alias');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.name_collision then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Create an alias.';

        rollback to start_name_tests;

        select name            into v_loc           from location      where rownum = 1;
        select state           into v_state         from State         where rownum = 1;
        select make, modelname into v_make, v_model from Model         where rownum = 1;
        select os              into v_os            from OS            where rownum = 1;
        select type            into intfc_type      from InterfaceType where rownum = 1;

        NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result       := NetDB_Test.get_generic_row (test_cursor);
        node_name_id := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        v_reserved   := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        select fullname into node_handle from Name where name = test_name and domain = NetDB_Test.domain_1.id;

        test_cursor := NetDB_Node.master ('insert', NetDB_Test.user_1.id, null, null, node_handle, NetDB_Test.dept_1_name, v_loc, room, v_state, v_make, v_model, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        node_id     := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor   := NetDB_Interface.master ('insert', NetDB_Test.user_1.id, v_lock, null, hw_addr, null, intfc_type, node_id);
        result        := NetDB_Test.get_generic_row (test_cursor);
        test_intfc_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_IPaddress.reserve (NetDB_Test.user_1.id, valid_addr, 1);
        result      := NetDB_Test.get_generic_row (test_cursor);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_IPaddress.accept (NetDB_Test.user_1.id, valid_addr, 1, 0, test_intfc_id, v_lock, 'interface');

        NetDB_Name.accept (NetDB_Test.user_1.id, node_name_id, null, 'Node', node_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, node_id, 'Node');
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_2.id, null, node_id, 'Node');
        NetDB_Node.admin ('add', NetDB_Test.user_1.id, v_lock, node_id, null, null, null, NetDB_Test.user_1.id);
        NetDB_Node.OS ('add', NetDB_Test.user_1.id, v_lock, node_id, null, v_os, null);

        test_cursor := NetDB_Node.master ('verify', NetDB_Test.user_1.id, null, node_id);
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, alias_name, NetDB_Test.domain_1.domain, 'Alias');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, v_name_id, null, 'Alias', node_name_id, null, v_lock);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete an alias.';

        rollback to start_name_tests;

        select name            into v_loc           from location      where rownum = 1;
        select state           into v_state         from State         where rownum = 1;
        select make, modelname into v_make, v_model from Model         where rownum = 1;
        select os              into v_os            from OS            where rownum = 1;
        select type            into intfc_type      from InterfaceType where rownum = 1;

        NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result       := NetDB_Test.get_generic_row (test_cursor);
        node_name_id := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        v_reserved   := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        select fullname into node_handle from Name where name = test_name and domain = NetDB_Test.domain_1.id;

        test_cursor := NetDB_Node.master ('insert', NetDB_Test.user_1.id, null, null, node_handle, NetDB_Test.dept_1_name, v_loc, room, v_state, v_make, v_model, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        node_id     := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor   := NetDB_Interface.master ('insert', NetDB_Test.user_1.id, v_lock, null, hw_addr, null, intfc_type, node_id);
        result        := NetDB_Test.get_generic_row (test_cursor);
        test_intfc_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_IPaddress.reserve (NetDB_Test.user_1.id, valid_addr, 1);
        result      := NetDB_Test.get_generic_row (test_cursor);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_IPaddress.accept (NetDB_Test.user_1.id, valid_addr, 1, 0, test_intfc_id, v_lock, 'interface');

        NetDB_Name.accept (NetDB_Test.user_1.id, node_name_id, null, 'Node', node_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, node_id, 'Node');
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_2.id, null, node_id, 'Node');
        NetDB_Node.admin ('add', NetDB_Test.user_1.id, v_lock, node_id, null, null, null, NetDB_Test.user_1.id);
        NetDB_Node.OS ('add', NetDB_Test.user_1.id, v_lock, node_id, null, v_os, null);

        test_cursor := NetDB_Node.master ('verify', NetDB_Test.user_1.id, null, node_id);
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, alias_name, NetDB_Test.domain_1.domain, 'Alias');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, v_name_id, null, 'Alias', node_name_id, null, v_lock);

        NetDB_Name.del (NetDB_Test.user_1.id, v_lock, v_name_id, null, 'alias');

        verified_id := NetDB_Name.verify_or_die (node_name_id, null, 'node');

        verified_id := null;
        begin
            verified_id := NetDB_Name.verify_or_die (v_name_id, null, 'alias');
        exception
            when NetDB.record_not_found then
                NetDB_Test.pass ();
        end;

        if (verified_id is not null) then
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    -- Undo the changes from the test.
    rollback to start_name_tests;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete an alias from a node without access to the domain.';

        rollback to start_name_tests;

        select name            into v_loc           from location      where rownum = 1;
        select state           into v_state         from State         where rownum = 1;
        select make, modelname into v_make, v_model from Model         where rownum = 1;
        select os              into v_os            from OS            where rownum = 1;
        select type            into intfc_type      from InterfaceType where rownum = 1;

        NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);

        test_cursor  := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result       := NetDB_Test.get_generic_row (test_cursor);
        node_name_id := result.value_string;
        result       := NetDB_Test.get_generic_row (test_cursor);
        v_reserved   := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        select fullname into node_handle from Name where name = test_name and domain = NetDB_Test.domain_1.id;

        test_cursor := NetDB_Node.master ('insert', NetDB_Test.user_1.id, null, null, node_handle, NetDB_Test.dept_1_name, v_loc, room, v_state, v_make, v_model, a_comment);
        result      := NetDB_Test.get_generic_row (test_cursor);
        node_id     := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor   := NetDB_Interface.master ('insert', NetDB_Test.user_1.id, v_lock, null, hw_addr, null, intfc_type, node_id);
        result        := NetDB_Test.get_generic_row (test_cursor);
        test_intfc_id := result.value_string;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_IPaddress.reserve (NetDB_Test.user_1.id, valid_addr, 1);
        result      := NetDB_Test.get_generic_row (test_cursor);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_IPaddress.accept (NetDB_Test.user_1.id, valid_addr, 1, 0, test_intfc_id, v_lock, 'interface');

        NetDB_Name.accept (NetDB_Test.user_1.id, node_name_id, null, 'Node', node_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_1.id, null, node_id, 'Node');
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, NetDB_Test.group_2.id, null, node_id, 'Node');
        NetDB_Node.admin ('add', NetDB_Test.user_1.id, v_lock, node_id, null, null, null, NetDB_Test.user_1.id);
        NetDB_Node.OS ('add', NetDB_Test.user_1.id, v_lock, node_id, null, v_os, null);

        test_cursor := NetDB_Node.master ('verify', NetDB_Test.user_1.id, null, node_id);
        NetDB_Test.empty_cursor (test_cursor);

        -- create the alias
        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, alias_name, NetDB_Test.domain_1.domain, 'Alias');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, v_name_id, null, 'Alias', node_name_id, null, v_lock);

        -- Now try and remove it.
        NetDB_Test.remove_from_group (NetDB_Test.user_1.id, NetDB_Test.group_1.id);

        NetDB_Name.del (NetDB_Test.user_1.id, v_lock, v_name_id, null, 'alias');

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.cannot_use_domain then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    -- Undo the changes from the test.
    rollback to start_name_tests;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a node name when the name exists as an MX.';

        NetDB_Test.add_to_group (NetDB_Test.user_1.id, NetDB_Test.group_2.id);

        -- Lock the node.
        test_cursor := NetDB_Node.master ('lock', NetDB_Test.user_1.id, null, NetDB_Test.node_4.id);
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        -- Add an MX to the node.
        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'MX');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, v_name_id, null, 'MX', NetDB_Test.name_node_4.id, null, v_lock, 10);

        -- Add another name to the node.
        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, NetDB_Test.domain_1.domain, 'Node');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, v_name_id, null, 'Node', NetDB_Test.node_4.id, null, v_lock);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    -- Undo the changes from the test.
    rollback to start_name_tests;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Verify that the user can''t assign names in the domain "COM".';

        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, com_domain, 'Node');
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.cannot_use_domain then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Reserve a node name, matching a domain name, with a user that doesn''t have access to the parent domain.';

        -- Create the domain
        test_cursor := NetDB_Name.reserve (NetDB_Test.user_0.id, test_name, com_domain, 'Domain');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Domain.master ('insert', NetDB_Test.user_0.id, null, null, NetDB_Util.fullname (test_name, com_domain), null, null, 0);
        result      := NetDB_Test.get_generic_row (test_cursor);
        domain_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_lock      := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_0.id, v_name_id, null, 'Domain', domain_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_0.id, v_lock, null, NetDB_Test.group_1_name, domain_id, 'Domain');
        NetDB_Owner.add (NetDB_Test.user_0.id, v_lock, null, NetDB_Test.group_1_name, domain_id, 'domain_use');

        test_cursor := NetDB_Domain.master ('verify', NetDB_Test.user_0.id, null, domain_id);
        NetDB_Test.empty_cursor (test_cursor);

        -- Create the node
        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, test_name, com_domain, 'Node');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_Node.master ('insert', NetDB_Test.user_1.id, null,                   null,
                                                    test_fullname,        NetDB_Test.dept_1_name, v_loc,
                                                   'the room',            v_state,                v_make,
                                                    v_model,             'comment' );
        result  := NetDB_Test.get_generic_row (test_cursor);
        node_id := result.value_string;
        result  := NetDB_Test.get_generic_row (test_cursor);
        v_lock  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, v_name_id, null, 'Node', node_id, null, v_lock);
        NetDB_Owner.add (NetDB_Test.user_1.id, v_lock, null, NetDB_Test.group_1_name, node_id, 'Node');
        NetDB_Node.admin ('add', NetDB_Test.user_1.id, v_lock, node_id, null, null, null, NetDB_Test.user_1.id);
        NetDB_Node.OS ('add', NetDB_Test.user_1.id, v_lock, node_id, null, v_os, null);

        test_cursor := NetDB_Node.master ('verify', NetDB_Test.user_1.id, null, node_id);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Rename a node whose name is shared with a domain.';

        -- Full test description:
        -- Rename a node whose name is shared with a domain. The user doing the renaming doesn't
        -- have access to the parent domain of the original name.

        -- Add the new name.
        test_cursor := NetDB_Name.reserve (NetDB_Test.user_1.id, alias_name, NetDB_Test.domain_1.domain, 'Node');
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_name_id   := result.value_string;
        result      := NetDB_Test.get_generic_row (test_cursor);
        v_reserved  := result.value_timestamp;
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Name.accept (NetDB_Test.user_1.id, v_name_id, null, 'Node', node_id, null, v_lock);

        -- Remove the old name
        NetDB_Name.del (NetDB_Test.user_1.id, v_lock, null, NetDB_Node.handle (node_id), 'Node');

        test_cursor := NetDB_Node.master ('verify', NetDB_Test.user_1.id, null, node_id);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    -- Undo the changes from the test.
    rollback to start_name_tests;

    -- fullname (...) tests
    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Generate the fullname of "null".';

        v_string := NetDB_Name.generate_fullname (null);

        if (v_string is null) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Genereate the fullname of "' || NetDB_Test.name_domain_1.name || '".';

        v_string := NetDB_Name.generate_fullname (NetDB_Test.name_domain_1.id);

        if (v_string = NetDB_Test.domain_1.domain) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
            dbms_output.put_line ('expected "' || NetDB_Test.domain_1.domain || '"     got "' || v_string || '"');
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Check the fullname column in the name table.';

        v_count := 0;
        for r in (select * from Name where fullname != NetDB_Name.generate_fullname (id))
        loop
            v_count := v_count + 1;

            if (v_count > 1) then
                dbms_output.put (rpad ('.', NetDB_Test.test_text_length, ' '));
            end if;

            dbms_output.put_line ('FAIL: id ' || rpad (r.id, 10, ' ') || ' expected "' || r.fullname || '"');
            dbms_output.put      (rpad ('.', NetDB_Test.test_text_length + 25, ' '));
            dbms_output.put_line ('got "' || NetDB_Name.fullname (r.id) || '"');
        end loop;

        if (v_count = 0) then
            NetDB_Test.pass ();
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    -- related_node (...) tests
    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Get the related node of "null".';

        v_int := NetDB_Name.related_node (null);

        NetDB_Test.failed (err_msg);
    exception
        when NO_DATA_FOUND then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Get the related node of "' || NetDB_Test.name_node_1.id || '".';

        v_int := NetDB_Name.related_node (NetDB_Test.name_node_1.id);

        if (v_int = NetDB_Test.name_node_1.node) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg || '  ' || v_int);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;
exception
    when unrecoverable_error then
        dbms_output.put_line ('ERROR: All remaining tests aborted. An unrecoverable error occured: ');
        dbms_output.put_line (err_msg);
    when others then
        dbms_output.put_line ('ERROR: Unhandled exception "' || sqlcode || '" with message "' || sqlerrm || '"');
        dbms_output.put_line (err_msg);
end;
/

rollback to start_name_tests;
