set serveroutput on size 100000;
savepoint start_state_tests;

declare
    test_number int          := 1;
    err_msg     varchar(256) := null;
    v_count     int          := 0;

    test_cursor NetDB.ref_cursor;
    result      result_row := result_row (null, null, null, null, null, null, null);

    -- values for insert and update.
    p_desc       State.description%TYPE := 'a description';
    test_state   State.state%TYPE := 'unlikely state name - test';
    test_state_2 State.state%TYPE := 'test - unlikely state name';
    test_id      State.state%TYPE := null;
    verified_id  State.state%TYPE := null;

    access_1 boolean;
    access_2 boolean;

    access_group   AccessGroup%ROWTYPE;
    existing_state State%ROWTYPE;

    unrecoverable_error EXCEPTION;
begin
    select count (*) into v_count from State where state in (test_state, test_state_2);

    if (v_count > 0) then
        err_msg := 'The test state "' || test_state || '" exists. If it is not a valid ' ||
                   'state, remove the entry and run the tests again. If it is valid, the ' ||
                   'tests must be modified.';
        raise unrecoverable_error;
    end if;

    select * into existing_state from State       where rownum = 1;
    select * into access_group   from AccessGroup where rownum = 1;

    NetDB_Test.revoke_access (NetDB_Test.user_1.id, 'State');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a state without permission';

        test_cursor := NetDB_State.master ('insert', NetDB_Test.user_1.id, test_state, null, 0, 0, 0, p_desc);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a state without permission.';

        test_cursor := NetDB_State.master ('update', NetDB_Test.user_1.id, existing_state.state, test_state_2, 0, 0, 0, p_desc);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete a state without permission.';

        test_cursor := NetDB_State.master ('delete', NetDB_Test.user_1.id, existing_state.state);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.no_access_to_record_type then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    NetDB_Test.grant_access (NetDB_Test.user_1.id, 'State');

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Insert a state';

        test_cursor := NetDB_State.master ('insert', NetDB_Test.user_1.id, test_state, null, 0, 0, 0, p_desc);
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_State.verify_or_die (test_state);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Update a state.';

        test_cursor := NetDB_State.master ('update', NetDB_Test.user_1.id, test_state, test_state_2, 0, 0, 0, p_desc);
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_State.verify_or_die (test_state_2);

        NetDB_Test.pass ();
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Delete a state.';

        test_cursor := NetDB_State.master ('delete', NetDB_Test.user_1.id, test_state_2);
        NetDB_Test.empty_cursor (test_cursor);

        verified_id := NetDB_State.verify_or_die (test_state_2);

        NetDB_Test.failed (err_msg);
    exception
        when NetDB.record_not_found then
            NetDB_Test.pass ();
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

    begin
        NetDB_Test.output_and_increment_test (test_number);
        err_msg := 'Check access to a state.';

        test_cursor := NetDB_State.master ('insert', NetDB_Test.user_1.id, test_state, null, 0, 0, 0, p_desc);
        NetDB_Test.empty_cursor (test_cursor);

        test_cursor := NetDB_State.master ('insert', NetDB_Test.user_1.id, test_state_2, null, 0, 0, 0, p_desc, access_group.label);
        NetDB_Test.empty_cursor (test_cursor);

        NetDB_Test.grant_access (NetDB_Test.user_1.id, access_group.label);
        NetDB_Test.revoke_access (NetDB_Test.user_2.id, access_group.label);

        NetDB_State.has_access_or_die (NetDB_Test.user_1.id, test_state);
        NetDB_State.has_access_or_die (NetDB_Test.user_2.id, test_state);

        access_1 := NetDB_State.check_access (NetDB_Test.user_1.id, test_state_2);
        access_2 := NetDB_State.check_access (NetDB_Test.user_2.id, test_state_2);

        if (access_1 and not access_2) then
            NetDB_Test.pass ();
        else
            NetDB_Test.failed (err_msg);
        end if;
    exception
        when others then
            NetDB_Test.failed (err_msg, sqlerrm);
    end;

exception
    when unrecoverable_error then
        dbms_output.put_line ('ERROR: All remaining tests aborted. An unrecoverable error occured: ' || err_msg);
    when others then
        dbms_output.put_line ('ERROR: Unhandled exception "' || sqlcode || '" with message "' || sqlerrm || '"');
        dbms_output.put_line (err_msg);
end;
/

rollback to start_state_tests;
