#######################################################################
#
# DNS_config.pm - DNS configuration data procedures
#
# Usage:
#
#  load_config(file)                ..load the config file
#
#  @nets = nets()                   ..list of nets
#  @delegated = delegated(net)      ..delegated subnets of net
#  @domains = domains()             ..list of domains
#  @delegated = delegated(domain)   ..delegated subdomains of domain
#
#  hacks(domain|net)                ..does domain have hacks
#  iserve(domain|net,server)        ..does server serve domain or net
#  imaster(domain|net,server)       ..is server master for domain or net
#
#  $server  = master(domain|net)    ..master NS for domain or net
#  $ipaddr  = ip(server)            ..ip address of server
#  @servers = servers(domain|net)   ..all NSs for domain or net
#  @servers = official(domain|net)  ..official NSs for domain or net
#  @servers = stealth(domain|net)   ..stealth NSs for domain or net
#
#  $zone = zone(domain|net)         ..zone name for domain or net
#  $ttl  = ttl(domain|net)          ..default ttl for domain or net
#  @soa  = soa(domain|net)          ..SOA info for domain or net
#                                        excluding the serial number
#
#----------------------------------------------------------------------
# Copyright (c) 1999 Board of Trustees, Leland Stanford Jr. University
#######################################################################

#######################################################################
#         Private bits to deal with DNS_config file entries
#######################################################################

# DNS configuration Objects and data structures
#
#   Entry->(name,type,comments)
#     type == {host|domain|net}
#
#   Zone->(zone,zonefile,ttl,hacks,master) isa(Entry)
#       ->(@delegated,@soa,@official,@stealth,@servers)
#
#   Host->(ip) isa(Entry)

package DNS_config::Entry;       # DNS configuration file entry

@scalars = qw( name type comments ip zone zonefile ttl hacks master );
@arrays  = qw( soa official stealth servers delegated );

sub new {                           # Name, Type, Comments
    my $class = shift;
    my $self = {};
    bless $self, $class;
    $self->{name} = shift;
    $self->{type} = shift;
    $self->{comments} = shift;
    return $self;
}

sub AUTOLOAD {                      # Get attributes by name
    my $self = shift;

    $AUTOLOAD =~ s/^.*:://;

    if (grep(/$AUTOLOAD/,@scalars)) {
	return exists $self->{$AUTOLOAD} ? $self->{$AUTOLOAD} : undef;

    } elsif (grep(/$AUTOLOAD/,@arrays)) {
	return exists $self->{$AUTOLOAD} ? @{$self->{$AUTOLOAD}} : ();

    } else {
	return ref($self) . " doesn't have an attribute '" . $AUTOLOAD . "'.";
    }
}


package DNS_config::Zone;        # Zone entry

@ISA = qw(DNS_config::Entry);

sub new {                           # Name, Type, Comments
    my $class = shift;
    my $self = $class->SUPER::new(@_);

    $self->{zone} = _zone_name($self->{name});
    $self->{zonefile} = _zonefile_name($self->{name});

    while(<DNS_config::CONF>) {
        $lineno++; chop; s/#.*//; s/^\s+//; s/\s+$//;
        next if /^$/;
        last if (/^\}/);

        if     (/^ttl:\s*(.*)/)       {  $self->{ttl} = $1;      }
        elsif  (/^hacks:\s*(.*)/)     {  $self->{hacks} = $1;    }
        elsif  (/^master:\s*(.*)/)    {  $self->{master} = $1;   }
        elsif  (/^zonefile:\s*(.*)/)  {  $self->{zonefile} = $1; }
        elsif  (/^delegated:\s*(.*)/) {  push (@{$self->{delegated}} ,$1); }
        elsif  (/^soa:\s*(.*)/)       {  push (@{$self->{soa}},split(/\s*,\s*/,$1)); }
        elsif  (/^official:\s*(.*)/)  {  push (@{$self->{official}},split(/\s*,\s*/,$1)); }
        elsif  (/^stealth:\s*(.*)/)   {  push (@{$self->{stealth}},split(/\s*,\s*/,$1));  }
        else   { die "syntax error at line $lineno of $config_file\n"; }
    }
    push (@{$self->{servers}},@{$self->{official}});
    push (@{$self->{servers}},@{$self->{stealth}});
    push (@{$self->{servers}},$self->{master})
      if (defined($self->{master}) && $self->{master} ne "ALL" &&
	  !grep(/$self->{master}/,@{$self->{servers}}));

    return $self;
}

sub _zone_name {
    my $n = shift;
    return $n unless $n =~ /^[0-9\.]+$/;
    return _inaddr_arpa($n);
}

sub _zonefile_name {
    my $n = shift;
    return lc($n) unless $n =~ /^[0-9\.]+$/;
    return $n . "inaddr";
}

sub _inaddr_arpa {
    my($net) = @_;
    return join(".",reverse(split(/\./,$net))) . ".IN-ADDR.ARPA";
}


package DNS_config::Host;        # Host entry

@ISA = qw(DNS_config::Entry);

sub new {                           # Name, Comments
    my $class = shift;
    my $self = $class->SUPER::new(shift,'host',shift);

    while(<DNS_config::CONF>) {
        $lineno++; chop; s/#.*//; s/^\s+//; s/\s+$//;
        next if /^$/;
        last if (/^\}/);

        if     (/^ip:\s*(.*)/) {  $self->{ip} = $1;    }
        else   { die "syntax error at line $lineno of $config_file\n"; }
    }
    return $self;
}


#######################################################################
#                        The Public Interface
#######################################################################

package DNS_config;

require Exporter;

@ISA = qw(Exporter);

@EXPORT_OK = qw(load_config nets domains delegated hacks iserve
                imaster master servers official ip stealth zone
                zonefile ttl soa name hostfqdn);

sub load_config {
    local($config_file,$save_comments)=@_;

    # find m4
    @m4 = qw(/usr/bin/m4 /usr/ccs/bin/m4);
    foreach (@m4) { $m4 = "$_"; last if -e $_; }
    die "DNS_config.pm can't find m4" unless -e $m4;

    if ($config_file =~ /\.m4$/ ) { $file = "$m4 $config_file |"; }
    else                          { $file = "< $config_file"      }

    open(CONF,$file) || die "Can't open $config_file: $!\n";
    $lineno = 0; $comments = "";

    while(<CONF>) {
        $line = $_;
        $lineno++; chop; s/#.*//; s/^\s+//; s/\s+$//;
	$comments .= $line if /^$/ and $save_comments;
	next if /^$/;
        last if (/^\}/);

        if (/^host:\s*(\S+)\s*{/) {
	    my $NAME = $1;  my $name = "\L$NAME";
	    $HOSTS{$name} = DNS_config::Host->new($NAME,$comments);
	    $ENTRIES{$name} = $HOSTS{$name};
	    push @entries, $HOSTS{$name};

	} elsif (/^domain:\s*(\S+)\s*{/) {
	    my $NAME = $1;  my $name = "\L$NAME";
	    $DOMAINS{$name} = DNS_config::Zone->new($NAME,'domain',$comments);
	    $ENTRIES{$name} = $DOMAINS{$name};
	    push @entries, $DOMAINS{$name};
	    push @domains, $NAME;

	} elsif (/^net:\s*(\S+)\s*{/) {
	    my $net = $1;  $net .= "." unless $net =~ /\.$/;
	    $NETS{$net} = DNS_config::Zone->new($net,'net',$comments);
	    $ENTRIES{$net} = $NETS{$net};
	    push @entries, $NETS{$net};

	} else {
	    die "syntax error at line $lineno of $config_file\n";
	}

	$comments = "";
    }

    close(CONF);

    if ($lineno == 0) {
	die "DNS_config: problems reading configuration file $config_file\n";
    }
}

sub domains {  return @domains;    }
sub nets    {  return keys %NETS;  }

sub name {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->name : 0;
}

sub ip {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->ip : 0;
}

sub ttl {
    my $n = lc(shift);
    return 0 unless exists $ENTRIES{$n};
    return $ENTRIES{$n}->ttl if $ENTRIES{$n}->ttl > 0;
    return ($ENTRIES{$n}->soa)[6];
}

sub soa {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->soa : 0;
}

sub hacks {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->hacks : 0;
}

sub delegated {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->delegated : 0;
}

sub servers {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->servers : 0;
}

sub official {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->official : 0;
}

sub stealth {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->stealth : 0;
}

sub master {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->master : 0;
}

sub zone {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->zone : 0;
}

sub zonefile {
    my $n = lc(shift);
    return exists $ENTRIES{$n} ? $ENTRIES{$n}->zonefile : 0;
}

sub iserve {
    my $n = lc(shift);
    my $server = shift;
    return 0 unless exists $ENTRIES{$n};
    return !!grep(/$server/i,$ENTRIES{$n}->servers);
}

sub imaster {
    my $n = lc(shift);
    my $server = shift;
    return 0 unless exists $ENTRIES{$n};
    return ((lc($server) eq lc($ENTRIES{$n}->master)) or ($ENTRIES{$n}->master eq "ALL"));
}

# Even little kids know their names, both first and last, but Unix is
# still so stupid we have to teach it its name over and over again.
# This is a one-shot sub to do that.  It's designed to be replaced by
# the hostfqdn from Net::Domain if that module ever makes it into the
# perl distribution.

sub hostfqdn {
    my($host,$domain);
    local *RES;

    chomp ($host = `hostname`);                           # host part
    $host =~ s/[\0\r\n]+//go;
    $host =~ s/(\A\.+|\.+\Z)//go;
    $host =~ s/\.\.+/\./go;

    if($host !~ /\./ and open(RES,"/etc/resolv.conf")) {  # domain part
        while(<RES>) {
            $domain = $1
                if(/\A\s*(?:domain|search)\s+(\S+)/);
        }
        close(RES);
	$host .= ".$domain";
    }

    return "\u$host";
}

# playground area -- interesting, but unnecessary procdures:
#
#  display_config($file), print_config()
#

sub display_config {

    # usage:  perl -mDNS_config -e 'DNS_config::display_config' [file]

    my($file) = @_;
    $file = $::ARGV[0] unless $file;
    load_config($file,1);
    print_config();
}

sub print_config {

    foreach $E (@entries) {

	# comments, type

	$comments = $E->comments;
	$comments =~ s/\n{2,}/\n\n/g;
	print $comments if $comments;

	print $E->type, ": ", $E->name, " { \n";

	# host - ip

	if ($E->type eq 'host') {
	    print "  ip: ", $E->ip, "\n" if ($E->ip);
	    print "}\n";
	    next;
	}

	# zone - ttl, soa, master, official, stealth, hacks, delegations

        print "  ttl:      ", $E->ttl,                 "\n" if ($E->ttl);
        print "  soa:      ", join(", ",$E->soa),      "\n" if ($E->soa);
        print "  master:   ", $E->master,              "\n" if ($E->master);
        print "  official: ", join(", ",$E->official), "\n" if ($E->official);
        print "  stealth:  ", join(", ",$E->stealth),  "\n" if ($E->stealth);
        print "  hacks:    ", $E->hacks,               "\n" if ($E->hacks);

	foreach $del ($E->delegated) { print "  delegated: $del\n"; }

	print "}\n"
    }
}

1;
