/************************************************************************
 *
 * FullSearchCommand.java - base class for all full searches
 *
 * $Id: FullSearchCommand.java,v 1.8 2006/12/11 20:27:09 priimak Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2003, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import stanford.netdb.utils.*;
import java.util.regex.*;
import java.util.*;
import java.text.*;

abstract class FullSearchCommand extends DBCall {

    protected Hashtable result = new Hashtable();
    protected Hashtable search_params;
    protected Comparator ip_cmp;
    protected Comparator string_cmp;

    public FullSearchCommand(stanford.netdb.utils.Logger logger,
                             Hashtable search_params)
    {
        super(logger);
        db_call_type = SELECT;
        this.search_params = search_params;

        ip_cmp = new Comparator() {
                public int compare(Object o1, Object o2) {
                    long cmp_val =
                        Utils.ipstr2long((String)o1) - Utils.ipstr2long((String)o2);
                    return cmp_val>0?1:(cmp_val<0?-1:0);
                }
                public boolean equals(Object obj) {
                    return true;
                }
            };

        string_cmp = new Comparator() {
                public int compare(Object o1, Object o2) {
                    return o1.toString().compareToIgnoreCase(o2.toString());
                }
                public boolean equals(Object obj) {
                    return true;
                }
            };
    }

    public String getSearchSQL() throws Exception {
        return "";
    }

    public int errorResultSetIndex() {
        return -1; // no errors in a result set
    }

    public String getSQL() { // unsed but has to be declared
        return "";
    }

    // Ensures that non null string value
    protected String n2s(Object src) {
        if( src == null )
            return "";
        else
            return ((String)src).trim();
    }

    // Builds 'Modified', 'Modified By', 'Created', 'Created By' parts in a select clause
    protected String modBlock(Hashtable search_params) {
        // first find appropriate keys
        String displayOrder_modby               = "";
        String displayOrder_modbyDate           = "";
        String displayOrder_createdby           = "";
        String displayOrder_createdbyDate       = "";

        Enumeration enm = search_params.keys();
        while( enm.hasMoreElements() ) {
            String key = (String)enm.nextElement();
            if(      key.matches("^displayOrder\\..*\\.Modified By$") )
                displayOrder_modby = key;
            else if( key.matches("^displayOrder\\..*\\.Date Modified$") )
                displayOrder_modbyDate = key;
            else if( key.matches("^displayOrder\\..*\\.Created By$") )
                displayOrder_createdby = key;
            else if( key.matches("^displayOrder\\..*\\.Date Created$") )
                displayOrder_createdbyDate = key;
        }

        return
            (n2s(search_params.get(displayOrder_modby)).equals("-1")?"":
             ", mod_by_name || '(' || mod_by_netid || ')' \"Modified By\"")+
            (n2s(search_params.get(displayOrder_modbyDate)).equals("-1")?"":
             ", modified \"Date Modified\"")+
            (n2s(search_params.get(displayOrder_createdby)).equals("-1")?"":
             ", created_by_name || '(' || created_by_netid || ')' \"Created By\"")+

            (n2s(search_params.get(displayOrder_createdbyDate)).equals("-1")?"":
             ", created \"Date Created\"");
    } // end of modBlock(...)

    // Convert brief field type into long one used by display engine
    protected String type(String brief_type) {
        if(      brief_type.equals("txt") ||
                 brief_type.equals("hws") ||
                 brief_type.equals("b") )
            return "String";
        else if( brief_type.equals("num") )
            return "Numeric";
        else if( brief_type.equals("ips") )
            return "IP";
        else if( brief_type.equals("tst") )
            return "Timestamp";
        else if( brief_type.equals("ns") )
            return "NoSortString";
        else if( brief_type.equals("nn") )
            return "NoSortNumber";
        else if( brief_type.equals("ni") )
            return "NoSortIP";
        else
            return brief_type;
    } // end of type(...)

    // goes over displayOrder.XXX parameters and assemble displayOrder Vector
    protected void assmDisplayOrder(Vector displayOrder, Hashtable search_params) {
        Hashtable entangled_columns = new Hashtable();
        Enumeration enm = search_params.keys();
        while( enm.hasMoreElements() ) {
            String field = (String)enm.nextElement();
            if( field.equals("displayOrder.Actions") || !field.startsWith("displayOrder.") )
                continue;
            Hashtable dsp = new Hashtable();
            Matcher m =
                Pattern.
                compile("^displayOrder\\.(txt|num|ip|hw|tst|b|ns|nn|ni)\\.(\\d+)\\.(.+)$").
                matcher(field);
            if( !m.matches() || m.group(3) == null  )
                continue;

            String name = m.group(3).trim();
            String requested_place = n2s(search_params.get(field));
            dsp.put("type", type(m.group(1)));
            dsp.put("def_place", m.group(2));
            dsp.put("requested_place", requested_place);
            dsp.put("name", name);
            displayOrder.addElement(dsp);

            if( ( ( name.equals("LowRes") || name.equals("HighRes") ) && !requested_place.equals("-1") ) ||
                name.equals("AddressSpace") )
                    entangled_columns.put(name, dsp);
        }

        // take care of entangled columns
        if( entangled_columns.get("AddressSpace") != null ) {
            Hashtable as = (Hashtable)entangled_columns.get("AddressSpace");
            Utils.nHash(entangled_columns.get("LowRes") ).put("def_place",       as.get("def_place"));
            Utils.nHash(entangled_columns.get("LowRes") ).put("requested_place", as.get("requested_place"));
            Utils.nHash(entangled_columns.get("HighRes")).put("def_place",       as.get("def_place"));
            Utils.nHash(entangled_columns.get("HighRes")).put("requested_place", as.get("requested_place"));
        }
    } // end of assmDisplayOrder(...)

    // assembles common sql 'where' clause
    protected String assmCommonSQL(Hashtable search_params) throws Exception {
        String sql = "";
        // fields where 'or' operator is implied need some special treatment
        Hashtable or_group_fields = new Hashtable();
        boolean has_search_parameters = false;
        Enumeration enm = search_params.keys();
        while( enm.hasMoreElements() ) {
            String table = (String)search_params.get("fs.table");
            String field = (String)enm.nextElement();
            String key = field;

            if( key.startsWith("fs.FS_NodeType.cb.i.not_") ) {
                sql += (sql.equals("")?"":" INTERSECT ")+
                    " ( SELECT ID FROM FS_Node MINUS ( SELECT ID FROM FS_NodeType WHERE type = '"+key.substring(24)+"' ) ) ";
                continue;
            }

            // ignore few fields
            if( key.equals("txt.customFields")   ||
                key.equals("txt.DHCPoptions")    ||
                key.equals("txt.NetDHCPoptions") ||
                key.equals("txt.ASDHCPoptions") )
                continue;

            // skip some keys
            if( field.endsWith(" error")          ||
                key.equals("cb.i.Name:Node")      ||
                key.equals("cb.i.Name:Alias")     ||
                key.equals("cb.i.Name:Interface") ||
                key.equals("cb.i.Name:IP")        ||
                key.equals("cb.i.Name:IPC") )
                continue;

            Matcher m = Pattern.compile("^fs\\.([^.]+)\\.(.*)$").matcher(field);
            if( m.matches() && m.group(2) != null ) {
                // field name contains table name as well
                table = m.group(1);
                field = m.group(2);
            }

            m = Pattern.compile("^(txt|num|tst|ip|hw|ns)\\.(group.\\.)?(.*)").matcher(field);
            if( m.matches() ) { // html text fields
                String value = n2s(search_params.get(key)).trim();
                if( value.length() == 0 )
                    continue;
                String column = m.group(3);
                if( column.matches(".*lower\\(.*") || column.matches(".*_lc$") )
                    value = value.toLowerCase();

                String sql_add = Utils.parseField(search_params, key, field, value, table, column);

                if( key.equals("fs.FS_NameNode.txt.fullname_lc") ) { // special handling for this field
                    sql += (sql.equals("")?"":" INTERSECT ")+
                        " ( select node_id id from FS_NameNode where id in "+sql_add+" and ";
                    String extra_where_part = "";
                    extra_where_part += search_params.get("cb.i.Name:Node")      ==null?"":
                        ((extra_where_part.length()>0?" or ":"")+search_params.get("cb.i.Name:Node"));
                    extra_where_part += search_params.get("cb.i.Name:Alias")     ==null?"":
                        ((extra_where_part.length()>0?" or ":"")+search_params.get("cb.i.Name:Alias"));
                    extra_where_part += search_params.get("cb.i.Name:Interface") ==null?"":
                        ((extra_where_part.length()>0?" or ":"")+search_params.get("cb.i.Name:Interface"));
                    extra_where_part += search_params.get("cb.i.Name:IP")        ==null?"":
                        ((extra_where_part.length()>0?" or ":"")+search_params.get("cb.i.Name:IP"));
                    extra_where_part += search_params.get("cb.i.Name:IPC")       ==null?"":
                        ((extra_where_part.length()>0?" or ":"")+search_params.get("cb.i.Name:IPC"));
                    sql += " ( " + extra_where_part + " ) ) ";

                } else if( m.group(2) != null ) {
                    // these keys are different and thus handled separately

                    String sql_group = (String)or_group_fields.get(m.group(2));
                    if( sql_group == null ) sql_group = "";
                    sql_group += (sql_group.length() > 0?" union ":"")+" ( "+sql_add+" ) ";
                    or_group_fields.put(m.group(2), sql_group);

                } else
                    sql += (sql.equals("")?"":" INTERSECT ")+sql_add;

                has_search_parameters = true;

            } else if( field.startsWith("rb.") ) { // radio buttons
                if( field.startsWith("rb.b.") ) { // boolean
                    String value = n2s(search_params.get(key)).trim();
                    if( value.equals("-1") ) // it is 'Either' condition
                        continue;
                    String column = field.substring(5);
                    sql += (sql.equals("")?"":" INTERSECT ")+
                        " ( SELECT ID FROM "+table+" WHERE "+column+" = "+value+" ) ";
                }

            } else if( field.startsWith("cb.") ) { // check box
                if( field.startsWith("cb.i.") ) { // inline sql
                    String full_sql = null;
                    String where_clause = n2s(search_params.get(key)).trim();
                    if( where_clause.length() == 0 )
                        continue;

                    if( field.startsWith("cb.i.f") ) // it is a full sql not just a where clause
                        full_sql = where_clause;

                    m = Pattern.compile("^(group.\\.)(.+)$").matcher(where_clause);
                    if( m.matches() ) {
                        // group of conditions which should be combined using union
                        String sql_group = (String)or_group_fields.get(m.group(1));
                        if( sql_group == null ) sql_group = "";
                        sql_group += (sql_group.length() > 0?" union ":"")+
                            (full_sql==null?" ( SELECT ID FROM "+table+" WHERE "+m.group(2)+" ) ":m.group(2));
                        or_group_fields.put(m.group(1), sql_group);
                    } else
                        sql += (sql.equals("")?"":" INTERSECT ")+
                            (full_sql==null?" ( SELECT ID FROM "+table+" WHERE "+where_clause+" ) ":where_clause);
                }
            }
        }

        if( search_params.get("tst.last_login >= ") != null && sql.trim().length() > 0 )
            // always do user full search whenever it is possible
            has_search_parameters = true;

        if( !has_search_parameters )
            throw new Exception("Please enter some search criteria.");

        // add grouped fields
        enm = or_group_fields.keys();
        while( enm.hasMoreElements() ) {
            sql += (sql.equals("")?"":" INTERSECT ")+
                " ( "+or_group_fields.get(enm.nextElement())+" ) ";
        }

        return sql;
    } // end of assmCommonSQL(...)

    // fills in initial data into result Hashtable
    protected void init_result(Hashtable search_params, Vector displayOrder, String type) {
        result.put("object",        n2s(search_params.get("searchtype")));
        result.put("displayOrder",  displayOrder);
        result.put("show actions",
                   n2s(search_params.get("displayOrder.Actions")).equals("Yes")?"true":"false");
        result.put("output_format", Utils.n2s(search_params.get("output_format")));
        result.put("search type",   type);
    }

    // this is a generic picker for a full search result
    public Hashtable processResultSet(Hashtable res) throws Exception {
        if( res.get("errors") != null ) {
            Hashtable errors = new Hashtable();
            errors.put("errors", res.get("errors"));
            return errors;
        }

        Vector column_names = (Vector)res.get("column_names");
        Hashtable result = new Hashtable();
        Vector result_set = (Vector)res.get("resultSet");
        for( int i = 0; i < result_set.size(); i++ ) {
            Vector row = (Vector)result_set.elementAt(i);

            String id = row.elementAt(0).toString();
            Hashtable rawH = (Hashtable)result.get(id);
            if( rawH == null ) {
                rawH = new Hashtable();
                result.put(id, rawH);
            }

            for( int j = 0; j < row.size(); j++ ) {
                Hashtable cell = (Hashtable)rawH.get(column_names.elementAt(j));
                if( cell == null ) {
                    cell = new Hashtable();
                    rawH.put(column_names.elementAt(j), cell);
                }
                if( row.elementAt(j) != null ) {
                    Object element = row.elementAt(j);
                    if( element instanceof oracle.sql.DATE )
                        cell.put(((oracle.sql.DATE)element).timestampValue(), "");
                    else if( element instanceof oracle.sql.TIMESTAMP )
                        cell.put(((oracle.sql.TIMESTAMP)element).timestampValue(), "");
                    else
                        cell.put(element, "");
                }
                rawH.put(column_names.elementAt(j), cell);
            }
        }

        // go over the result and convert cells into vectors
        Enumeration enm = result.keys();
        while( enm.hasMoreElements() ) {
            Hashtable rawH = (Hashtable)result.get(enm.nextElement());
            Enumeration cells = rawH.keys();
            while( cells.hasMoreElements() ) {
                Object column_name = cells.nextElement();
                Hashtable cell = (Hashtable)rawH.get(column_name);
                Vector vCell = new Vector();
                Enumeration cell_enm = cell.keys();
                while( cell_enm.hasMoreElements() ) {
                    vCell.addElement(cell_enm.nextElement());
                }
                rawH.put(column_name, vCell);
            }
        }

        this.result.put("result", result);

        return this.result;
    } // end of processResultSet(...)

    // given display key returns boolean indicating if this display key is selected
    protected boolean selected(String display) {
        return !n2s(search_params.get(display)).equals("-1");
    }

    protected boolean displayed(String display) {
        return rselected("^displayOrder.+\\."+display+"$");
    }

    protected boolean rselected(String display) {
        Iterator it = search_params.keySet().iterator();
        while( it.hasNext() ) {
            Object key = it.next();
            if( key.toString().matches(display) )
                return !n2s(search_params.get(key)).equals("-1");
        }
        return false;
    }

    // given a key find if it has non null value in search_parmeters
    protected boolean spNull(String key) {
        return (search_params.get(key) == null);
    }

    protected Object[] sorted_keys(Hashtable src, Comparator cmp) {
        return Utils.sorted_keys(src, cmp);
    }
}
