/************************************************************************
 *
 * NetworkInsertCommand.java - complex db call pNetwork 'insert'
 *
 * $Id: NetworkInsertCommand.java,v 1.3 2006/09/26 17:51:49 priimak Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2002, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import stanford.netdb.utils.*;
import stanford.netdb.lib.*;
import java.util.*;
import java.lang.*;
import java.sql.*;

/** complex db call pNetwork 'update' */
class NetworkInsertCommand extends NetworkCommand implements ComplexDBCall {

    private Hashtable newNet;
    private NetDBii netdb;
    private Hashtable acceptedNames = new Hashtable();

    public NetworkInsertCommand(NetDBii netdb, Hashtable newNet, Integer op_uid)
        throws Exception {
        super((stanford.netdb.utils.Logger)netdb);

        this.netdb = netdb;
        this.newNet = Utils.copyHash(newNet);
        this.op_uid = op_uid;

        values[ACTION_INDEX] = "insert";
        values[ID_OP_USER_INDEX] = op_uid;
        String service = (String)newNet.get("dhcp_service");
        if( service != null )
            values[SERVICE_INDEX] = new String(service);
        String comment = (String)newNet.get("comment");
        if( comment != null ) {
            if( comment.length() > 255 )
                comment = comment.substring(0, 254);
            values[COMMENT_INDEX] = new String(comment);
        }
        values[ID_NETWORK_INDEX] = new OutParam(Types.NUMERIC);
    }

    public Hashtable doIt(NetDBConnection con) throws NetDBException, Exception {
        boolean commit_finally = false;
        Hashtable warnings = new Hashtable(); // contains warnings
        Hashtable accepted_ip_addresses = new Hashtable();

        try {
            logger.debug("inserting network");
            BeginTransaction(con.con);

            Hashtable result = processResultSet(con.callDB(this));
            checkErrors("Failed to insert network.", result);

            // get network id and lock
            idNet = new Integer((String)result.get("id"));
            lock = (Timestamp)result.get("locked");

            DBCall call = null;

            // insert network names
            Hashtable names = (Hashtable)newNet.get("names");
            Enumeration enm = names.keys();
            while( enm.hasMoreElements() ) {
                String key = (String)enm.nextElement();
                Hashtable name = (Hashtable)names.get(key);

                // first reserve name
                Hashtable reservedName =
                    netdb.reserveName(con, (String)name.get("name")+"."+
                                      (String)name.get("domain"), Constants.NETWORK_NAME_TYPE);

                name.put("name", Utils.name((String)reservedName.get("name")));
                name.put("domain", Utils.domain((String)reservedName.get("name"), netdb.getDefaultDomain()));

                call = new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null,
                                             Constants.NETWORK_NAME_TYPE, idNet, null, lock, null, op_uid);
                checkErrors("Failed to accept name \""+
                            name.get("name")+"."+name.get("domain")+"\"",
                            call.processResultSet(con.callDB(call)));
            }

            // insert locations
            logger.debug("inserting locations");
            Hashtable locations = Utils.nHash(newNet.get("locations"));
            Object[] locations_to_insert = locations.keySet().toArray();

            for( int i = 0; i < locations_to_insert.length; i++ ) {
                call = new NetworkLocationCommand(logger, NetworkLocationCommand.INSERT_ACTION,
                                                  idNet, lock,
                                                  Utils.nStr(((Hashtable)locations.get(locations_to_insert[i])).get("name")),
                                                  op_uid);
                checkErrors("Failed to insert location.", call.processResultSet(con.callDB(call)));
            }

            // insert groups
            logger.debug("inserting groups");

            // copy hash into vector
            Hashtable newGroups = (Hashtable)newNet.get("groups");
            Vector groups = new Vector();
            enm = newGroups.keys();
            while( enm.hasMoreElements() ) {
                groups.addElement(enm.nextElement());
            }
            // now call db
            ComplexDBCall cmd =
                new ObjectGroupCommand(logger, ObjectGroupCommand.ADD_GROUPS, idNet.intValue(),
                                       ObjectGroupCommand.NETWORK, lock, groups, op_uid);

            checkErrors("Failed to insert groups", cmd.doIt(con));

            // insert dhcp options
            Hashtable options = (Hashtable)newNet.get("dhcp_options");
            if( options != null ) {
                cmd = new DHCPOptionsActionCommand(logger, DHCPOptionsActionCommand.INSERT_ACTION,
                                                   lock, idNet, Constants.NETWORK, options, op_uid);
                checkErrors("Failed to insert dhcp options.", cmd.doIt(con));
            }

            // insert address spaces
            // adding dhcp ip addresses now allowed in create thus we do nothing about them
            // even if they are present in a hash
            Hashtable address_spaces = (Hashtable)newNet.get("address_spaces");
            if( address_spaces == null )
                address_spaces = new Hashtable();
            enm = address_spaces.keys();
            while( enm.hasMoreElements() ) {
                Hashtable as = (Hashtable)address_spaces.get(enm.nextElement());
                cmd = new AddrSpaceInsertCommand(logger, this, as, idNet, op_uid);

                result = cmd.doIt(con);
                checkErrors("Failed to insert address space.", result);
                Utils.appendHash(accepted_ip_addresses, Utils.nHash(result.get("accepted_ip_addresses")));
            }

            // verify whole stuff
            checkErrors("Failed to update network.", doCheck(con));

            commit_finally = true;
            return warnings;

        } catch(MiniException ex) {
            return ex.result;

        } finally {
            if( commit_finally ) {
                try {
                    logger.info("inserting network done");
                    // Unlock the network now that we're done.
                    try {
                        DBCall call =
                            new NetworkLockUnLockCommand(logger, NetworkLockUnLockCommand.UNLOCK_ACTION, null,
                                                         idNet.intValue(), lock, op_uid);
                        call.processResultSet(con.callDB(call));
                    } catch (Exception exxx) { }
                    logger.debug("CommitTransaction");
                    CommitTransaction(con.con);
                } catch(Exception exx) { }

                // unregister accept ip addresses from list of registered
                Enumeration enm = accepted_ip_addresses.keys();
                while( enm.hasMoreElements() )
                    netdb.unregister_reserved_ip((String)enm.nextElement());

                // remove accepted names from list of reserved
                enm = acceptedNames.keys();
                while( enm.hasMoreElements() )
                    netdb.removeReservedName((Hashtable)enm.nextElement());
            } else
                try {
                    logger.info("inserting network failed");
                    logger.debug("RollbackTransaction");
                    RollbackTransaction(con.con);
                } catch(Exception exx) { }
        }
    } // end of doIt(...)
}
