/************************************************************************
 *
 * NetworkModifyCommand.java - complex db call pNetwork 'update'
 *
 * $Id: NetworkModifyCommand.java,v 1.3 2006/09/26 17:51:32 priimak Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2002, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.lib;

import stanford.netdb.utils.*;
import stanford.netdb.lib.*;
import java.util.*;
import java.sql.*;
import java.lang.*;

/** complex db call pNetwork 'update' */
class NetworkModifyCommand extends NetworkCommand implements ComplexDBCall {

    private NetDBii netdb;
    private Hashtable acceptedNames = new Hashtable();
    private Hashtable accepted_ip_addresses = new Hashtable();

    public NetworkModifyCommand(NetDBii netdb, Hashtable oldNet,
                                Hashtable newNet, Integer op_uid)
        throws Exception
    {
        super((stanford.netdb.utils.Logger)netdb);

        db_call_type = ORA_FUN_IGNORE_RET_CALL;

        this.netdb  = netdb;
        this.oldNet = Utils.copyHash(oldNet);
        this.newNet = Utils.copyHash(newNet);
        this.op_uid = op_uid;

        values[ACTION_INDEX] = "update";
        values[ID_OP_USER_INDEX] = op_uid;
        values[SERVICE_INDEX] = o2v(newNet.get("dhcp_service"));
        String comment = (String)newNet.get("comment");
        if( comment != null ) {
            if( comment.length() > 255 )
                comment = comment.substring(0, 254);
            values[COMMENT_INDEX] = new String(comment);
        }
        values[ID_NETWORK_INDEX] = idNet  = new Integer((String)newNet.get("id"));
        values[LOCKED_INDEX]     = lock = (Timestamp)newNet.get("locked");

    }

    public Hashtable doIt(NetDBConnection con) throws NetDBException, Exception {
        boolean commit_finally = false;
        Hashtable warnings = new Hashtable(); // contains warnings
        try {
            logger.info("modifying network id "+idNet);
            logger.debug("BeginTransacton");
            BeginTransaction(con.con);
            checkErrors("Failed to update network.", processResultSet(con.callDB(this)));

            // locations
            updateLocations(con);

            // dhcp options
            updateDHCPOptions(con);

            // groups
            updateGroups(con);

            // collect address spaces to delete, insert and update
            Hashtable ases = collectASes();
            deleteASes(con, (Hashtable)ases.get("ases_to_delete"), lock);
            insertASes(con, (Hashtable)ases.get("ases_to_insert"), lock);
            updateASes(con, (Hashtable)ases.get("ases_to_update"), lock);

            Hashtable names = Utils.sortOutNames(oldNet, newNet);
            deleteNames(con, (Hashtable)names.get("names_to_delete"));
            insertNames(con, (Hashtable)names.get("names_to_insert"));

            // verify whole stuff
            checkErrors("Failed to update network.", doCheck(con));

            DBCall call = new NetworkLockUnLockCommand(logger, NetworkLockUnLockCommand.UNLOCK_ACTION, null,
                                                       idNet.intValue(), lock, op_uid);
            Hashtable result = call.processResultSet(con.callDB(call));

            Vector error_set = (Vector)result.get("errors");
            if( error_set != null && error_set.size() > 0 )
                return result;

            ((NetDBii)logger).unregister_lock(lock, new Integer(Constants.NETWORK));

            commit_finally = true;
            return warnings;

        } catch(MiniException ex) {
            return ex.result;

        } finally {
            if( commit_finally ) {
                try {
                    logger.info("modifying network id "+idNet+" done");
                    logger.debug("CommitTransaction");
                    CommitTransaction(con.con);
                } catch(Exception exx) { }

                // unregister accept ip addresses from list of registered
                Enumeration enm = accepted_ip_addresses.keys();
                while( enm.hasMoreElements() )
                    netdb.unregister_reserved_ip((String)enm.nextElement());

                // remove accepted names from list of reserved
                enm = acceptedNames.keys();
                while( enm.hasMoreElements() )
                    netdb.removeReservedName((Hashtable)enm.nextElement());
            } else
                try {
                    logger.info("modifying network id "+idNet+" failed");
                    logger.debug("RollbackTransaction");
                    RollbackTransaction(con.con);
                } catch(Exception exx) { }
        }
    }

    private void insertASes(NetDBConnection con, Hashtable ases, Timestamp lock)
        throws NetDBException, Exception, MiniException {
        Enumeration enm = ases.keys();
        while( enm.hasMoreElements() ) {
            String key = (String)enm.nextElement();

            Hashtable result =
                (new AddrSpaceInsertCommand(logger, this, (Hashtable)ases.get(key),
                                            new Integer((String)newNet.get("id")),
                                            op_uid)).doIt(con);
            checkErrors("Failed to inser address space.", result);
            Utils.appendHash(accepted_ip_addresses, Utils.nHash(result.get("accepted_ip_addresses")));
        }

    }

    private void updateASes(NetDBConnection con, Hashtable ases, Timestamp lock)
        throws NetDBException, Exception, MiniException {
        Enumeration enm = ases.keys();
        while( enm.hasMoreElements() ) {
            Hashtable result =
                (new AddrSpaceModifyCommand(logger, this, (Hashtable)ases.get(enm.nextElement()),
                                            new Integer((String)newNet.get("id")),
                                            op_uid)).doIt(con);
            checkErrors("Failed to modify address space.", result);
            Utils.appendHash(accepted_ip_addresses, Utils.nHash(result.get("accepted_ip_addresses")));
        }

    }

    private void deleteASes(NetDBConnection con, Hashtable ases, Timestamp lock)
        throws NetDBException, Exception, MiniException {
        Enumeration enm = ases.keys();
        while( enm.hasMoreElements() ) {
            String key = (String)enm.nextElement();
            Hashtable as = (Hashtable)ases.get(key);
            String asPrefix = (String)as.get("addr_space_prefix");
            DBCall call = new AddrSpaceDeleteCommand(logger, asPrefix, lock, op_uid);
            checkErrors("Failed to delete address space.", call.processResultSet(con.callDB(call)));
            ((Hashtable)oldNet.get("address_spaces")).remove(as.get("key"));
        }
    }

    private void updateLocations(NetDBConnection con)
        throws NetDBException, Exception, MiniException {
        // delete old locations
        DBCall call;
        Hashtable oldLocations = Utils.nHash(oldNet.get("locations"));
        Hashtable newLocations = Utils.nHash(newNet.get("locations"));

        Vector locations_to_insert = new Vector();
        Vector locations_to_delete = new Vector();
        Utils.keysIntersect(oldLocations, newLocations, locations_to_insert, locations_to_delete, new Vector());

        for( int i = 0; i < locations_to_insert.size(); i++ ) {
            call = new NetworkLocationCommand(logger, NetworkLocationCommand.INSERT_ACTION,
                                              (Integer)values[ID_NETWORK_INDEX], lock,
                                              Utils.nStr(((Hashtable)newLocations.get(locations_to_insert.elementAt(i))).get("name")),
                                              op_uid);
            checkErrors("Failed to insert location.", call.processResultSet(con.callDB(call)));
        }

        for( int i = 0; i < locations_to_delete.size(); i++ ) {
            call = new NetworkLocationCommand(logger, NetworkLocationCommand.DELETE_ACTION,
                                              (Integer)values[ID_NETWORK_INDEX], lock,
                                              Utils.nStr(((Hashtable)oldLocations.get(locations_to_delete.elementAt(i))).get("name")),
                                              op_uid);
            checkErrors("Failed to delete location.", call.processResultSet(con.callDB(call)));
        }
    } // end of updateLocations(...)

    private Hashtable updateDHCPOptions(NetDBConnection con)
        throws NetDBException, Exception, MiniException {
        // delete old options
        ComplexDBCall call;
        Hashtable options = (Hashtable)oldNet.get("dhcp_options");
        if( options != null ) {
            call = new DHCPOptionsActionCommand(logger, DHCPOptionsActionCommand.DELETE_ACTION,
                                                lock, (Integer)values[ID_NETWORK_INDEX],
                                                Constants.NETWORK, options, op_uid);
            checkErrors("Failed to delete dhcp options.", call.doIt(con));
        }

        // insert new options
        options = (Hashtable)newNet.get("dhcp_options");
        if( options != null ) {
            call = new DHCPOptionsActionCommand(logger, DHCPOptionsActionCommand.INSERT_ACTION,
                                                lock, (Integer)values[ID_NETWORK_INDEX],
                                                Constants.NETWORK, options, op_uid);
            checkErrors("Failed to insert dhcp options.", call.doIt(con));
        }

        // no errors here
        return new Hashtable();
    }

    private void updateGroups(NetDBConnection con)  throws NetDBException, Exception, MiniException {
        Hashtable oldGroups = (Hashtable)oldNet.get("groups");
        if( oldGroups == null )
            oldGroups = new Hashtable();
        Hashtable newGroups = (Hashtable)newNet.get("groups");
        if( newGroups == null )
            newGroups = new Hashtable();
        Vector groups_to_insert = new Vector();
        Vector groups_to_delete = new Vector();
        Vector groups_to_keep = new Vector();
        Utils.keysIntersect(oldGroups, newGroups, groups_to_insert, groups_to_delete, groups_to_keep);

        // insert groups_to_insert
        logger.debug("inserting groups_to_insert");
        ComplexDBCall cmd =
            new ObjectGroupCommand(logger, ObjectGroupCommand.ADD_GROUPS, idNet.intValue(),
                                   ObjectGroupCommand.NETWORK, lock, groups_to_insert, op_uid);
        checkErrors("Failed to insert groups", cmd.doIt(con));

        logger.debug("deleting groups_to_delete");
        cmd = new ObjectGroupCommand(logger, ObjectGroupCommand.REMOVE_GROUPS, idNet.intValue(),
                                     ObjectGroupCommand.NETWORK, lock, groups_to_delete, op_uid);
        checkErrors("Failed to delete groups", cmd.doIt(con));
    }

    private Hashtable collectASes() {
        Hashtable ases_to_insert = new Hashtable();
        Hashtable ases_to_delete = new Hashtable();
        Hashtable ases_to_update = new Hashtable();

        Hashtable newASes = (Hashtable)newNet.get("address_spaces");
        if( newASes != null ) {
            Enumeration enm = newASes.keys();
            while( enm.hasMoreElements() ) {
                Hashtable as = (Hashtable)newASes.get(enm.nextElement());
                ases_to_insert.put(as.get("addr_space_prefix"), as);
            }
        }

        Hashtable oldASes = (Hashtable)oldNet.get("address_spaces");
        if( oldASes != null ) {
            Enumeration enm = oldASes.keys();
            while( enm.hasMoreElements() ) {
                String key = (String)enm.nextElement();
                Hashtable oldAS = (Hashtable)oldASes.get(key);
                String asPrefix = (String)oldAS.get("addr_space_prefix");
                Hashtable newAS = (Hashtable)ases_to_insert.get(asPrefix);
                if( newAS == null ) {
                    ases_to_delete.put(asPrefix, oldAS);
                    oldAS.put("key", key);
                    oldASes.remove(asPrefix); // it is important !!!
                } else {
                    newAS = (Hashtable)ases_to_insert.remove(asPrefix);
                    ases_to_update.put(asPrefix, newAS);

                    String oldComment = (String)oldAS.get("comment");
                    if( oldComment == null ) oldComment = "";
                    newAS.put("old.comment", oldComment);

                    Hashtable oldGroups = (Hashtable)oldAS.get("groups");
                    if( oldGroups == null )
                        oldGroups = new Hashtable();
                    newAS.put("old.groups", oldGroups);

                    Hashtable oldDHCP_options = (Hashtable)oldAS.get("dhcp_options");
                    if( oldDHCP_options == null )
                        oldDHCP_options = new Hashtable();
                    newAS.put("old.dhcp_options", oldDHCP_options);

                    Hashtable oldDHCPips = (Hashtable)oldAS.get("dhcp_addresses");
                    if( oldDHCPips == null )
                        oldDHCPips = new Hashtable();
                    newAS.put("old.dhcp_addresses", oldDHCPips);

                    newAS.put("old.low_res", oldAS.get("low_res"));
                    newAS.put("old.high_res", oldAS.get("high_res"));
                }
            }
        }

        Hashtable result = new Hashtable();
        result.put("ases_to_insert", ases_to_insert);
        result.put("ases_to_delete", ases_to_delete);
        result.put("ases_to_update", ases_to_update);

        return result;
    } // end of collectASes()

    private void deleteNames(NetDBConnection con, Hashtable names_to_delete)
        throws NetDBException, Exception, MiniException {
        Enumeration enm = names_to_delete.keys();
        while( enm.hasMoreElements() ) {
            String key = (String)enm.nextElement();
            Hashtable name = (Hashtable)names_to_delete.get(key);

            String nameType = (String)name.get("type");
            if( nameType.equals("network_name") )
                nameType = Constants.NETWORK_NAME_TYPE;
            else if( nameType.equals("alias") )
                nameType = Constants.NAME_ALIAS_TYPE;
            else if( nameType.equals("mx") )
                nameType = Constants.MX_TYPE;
            else if( nameType.equals("ip_name") )
                nameType = Constants.ADDRESS_NAME_TYPE;

            DBCall call = new NameDeleteCommand(logger, (String)name.get("name"),
                                                (String)name.get("domain"),
                                                nameType, lock, op_uid);
            checkErrors("Failed to delete name ["+name.get("name")+"."+name.get("domain")+"]",
                        call.processResultSet(con.callDB(call)));
        }
    } // end of deleteNames(...)

    private void insertNames(NetDBConnection con, Hashtable names_to_insert)
        throws NetDBException, Exception, MiniException {
        Enumeration enm = names_to_insert.keys();
        while( enm.hasMoreElements() ) {
            String key = (String)enm.nextElement();
            // first go over parenting names,
            // i.e. names which may have aliases and/or mx's
            Hashtable name = (Hashtable)names_to_insert.get(key);
            if( name.get("refs") == null )
                continue;

            String nameType = (String)name.get("type");
            if( nameType.equals("network_name") )
                nameType = Constants.NETWORK_NAME_TYPE;
            else if( nameType.equals("alias") )
                nameType = Constants.NAME_ALIAS_TYPE;
            else if( nameType.equals("mx") )
                nameType = Constants.MX_TYPE;
            else if( nameType.equals("ip_name") )
                nameType = Constants.ADDRESS_NAME_TYPE;

            // first reserve name
            Hashtable reservedName =
                netdb.reserveName(con, (String)name.get("name")+"."+
                                  (String)name.get("domain"), nameType);

            name.put("name", Utils.name((String)reservedName.get("name")));
            name.put("domain", Utils.domain((String)reservedName.get("name"),
                                            (String)oldNet.get("default_domain")));
            DBCall call =
                new NameAcceptCommand(logger, new Integer((String)reservedName.get("id")), null,
                                      nameType,
                                      nameType.equals(Constants.NETWORK_NAME_TYPE)?idNet:null,
                                      (nameType.equals(Constants.ADDRESS_NAME_TYPE)?
                                       (String)name.get("ip"):null),
                                       lock, null, op_uid);
            checkErrors("Failed to accept name ["+name.get("name")+"."+name.get("domain")+"]",
                        call.processResultSet(con.callDB(call)));
            Hashtable nameH = new Hashtable();
            nameH.put("fqdn", ((String)name.get("name")).trim().toLowerCase()+"."+
                      ((String)name.get("domain")).trim().toLowerCase());
            nameH.put("type", new String(nameType));
            acceptedNames.put(nameH, "");
        }

        // now we should go over aliases and mxs, but since they are not
        // allowed in Network object we do nothing

    } // end of insertNames(...)
}
