/************************************************************************
 *
 * NetDBServer.java - start this class if using jetty http server engine.
 *
 * $Id: NetDBServer.java,v 1.4 2006/09/26 22:47:39 priimak Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2002, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.html;

import java.lang.*;
import java.util.*;
import java.io.*;
import java.net.*;
import javax.servlet.*;
import org.mortbay.util.*;
import org.mortbay.http.*;
import org.mortbay.jetty.servlet.*;
import org.mortbay.http.handler.*;
import org.mortbay.servlet.*;
import stanford.netdb.lib.*;
import stanford.netdb.utils.*;

public class NetDBServer
{
    public static void main(String[] args)
    {
        // it may not be a good idea to fork depending on a platform
        if (args.length > 0 && args[0].equals("-d"))
        {
            // fork if -d is present; daemon mode
            if (Posix.fork() != 0)
                return;
        }

        // disable SIGHUP
        Posix.disableSIGHUP();

        // print pid, needed for manage
        Utils.write_pid_file();

        // now pick up data from properties file
        Properties netdb_props = new Properties();
        FileInputStream in = null;
        try {

            // if system property 'config.file' is not set in a command line then
            // use Constants.NETDB_PROPERTIES_FILENAME to get properties file name
            in = new FileInputStream(System.getProperty("config.file",
                                                Constants.NETDB_PROPERTIES_FILENAME));
            netdb_props.load(in);
        } catch (Exception ex) {
            System.err.println("ERROR: Failed to load properties. "+ex);
            System.exit(-1);
        } finally {
            try {
                if (in != null)
                    in.close();
            } catch (Exception ex) {}
        }

        // override by system properties
        Properties system_props = System.getProperties();
        for( Object key : system_props.keySet() )
            netdb_props.put(key, system_props.get(key));

        // now override system properties
        for( Object key : netdb_props.keySet() )
            system_props.put(key, netdb_props.get(key));

        // parse PTRpref string
        CommonPage.ptr_pref = Utils.parsePTRprefConfigString();

        // set up logger
        if (Boolean.getBoolean(Constants.DO_SYSLOG_KEY)) {
            Utils.logger.set(Constants.DO_SYSLOG_KEY, new Boolean(true));
            Utils.logger.set(Constants.SYSLOG_PORT_KEY,
                       new Integer(netdb_props.getProperty(Constants.SYSLOG_PORT_KEY, "514")));
            Utils.logger.set(Constants.SYSLOG_HOST_KEY,
                       netdb_props.getProperty(Constants.SYSLOG_HOST_KEY, "localhost"));
            Utils.logger.set(Constants.SYSLOG_FACILITY_KEY,
                       netdb_props.getProperty(Constants.SYSLOG_FACILITY_KEY, "local3"));
            Utils.logger.set(Constants.SYSLOG_PRIORITY_KEY,
                       netdb_props.getProperty(Constants.SYSLOG_PRIORITY_KEY, "info"));
        }

        // change euid and egid
        if (netdb_props.getProperty(Constants.GID_KEY) != null) {
            try {
                Posix.setgid(netdb_props.getProperty(Constants.GID_KEY));
            } catch(Exception ex) {
                // Something went wrong
                Utils.error("failed to set gid ["+
                            netdb_props.getProperty(Constants.GID_KEY)+"]");
            }
        }

        if (netdb_props.getProperty(Constants.UID_KEY) != null) {
            try {
                Posix.setuid(netdb_props.getProperty(Constants.UID_KEY));
            } catch (Exception ex) {
                // Something went wrong
                Utils.error("failed to set uid ["+
                            netdb_props.getProperty(Constants.UID_KEY)+"]");
            }
        }

        // Create the jetty http server
        HttpServer server = new HttpServer();

        // Create a port listener
        SocketListener listener;
        try {
            // if ip interface address specified then use it
            if( netdb_props.getProperty(Constants.IP_INTERFACE_KEY) != null)
                listener = new SocketListener(
                  new InetAddrPort(Inet4Address.getByAddress(Utils.ipstr2bytes(netdb_props.getProperty(Constants.IP_INTERFACE_KEY))),
                                   Integer.parseInt(netdb_props.getProperty(Constants.NETDB_HTTP_PORT_KEY, "8080"))));
            else
                listener = new SocketListener(
                  new InetAddrPort(Integer.parseInt(netdb_props.getProperty(Constants.NETDB_HTTP_PORT_KEY, "8080"))));

            server.addListener(listener);
        } catch (Exception ex) {
            Utils.error("Failed to create listener. "+ex);
            System.exit(-1);
        }

        ServletHttpContext context = new ServletHttpContext();
        context.setContextPath("/");

        // set current workig directory
        System.setProperty("user.dir",
            System.getProperty("user.dir")+"/"+System.getProperty("working.dir"));

        // enable selenium test directory if needed
        if( System.getProperty("enable.tests", "false").equals("true") ) {
            context.setResourceBase("./test/");
            context.addHandler(new ResourceHandler());
        }

        // hard code all the servlets below
        try {
            context.addServlet("/",
                               "stanford.netdb.html.IndexServlet");
            context.addServlet("/netdb.css",
                               "stanford.netdb.html.NetDBCSS");
            context.addServlet("/index.html",
                               "stanford.netdb.html.IndexServlet");
            context.addServlet("/login",
                               "stanford.netdb.html.LoginServlet");
            context.addServlet("/home",
                               "stanford.netdb.html.HomeServlet");
            context.addServlet("/logout",
                               "stanford.netdb.html.LogoutServlet");
            context.addServlet("/qsearch",
                               "stanford.netdb.html.QuickSearch");
            context.addServlet("/qresults",
                               "stanford.netdb.html.QSearchResults");
            context.addServlet("/InfoGroup",
                               "stanford.netdb.html.InfoGroup");
            context.addServlet("/InfoDomain",
                               "stanford.netdb.html.InfoDomain");
            context.addServlet("/InfoAdminTeam",
                               "stanford.netdb.html.InfoAdminTeam");
            context.addServlet("/images/NetDBlogo.gif",
                               "stanford.netdb.html.NetDBLogo");
            context.addServlet("/InfoUser",
                               "stanford.netdb.html.InfoUser");
            context.addServlet("/InfoNetwork",
                               "stanford.netdb.html.InfoNetwork");
            context.addServlet("/InfoNode",
                               "stanford.netdb.html.InfoNode");
            context.addServlet("/FullSearchForm",
                               "stanford.netdb.html.FullSearchForm");
            context.addServlet("/Department",
                               "stanford.netdb.html.Department");
            context.addServlet("/ModifyDepartment",
                               "stanford.netdb.html.Department");
            context.addServlet("/AddDepartment",
                               "stanford.netdb.html.Department");
            context.addServlet("/DepartmentsList",
                               "stanford.netdb.html.Department");
            context.addServlet("/Location",
                               "stanford.netdb.html.Location");
            context.addServlet("/AddLocation",
                               "stanford.netdb.html.Location");
            context.addServlet("/ModifyLocation",
                               "stanford.netdb.html.Location");
            context.addServlet("/LocationsList",
                               "stanford.netdb.html.Location");
            context.addServlet("/Make",
                               "stanford.netdb.html.Make");
            context.addServlet("/AddMake",
                               "stanford.netdb.html.Make");
            context.addServlet("/ModifyMake",
                               "stanford.netdb.html.Make");
            context.addServlet("/Model",
                               "stanford.netdb.html.Model");
            context.addServlet("/AddModel",
                               "stanford.netdb.html.Model");
            context.addServlet("/ModifyModel",
                               "stanford.netdb.html.Model");
            context.addServlet("/OS",
                               "stanford.netdb.html.OS");
            context.addServlet("/AddOS",
                               "stanford.netdb.html.OS");
            context.addServlet("/ModifyOS",
                               "stanford.netdb.html.OS");
            context.addServlet("/State",
                               "stanford.netdb.html.State");
            context.addServlet("/AddState",
                               "stanford.netdb.html.State");
            context.addServlet("/ModifyState",
                               "stanford.netdb.html.State");
            context.addServlet("/FS",
                               "stanford.netdb.html.FullSearch");
            context.addServlet("/results.csv",
                               "stanford.netdb.html.FullSearch");
            context.addServlet("/Dispatch",
                               "stanford.netdb.html.Dispatch");
            context.addServlet("/List",
                               "stanford.netdb.html.List");

            context.addServlet("/ModifyAdminTeam",
                               "stanford.netdb.html.ModifyAdminTeam");
            context.addServlet("/CreateAdminTeam",
                               "stanford.netdb.html.ModifyAdminTeam");

            context.addServlet("/ModifyNode",
                               "stanford.netdb.html.ModifyNode");
            context.addServlet("/CreateNode",
                               "stanford.netdb.html.ModifyNode");

            context.addServlet("/ModifyNetwork",
                               "stanford.netdb.html.ModifyNetwork");
            context.addServlet("/CreateNetwork",
                               "stanford.netdb.html.ModifyNetwork");
            context.addServlet("/MoveAS",
                               "stanford.netdb.html.MoveAS");

            context.addServlet("/ModifyDomain",
                               "stanford.netdb.html.ModifyDomain");
            context.addServlet("/CreateDomain",
                               "stanford.netdb.html.ModifyDomain");

            context.addServlet("/ModifyGroup",
                               "stanford.netdb.html.ModifyGroup");
            context.addServlet("/CreateGroup",
                               "stanford.netdb.html.ModifyGroup");

            context.addServlet("/ModifyUser",
                               "stanford.netdb.html.ModifyUser");
            context.addServlet("/CreateUser",
                               "stanford.netdb.html.ModifyUser");

            context.addServlet("/Delete",
                               "stanford.netdb.html.Delete");

            context.addServlet("/LoginFailed",
                               "stanford.netdb.html.LoginFailed");

            context.addServlet("/FailTest",
                               "stanford.netdb.html.FailTest");

            context.addServlet("/Profile",
                               "stanford.netdb.html.Profile");

            context.addServlet("/stat",
                               "stanford.netdb.html.Stat");

        } catch (Exception ex) {
            Utils.error("Failed to register servlets. "+ex);
            System.exit(-1);
        }

        // add context
        server.addContext(context);

        // Start the http server
        try {
            server.start();
        } catch (Exception ex) {
            Utils.error("Failed to start server. "+ex);
            System.exit(-1);
        }
    }
}
