/************************************************************************
 *
 * WhoIsServer.java -
 *
 * $Id: WhoIsServer.java,v 1.3 2006/09/27 19:28:13 priimak Exp $
 *
 *-----------------------------------------------------------------------
 * Copyright (c) 2006, Board of Trustees, Leland Stanford Jr. University
 ************************************************************************/
package stanford.netdb.whois;

import java.util.*;
import java.net.*;
import java.io.*;
import java.lang.*;
import stanford.netdb.lib.*;
import stanford.netdb.utils.*;
import su.netdb.parser.Parser;

public class WhoIsServer implements Runnable {

    // all keys which can be present in config file
    public static final String WHOIS_PROPERTIES_FILENAME       = "whois.properties";
    public static final String DBCONNECTION_URL_KEY            = "dbconnection.url";
    public static final String DBCONNECTION_SEARCHUSER_KEY     = "dbconnection.searchuser";
    public static final String DBCONNECTION_SEARCHPASSWORD_KEY = "dbconnection.searchpassword";
    public static final String LISTEN_PORT_KEY                 = "listen.port";
    public static final String HTTP_PORT_KEY                   = "http.port";
    public static final String HTTP_BIND_ADDR_KEY              = "http.bind.addr";
    public static final String HTML_PORT_KEY                   = "html.port";
    public static final String NETDB_UID_KEY                   = "whois.netdb.uid";
    public static final String DO_SYSLOG_KEY                   = "do.syslog";
    public static final String SYSLOG_HOST_KEY                 = "syslog.host";
    public static final String SYSLOG_PORT_KEY                 = "syslog.port";
    public static final String SYSLOG_FACILITY_KEY             = "syslog.facility";
    public static final String SYSLOG_PRIORITY_KEY             = "syslog.priority";
    public static final String UID_KEY                         = "set.euid";
    public static final String GID_KEY                         = "set.egid";
    public static final String USER_KEY                        = "user";
    public static final String PASSWORD_KEY                    = "password";
    public static final String SERVER_KEY                      = "server";
    public static final String DB_KEY                          = "db";
    public static final String ST_WHO_KEY                      = "stanfordwho.url";
    public static final String ST_WHAT_KEY                     = "stanfordwhat.url";
    public static final String LABEL_COLOR_KEY                 = "label.color";
    public static final String SHOW_STATE_KEY                  = "show.state";
    public static final String SHOW_ROOM_KEY                   = "show.room";
    public static final String SHOW_TAGS_KEY                   = "show.custom.fields";
    public static final String ALLOW_KEY                       = "allow";

    // search targets
    public static final int SEARCH_FOR_NAME    = 1;
    public static final int SEARCH_FOR_IP_ADDR = 2;
    public static final int SEARCH_FOR_HW_ADDR = 3;

    // log levels
    public static final int DEBUG = 1;
    public static final int WARN  = 2;
    public static final int ERROR = 3;
    public static final int INFO  = 4;

    // default logger class
    public static stanford.netdb.utils.Logger logger = new DefaultLogger();

    public static String default_domain;

    // print all log messages in stdout
    public static boolean debug = false;

    public static Hashtable ptr_pref = new Hashtable();

    // handle to netdb
    public static NetDBpi netdb = null;

    public  Socket client = null;
    public  Parser parser = new Parser();

    public WhoIsServer(Socket client) {
        this.client = client;
    }

    public static void main(String args[]) {
        if( args.length > 0 && args[0].equals("-d") )
            debug = true; // lots of messages will go to stdout
        else if( args.length > 0 && args[0].equals("-fork") ) {
            // this creates some strange problems on some machines.
            // therefore it is not recommended to do so
            if( Posix.fork() != 0 )
                return;
        }

        // disable SIGHUP
        Posix.disableSIGHUP();

        // print pid, needed for manage
        Utils.write_pid_file();

        try {
            loadConfig();
        } catch(Exception ex) {
            error("Failed to load whois properties ["+ex+"]");
            System.exit(1);
        }

        ptr_pref = Utils.parsePTRprefConfigString();

        // get netdb handle
        try {
            netdb =
                NetDBHook.createInstance(System.getProperty(NETDB_UID_KEY, "system/whois"), false);
        } catch (Exception ex) {
            error("Failed to instantiate netdb handle ["+ex.getMessage()+"]");
            System.exit(1);
        }

        try {
            default_domain = (String)netdb.getCurrentUser().get("default_domain");
            if( default_domain == null ) default_domain = "";
            default_domain = default_domain.trim().toLowerCase();
        } catch (Exception ex) {
            error("Failed to get default domain ["+ex.getMessage()+"]. Not using any.");
            default_domain = "";
        }

        try {
            String htmlPort = System.getProperty(HTML_PORT_KEY);
            if( htmlPort == null || htmlPort.length() == 0 )
                throw new Exception("html port is not set. html server is not started");

            // try to start html server
            (new Thread(new HtmlServer(Integer.parseInt(htmlPort)))).start();
        } catch(Exception ex) {
            ex.printStackTrace();
            error("Exception while starting html server ["+ex+"]");
        }

        // start whois server
        listen();

        return;
    } // end of main(...)

    /**
     * whois server.
     */
    static void listen() {
        try{
            String port = System.getProperty(LISTEN_PORT_KEY, "43");
            int port_int = 43; // default whois port
            try {
                port_int = Integer.parseInt(port);
            } catch(Exception exx) {
                warn("port value["+port+"] is not integer, using deault port 43");
                port_int = 43;
            }
            info("Listening on port "+port_int);

            // open port
            ServerSocket server = new ServerSocket(port_int); // 43 - whois port

            // change gid and uid
            if( System.getProperty(Constants.GID_KEY) != null ) {
                try {
                    Posix.setgid(System.getProperty(Constants.GID_KEY));
                } catch(Exception ex) {
                    // Something went wrong
                    System.out.println("failed to set gid ["+
                                       System.getProperty(Constants.GID_KEY)+"]");
                }
            }

            if( System.getProperty(Constants.UID_KEY) != null ) {
                try {
                    Posix.setuid(System.getProperty(Constants.UID_KEY));
                } catch(Exception ex) {
                    // Something went wrong
                    System.out.println("failed to set uid ["+
                                       System.getProperty(Constants.UID_KEY)+"]");
                }
            }
            // start connection handler
            for(;;) {
                (new Thread(new WhoIsServer(server.accept()))).start();
            }

        } catch(Exception ex) {
            error(ex.getMessage());
        }
    } // end of listen()

    /**
     * Spawned thread on accept() connection.
     */
    public void run() {
        try {
            InetAddress rAddr = client.getInetAddress();
            String remoteAddr = rAddr.toString();
            String remoteName = ((String)Utils.n2s(rAddr.getHostName())).toLowerCase();
            InputStream istr =  client.getInputStream();
            OutputStream ostr =  client.getOutputStream();
            BufferedReader reader = new BufferedReader(new InputStreamReader(istr));
            PrintWriter writer = new PrintWriter(new OutputStreamWriter(ostr));
            String line = reader.readLine();
            Hashtable parsing_result = new Hashtable();
            boolean no_mx_count = false;
            if( line.length() > 0 && line.charAt(0) == '!' ) {
                no_mx_count = true;
                line = line.substring(1);
            }
            if( System.getProperty("allow") != null &&
                !remoteName.endsWith(System.getProperty("allow")) ) {
                info("denied query ["+line+"] from "+remoteAddr);
                client.close();
                return;
            }
            try {
                parsing_result = parseInputString(parser, line);
            } catch(Exception exx) {
                writer.println(exx.getMessage());
                writer.flush();
                client.close();
                return;
            }

            boolean do_search = false;
            int search_for = -1;

            if( parsing_result.get("hw") != null ) { // hardware address search
                search_for = SEARCH_FOR_HW_ADDR;
                do_search = true;

            } else if( parsing_result.get("name") != null ) { // dns name
                search_for = SEARCH_FOR_NAME;
                do_search = true;

            } else if( parsing_result.get("ip_low") != null ) { // ip address search
                search_for = SEARCH_FOR_IP_ADDR;
                do_search = true;
            }

            if( do_search )
                (new WhoIsFormatter(search(line), search_for)).
                    format(new PrintStream(ostr, true), no_mx_count);
            else {
                writer.println("NO MATCH");
                writer.flush();
            }

            writer.flush(); // just in case
            client.close();

            info("query ["+line+"] from "+remoteAddr);

        } catch(Exception ex) {
            ex.printStackTrace();
            error("error:run: "+ex.getMessage());
        }
    } // end of run()


    /** logs debug message */
    public static void debug(Object msg) {
        doLog(Constants.DEBUG, msg.toString());
    }

    /** logs warning message */
    public static void warn(Object msg) {
        doLog(Constants.WARN, msg.toString());
    }

    /** logs error message */
    public static void error(Object msg) {
        doLog(Constants.ERROR, msg.toString());
    }

    /** logs info message */
    public static void info(Object msg) {
        doLog(Constants.INFO, msg.toString());
    }

    /** Does actuall logging */
    public static void doLog(int level, String msg) {
        if( debug ) {
            System.out.println(msg);
            return;
        }
        StackTraceElement[] st = (new Throwable()).getStackTrace();
        synchronized (logger) {
            switch( level ) {
            case DEBUG:
                logger.debug(st[2].getClassName()+"::"+st[2].getMethodName()+"() "+
                             st[2].getLineNumber()+" | "+msg);
                break;

            case WARN:
                logger.warn(st[2].getClassName()+"::"+st[2].getMethodName()+"() "+
                            st[2].getLineNumber()+" | "+msg);
                break;

            case ERROR:
                logger.error(st[2].getClassName()+"::"+st[2].getMethodName()+"() "+
                             st[2].getLineNumber()+" | "+msg);
                break;

            case INFO:
                logger.info(st[2].getClassName()+"::"+st[2].getMethodName()+"() "+
                            st[2].getLineNumber()+" |  "+msg);
                break;
            }
        }
    }

    public static void loadConfig() throws Exception {
        FileInputStream in = null;
        String exceptionString = null;
        Properties system_props = System.getProperties();
        Properties whois_props = null;
        try {
            // Load all properties from config file.
            whois_props = new Properties();
            in = new FileInputStream(System.getProperty("config.file",
                                                        WHOIS_PROPERTIES_FILENAME));
            whois_props.load(in);
        } finally {
            try {
                if (in != null)
                    in.close();
            } catch (Exception ex) { }
        }

        // override by system properties
        for( Object key : system_props.keySet() )
            whois_props.put(key, system_props.get(key));

        // now override system properties
        for( Object key : whois_props.keySet() )
            system_props.put(key, whois_props.get(key));

        // syslog related data
        if( Boolean.getBoolean(DO_SYSLOG_KEY) ) {
            logger.set(DO_SYSLOG_KEY, new Boolean(true));
            logger.set(SYSLOG_PORT_KEY,
                       new Integer(whois_props.getProperty(SYSLOG_PORT_KEY, "514")));
            logger.set(SYSLOG_HOST_KEY,
                       whois_props.getProperty(SYSLOG_HOST_KEY, "localhost"));
            logger.set(SYSLOG_FACILITY_KEY,
                       whois_props.getProperty(SYSLOG_FACILITY_KEY, "local3"));
            logger.set(SYSLOG_PRIORITY_KEY,
                       whois_props.getProperty(SYSLOG_PRIORITY_KEY, "info"));
        }

        Formatter.st_what_url = System.getProperty(ST_WHAT_KEY, "/search?key=").trim();
        Formatter.st_who_url  = System.getProperty(ST_WHO_KEY, "").trim();
        Formatter.ncolor      = System.getProperty(LABEL_COLOR_KEY, "#990000").trim();
    } // end of loadConfig(...)

    /**
     * This is a wrapper around su.netdb.parser.Parser which provides
     * corrections specific to whoisd.
     */
    public static Hashtable parseInputString(Parser parser, String line) throws Exception {
        Hashtable result = parser.parse(line);
        if( result.get("string") != null )
            throw new Exception("Invalid input string");
        return result;
    }

    public static Hashtable search(String searchStr) throws Exception {
        searchStr = searchStr.trim();
        if( searchStr.endsWith(".") )
            searchStr = searchStr.substring(0, searchStr.length() - 1);
        Hashtable result = netdb.simpleSearch(Constants.NODE, searchStr);
        return Utils.appendHash(result, netdb.simpleSearch(Constants.NETWORK, searchStr));
    }
}
